/*global apex*/
/* jshint forin: false */
/*!
 Actions - encapsulate action identity and state for use in keyboard shortcuts, buttons, and menus
 Copyright (c) 2014, 2017, Oracle and/or its affiliates. All rights reserved.
 */
/**
 * @fileOverview
 * An Action encapsulates the identity, state and behavior of a named operation or procedure that the user
 * initiates directly. Actions are most useful when an operation can be initiated in multiple ways such as with a button
 * or toolbar button, menu, or keyboard shortcut. The operation should be labeled consistently and if it can be
 * enabled and disabled that state must be kept consistent. By using an Action and then associating a button and/or
 * menu item with that action all aspects of the action are centralized and kept in sync. This avoids duplicating
 * labels, icons etc.
 *
 * @typedef {Object} action
 *
 * There are 3 kinds of actions:
 * - action: this is typically associated with a button or action menu item. The action must have an action function.
 * - toggle: this is typically associated with a checkbox input, button, or toggle menu item. The action must have
 *      get and set functions and not have a choices property. Toggle actions update an external Boolean state variable
 *      by means of the get and set functions. It is also possible to keep the state in the action by using 'this'
 *      in the get and set functions.
 * - radio group: this is typically associated with radio inputs, select, or radioGroup menu item. The action
 *      must have get and set functions and a choices property. Radio group actions update an external state variable
 *      with the currently selected value of the group by means of the get and set functions. It is also possible to
 *      keep the state in the action by using 'this' in the get and set functions.
 *
 * An action object has these properties:
 * @prop name {string} (required) a unique name for the action. By convention the style of names uses a dash to separate
 * words as in "clear-log".
 * @prop label {string} translatable label for action used in buttons, menus etc. Note: if this is a
 * radio group action (action has choices property) the label is optional. It is used in results of the list
 * and listShortcuts methods. Depending on what kind of UI control the action is bound to it may be used as a label
 * for the whole group. For example using aria-label.
 * @prop icon {string} icon for action may be used in buttons and menus
 * @prop iconType {string} icon type. defaults to a-Icon. Updates to the iconType
 *                         may not be supported by all control types that can be associated with actions
 * @prop disabled {boolean} true if the action is disabled and false if it is enabled. Default is enabled
 * @prop hide {boolean} true if UI controls connected to this action should be hidden and false otherwise.
 *                   Default is false (show)
 * @prop title {string} used as the title attribute when appropriate
 * @prop shortcut {string} keyboard shortcut to invoke action (not allowed for radio group actions)
 * These next properties are for action actions
 * @prop href {string} for actions that navigate set href to the URL and don't set an action function
 * @prop action {function} function(event, focusElement) return true if focus set
 * The following properties are for toggle and radio group actions only
 * @prop get {function} for toggle actions this function should return true or false. For radio group actions this
 *                      should return the current value.
 * @prop set {function} for toggle actions this receives a boolean value. For radio group actions this function receives
 *                      the new value.
 * @prop choices {Array} Only for radio group actions. Array of objects. Each object has properties:
 *                       label, value, icon, iconType, shortcut, disabled, group (for select lists only)
 * @prop labelClasses Only for radio group actions. Classes to add to all radio labels. This and the next two label
 * properties are only used when rendering radio group choices.
 * @prop labelStartClasses Only for radio group actions. Classes to add to last radio label
 * @prop labelEndClasses Only for radio group actions. Classes to add to last radio label
 * @prop itemWrapClasses Only for radio group actions. Classes to add to a span wrapper element. Or to change the
 *                       span use one of these prefixes: p:, li:, div:, span:
 *                       for example "li:myRadio"
 * @prop onLabel {string} Only for dynamic antonyms toggle actions. This is the label when the value is true
 * @prop offLabel {string} Only for dynamic antonyms toggle actions. This is the label when the value is false
 *
 * Note: The disabled and hide properties cannot be functions. Menu widget can use actions and non-action based menu
 * items allow hide and disabled to be functions. But when a menu uses an action that action still must not use
 * functions for disabled and hide.
 *
 * As an alternative to label (or onLabel, offLabel) you can specify labelKey (or onLabelKey, offLabelKey) and
 * the apex.lang.getMessage function will be used to lookup the label text. The localized label text is then stored in
 * the normal label/onLabel/offLabel property. This happens when the action is added. The same applies to titleKey
 * groupKey, and labelKey of each object in the choices array.
 *
 * The apex.actions singleton manages all the global (page level) actions. For components that can have multiple
 * instances on a page the global actions will not work because it is not clear which instance of the component the
 * action applies to. To support components the apex.actions.createContext function is used to create an actions
 * object that is scoped to a specific component instance (the context). Typically the component (e.g. widget) would
 * call createContext when it is created and removeContext when it is destroyed.
 *
 * For global actions and any other created actions contexts the methods on the actions object are used to add,
 * remove, lookup, and invoke actions. There are also methods to manage keyboard shortcuts. Additional state can be
 * stored in the action if desired. If any of the Action properties change then apex.actions.update must be called.
 *
 * Actions are associated with other controls that invoke the action. It is also possible to invoke
 * the action explicitly with the apex.actions.invoke method. Or for toggle actions the toggle method is used
 * and for radio group actions the set method is used to change the value. The following describes how to associate
 * actions with various controls.
 *
 * Buttons:
 * To associate a button element with an action give it a class of js-actionButton and a data-action
 * attribute with the name of the action as its value. The button icon, label text, title, aria-label, hide/show,
 * and disabled state are all updated automatically.
 * For this automatic updating to work buttons should use the following classes:
 *   t-Button-label if a button has a text label this class should be on an element that wraps the text.
 *       This is useful when the button also has an icon or other non text label content. This class does not
 *       go on the button element. If this class is not used then the content of the button element will be the
 *       label text.
 *   t-Button--icon if a button has an icon this class should be on the button element. If the action has an
 *       icon and the button has this class then any elements with the icon type class will be updated with
 *       the icon. Any classes on the icon element that are not the icon, the icon type or start with "t-"
 *       will get removed.
 *   t-Button--noLabel if a button has no visible label this class should be on the button element. A button with
 *       no visible label text will have the button's aria-label attribute set to the button label. Also if there
 *       is no title the label will be used as the title.
 * If the action label or title are null they will be initialized with the text and title attribute value respectively
 * from the first button (in document order) associated with the action. This is useful if the server has
 * already rendered a localized button for the action. The title comes from the button title attribute. The label comes
 * from the first found of; aria-label attribute, title attribute if button has class t-Button--noLabel,
 * content of the descendent element with class t-Button-label, and finally the button element content. If disabled
 * is null it will be taken from the button disabled property. If you don't want the label, title, or icon to be updated
 * add attribute data-no-update="true".
 *
 *    Example:
 *      <button class="js-actionButton" data-action="undo">Undo</button>
 *
 * Checkboxes:
 * A checkbox can be associated with a toggle action by giving the input element (or a wrapping parent element)
 * a class of js-actionCheckbox and a data-action attribute with the name of the action as its value. The checkbox
 * should have a label element. The icon, label text, title, hide/show, disabled state, and checked state are all
 * updated automatically. For hide to work correctly use a wrapping element.
 * If the label has the class t-Button then it should be marked up like a button and the same classes described for
 * a button are used to update the label and icon (except a visually hidden child element is used for the label
 * in place of aria-label). Otherwise the label element content will be updated with the label and the icon is not used.
 * If the action label or title are null they will be initialized from the markup. If the checkbox label is marked up
 * like a button then the label comes from the text of a child element with class t-Button-label or if the label
 * has class t-Button--noLabel then from a child element with class u-VisuallyHidden. If you don't want the label,
 * title, or icon to be updated add attribute data-no-update="true".
 *
 *    Example:
 *      <input id="abc" type="checkbox" class="js-actionButton" data-action="optionA"><label for="abc">Option A</label>
 *    or
 *      <div class="js-actionButton" data-action="optionA">
 *          <input id="abc" type="checkbox"><label for="abc">Option A</label>
 *      </div>
 *
 * Radio groups:
 * A radio group is a set of input elements of type radio and sharing the same name value. A radio group can be
 * associated with a radio group action by giving the element that wraps the radio group a class of
 * js-actionRadioGroup and a data-action attribute with the name of the action as its value. The wrapper element
 * aria label is kept in sync with the action label. The radio group as a whole does not have a disabled state,
 * icon or title. When the radio action value changes (or when update is called) the checked state (and disabled state)
 * of each radio input is updated.
 * The element with class js-actionRadioGroup can also have attributes: data-item-start, data-item, data-item-end,
 * data-item-wrap to override action properties labelStartClasses, labelClasses, labelEndClasses, and itemWrapClasses
 * respectively.
 * If the action label is null it will be initialized from the wrapper element aria-label. If the wrapping element
 * has no children when the action is added or when updateChoices is called (after the choices have been changed) then
 * the choices are rendered as radio input, label pairs elements. The action labelStartClasses, labelClasses,
 * labelEndClasses values are used for the classes of the label elements. If there is an icon it will be used
 * as the label. The label will be included as a hidden but accessible label. The label element will have a title
 * if there is a title property or if the choice has an icon. The title comes from the label property if the title
 * property isn't given.
 *
 *    Example:
 *      <div class="js-actionRadioGroup" data-action="itemSize">
 *        <input id="lc1" type="radio" name="RG1" value="s"><label for="lc1">Small</label>
 *        <input id="lc2" type="radio" name="RG1" value="m"><label for="lc2">Medium</label>
 *        <input id="lc3" type="radio" name="RG1" value="l"><label for="lc3">Large</label>
 *      </div>
 *
 * Select List:
 * To associate a select element with a radio group action give it a class of js-actionSelect and a data-action
 * attribute with the name of the action as its value. Select lists used with actions are assumed to not have an
 * associated label element. The select element aria label, title, value, and disabled state are kept in sync with the
 * action. When the radio action value changes (or when update is called) the select element value is updated and
 * also the disabled state of each option element.
 * If the action label or title are null they will be initialized from the select element aria-label and
 * title attributes. If the select element has no children when the action is added or when updateChoices is called
 * (after the choices have been changed) then the choices are rendered as option elements. The choice group property
 * is used to put options in optgroup elements. The group property value is used as the optgroup label. The choices
 * need to be sorted first by group value.
 *
 *    Example:
 *      <select class="js-actionSelect" data-action="itemSize">...</select>
 *
 * Menu items:
 * For menu items of type action or toggle simply specify the action name as the value of the action property.
 * Values for label, icon, iconType, disabled, hide, and accelerator are taken from the action (accelerator is taken from
 * the action shortcut property). It is possible to override action values such as label and icon by specifying them
 * in the menu item.
 *    Examples:
 *      { type: "action", action: "undo" },
 *      { type: "toggle", action: "myToggleAction" },
 *      { type: "radioGroup", action: "myRadioAction" }
 *
 * Shortcuts:
 * Shortcuts are not an actual widget. The keyboard event handler for invoking actions in response to shortcut keys
 * is in this module and is registered on the body element. The shortcut name must be given in the following format:
 *   [Ctrl+][Alt+][Meta+][Shift+]<key>
 * Order and case is important. See mapKeyToName for a list of keys. Key names and modifiers are not localized.
 * The primary shortcut for an action is specified in the shortcut property of the action object. This is so that
 * it can be shown in associated menu items. Additional shortcuts can be added with addShortcut.
 *
 * To integrate actions with other UI controls:
 * - devise a way to specify the action name
 * - register an observer call back to get notified when the action is added, removed, or updated
 * - call the invoke method when it is time to invoke the action
 *
 * todo:
 * - menubar support for action menu items in menu bar
 * - shortcut configuration, get/setKeyBindings, THINK is there a need to disable all shortcuts permanently?
 *
 * Depends on these strings being defined in the apex.lang message facility
 *     APEX.ACTIONS.TOGGLE
 *
 * Depends:
 *      debug.js
 *      lang.js
 *      util.js
 *      navigation.js
 */
(function ( apex, debug, lang, util, $, undefined ) {
    "use strict";

    var SEL_ACTION_BUTTON = ".js-actionButton",
        SEL_ACTION_RADIO_GROUP = ".js-actionRadioGroup",
        SEL_ACTION_SELECT = ".js-actionSelect",
        SEL_ACTION_CHECKBOX = ".js-actionCheckbox",
        A_LABEL = "aria-label",
        ACTION_SHORTCUT_SEPARATOR = ">";

    var mapKeyToName = {
        6: "Help",
        8: "Backspace",
        9: "Tab",
        13: "Enter",
        27: "Escape",
        32: "Space",
        33: "Page Up",
        34: "Page Down",
        35: "End",
        36: "Home",
        37: "Left",
        38: "Up",
        39: "Right",
        40: "Down",
        45: "Insert",
        46: "Delete",
        48: "0",
        49: "1",
        50: "2",
        51: "3",
        52: "4",
        53: "5",
        54: "6",
        55: "7",
        56: "8",
        57: "9",
        65: "A",
        66: "B",
        67: "C",
        68: "D",
        69: "E",
        70: "F",
        71: "G",
        72: "H",
        73: "I",
        74: "J",
        75: "K",
        76: "L",
        77: "M",
        78: "N",
        79: "O",
        80: "P",
        81: "Q",
        82: "R",
        83: "S",
        84: "T",
        85: "U",
        86: "V",
        87: "W",
        88: "X",
        89: "Y",
        90: "Z",
        96: "Keypad 0",
        97: "Keypad 1",
        98: "Keypad 2",
        99: "Keypad 3",
        100: "Keypad 4",
        101: "Keypad 5",
        102: "Keypad 6",
        103: "Keypad 7",
        104: "Keypad 8",
        105: "Keypad 9",
        106: "Keypad *",
        107: "Keypad +",
        109: "Keypad -",
        110: "Keypad .",
        111: "Keypad /",
        112: "F1",
        113: "F2",
        114: "F3",
        115: "F4",
        116: "F5",
        117: "F6",
        118: "F7",
        119: "F8",
        120: "F9",
        121: "F10",
        122: "F11",
        123: "F12",
        188: "Comma",
        190: "Period",
        191: "/",
        192: "`",
        219: "[",
        220: "\\",
        221: "]"
    };

    var isMac = navigator.appVersion.indexOf("Mac") >= 0;

    function makeKeyName(event) {
        var name = "",
            k = mapKeyToName[event.which];

        // keys that produce a character require at least one modifier other than shift
        if ( k && ((event.which >= 112 && event.which <= 123) || ( event.which >= 33 && event.which <= 46 ) || (event.ctrlKey || event.altKey || event.metaKey)) ) {
            if ( event.ctrlKey ) {
                name += "Ctrl+";
            }
            if ( event.altKey ) {
                name += "Alt+";
            }
            if ( event.metaKey ) {
                name += "Meta+";
            }
            if ( event.shiftKey ) {
                name += "Shift+";
            }
            name += k;
            return name;
        }
        return null;
    }

    function getActionLabel( action ) {
        var label = action.label;
        if ( !label && action.onLabel && action.offLabel ) {
            label = action.onLabel + "/" + action.offLabel;
        } else if ( action.set && action.get && !action.choices ) {
            label = lang.formatMessage( "APEX.ACTIONS.TOGGLE", label );
        }
        return label;
    }

    function getLabelFor( input ) {
        var label$;

        if ( input.id ) {
            label$ = $( "[for=" + input.id + "]" );
        } else {
            label$ = $( input ).closest( "label" );
        }
        return label$;
    }

    function translateMessages( obj, keys ) {
        var i, key, msgKey;
        for ( i = 0; i < keys.length; i++ ) {
            key = keys[i];
            msgKey = key + "Key";
            if ( obj[msgKey] ) {
                obj[key] = lang.getMessage( obj[msgKey] );
                delete obj[msgKey];
            }
        }
    }

    /**
     * A DOM element
     * @typedef {Object} DOMElement
     */

    function makeActionContext( typeName, context ) {
        var actionMap = {},
            actionShortcut = {}, // Map actionName -> shortcut, or actionName + sep + choiceValue -> shortcut
            shortcutMap = {}, // Map shortcut -> actionName, or shortcut -> { actionName: actionName,  value: choiceValue }
            observers = [],
            shortcutsDisabled = false,
            disableDepth = 0;

        function notifyObservers( action, operation ) {
            var i, callback;

            for ( i = 0; i < observers.length; i++ ) {
                callback = observers[i];
                callback( action, operation );
            }
        }

        /**
         * @typedef actions
         * @prop typeName {string}
         * @prop context {DOMElement}
         */
        return {
            typeName: typeName,
            context: context,
            /**
             * Add an action or an array of actions to this actions context.
             * The action name must be unique and the shortcut if any must be unique and valid. Debug warnings are
             * logged if any of these conditions are not met.
             *
             * @param pActions {action, action[]} the action or an array of actions to add
             * @return {boolean} true if all the actions and shortcuts are added without errors or warnings,
             *                      false otherwise.
             * @memberOf actions
             */
            add: function( pActions ) {
                var i, c, a, choice,
                status = true;

                if ( !$.isArray( pActions ) ) {
                    pActions = [pActions];
                }
                for ( i = 0; i < pActions.length; i++ ) {
                    a = pActions[i];
                    if ( a.name ) {
                        if ( !actionMap[a.name] ) {
                            actionMap[a.name] = a;
                            if ( a.shortcut ) {
                                if ( !this.isValidShortcut( a.shortcut )) {
                                    debug.warn( "Invalid shortcut '" + a.shortcut + "' for action '" + a.name + "' ignored.");
                                    a.shortcut = null;
                                    status = false;
                                } else if ( a.choices ) {
                                    debug.warn( "Shortcut '" + a.shortcut + "' for radio group action '" + a.name + "' ignored.");
                                    a.shortcut = null;
                                    status = false;
                                } else if ( !shortcutMap[a.shortcut] ) {
                                    actionShortcut[a.name] = a.shortcut;
                                    shortcutMap[a.shortcut] = a.name;
                                } else {
                                    debug.warn( "Duplicate shortcut '" + a.shortcut + "' for action '" + a.name + "' ignored.");
                                    a.shortcut = null;
                                    status = false;
                                }
                            }
                            translateMessages( a, ["label", "onLabel", "offLabel", "title"] );

                            if ( a.choices ) {
                                for ( c = 0; c < a.choices.length; c++ ) {
                                    choice = a.choices[c];

                                    if ( choice.shortcut ) {
                                        if ( !this.isValidShortcut( choice.shortcut )) {
                                            debug.warn( "Invalid shortcut '" + choice.shortcut + "' for action '" + a.name + "' ignored.");
                                            choice.shortcut = null;
                                            status = false;
                                        } else if ( !shortcutMap[choice.shortcut] ) {
                                            actionShortcut[a.name + ACTION_SHORTCUT_SEPARATOR + choice.value] = choice.shortcut;
                                            shortcutMap[choice.shortcut] = { actionName: a.name, value: choice.value };
                                        } else {
                                            debug.warn( "Duplicate shortcut '" + choice.shortcut + "' for action '" + a.name + "' ignored.");
                                            choice.shortcut = null;
                                            status = false;
                                        }
                                    }

                                    translateMessages( choice, ["label", "group", "title"] );
                                }
                            }
                            notifyObservers( a, "add" );
                        } else {
                            debug.warn( "Duplicate action '" + a.name + "' ignored.");
                            status = false;
                        }
                    }
                }
                return status;
            },

            /**
             * Add one or more actions from simple list markup. This is useful in cases where it is easier
             * to render list markup than an array of action objects.
             * The markup expected by this method overlaps with what a menu expects. Keep this code in sync with menu widget.
             *
             * Expected markup:
             * An element with a ul child. The ul has one or more li elements each one representing an action.
             * The li element can have either an "a" or span element.
             * Action property      Comes from
             * name                 li[data-id]
             * label                a or span content
             * title                a[title] or span[title]
             * href                 a[href]
             * disabled             true iff li[data-disabled=true]
             * hide                 true iff li[data-hide=true]
             * shortcut             li[data-shortcut]
             * icon                 li[data-icon] if the value has a space the icon is the word after the space
             *                      otherwise it is the whole value
             * iconType             li[data-icon] if the value has a space the type is the word before the space
             *
             * If there is no name or label or the value of href equals "separator" then no action is created for that li.
             * If the li has a ul child the ul is processed recursively.
             *
             * @param pList$
             * @memberOf actions
             */
            addFromMarkup: function ( pList$ ) {
                var self = this;

                pList$.children( "ul" ).children( "li" ).each(function() {
                    var action, icon, index,
                        title = null,
                        $item = $( this ),
                        a$ = $item.children( "a" ).eq( 0 ),
                        span$ = $item.children( "span" ).eq( 0 );

                    action = { };
                    if ( a$.length > 0 ) {
                        action.label = a$.text();
                        action.href = a$.attr("href");
                        title = a$.attr("title");
                    } else if ( span$.length > 0 ) {
                        action.label = span$.text();
                        title = span$.attr("title");
                    }
                    if ( title ) {
                        action.title = title;
                    }
                    action.name = $item.attr( "data-id" );
                    if ( $item.attr( "data-hide" ) === "true" ) {
                        action.hide = true;
                    }
                    if ( $item.attr("data-disabled") === "true" ) {
                        action.disabled = true;
                    }
                    action.shortcut = $item.attr("data-shortcut") || null;

                    icon = $item.attr("data-icon");
                    if ( icon ) {
                        index = icon.indexOf( " " );
                        if ( index >= 0 ) {
                            action.iconType = icon.substring( 0, index );
                            action.icon = icon.substring( index + 1 );
                        } else {
                            action.icon = icon;
                        }
                    }
                    // without a name or a label there is no action also separators are not actions
                    if ( action.name && action.label && action.href !== "separator" ) {
                        self.add( action );
                    }
                    if ( $item.children( "ul" ).length > 0 ) {
                        self.addFromMarkup( $item );
                    }
                });
            },

            /**
             * Remove one or more actions from this actions context.
             *
             * @param pActions {action | string | (action | string)[]} the action or action name or an array of actions
             *                                      or action names to remove
             * @memberOf actions
             */
            remove: function( pActions ) {
                var i, c, choice, a, name, shortcut;

                if ( !$.isArray( pActions ) ) {
                    pActions = [pActions];
                }
                for ( i = 0; i < pActions.length; i++ ) {
                    a = pActions[i];
                    name = typeof a === "string" ? a : a.name;
                    if ( name && actionMap[name] ) {
                        a = actionMap[name];
                        shortcut = a.shortcut;
                        if ( shortcut ) {
                            delete shortcutMap[shortcut];
                        }
                        if ( a.choices ) {
                            // go through choices to clear out any shortcuts
                            for ( c = 0; c < a.choices.length; c++ ) {
                                choice = a.choices[c];
                                shortcut = choice.shortcut;
                                if ( shortcut ) {
                                    delete shortcutMap[shortcut];
                                    delete actionShortcut[name + ACTION_SHORTCUT_SEPARATOR + choice.value];
                                }
                            }
                        }
                        delete actionMap[name];
                        delete actionShortcut[name];
                        notifyObservers( a, "remove" );
                    }
                }
            },

            /**
             * Remove all actions from this actions context.
             *
             * @memberOf actions
             */
            clear: function() {
                var name, a;

                // unfiltered loop OK because actionMap is private
                for ( name in actionMap ) {
                    a = actionMap[name];
                    notifyObservers( a, "remove" );
                }
                actionMap = {};
                actionShortcut = {};
                shortcutMap = {};
            },

            /**
             * Lookup and return an action by name.
             *
             * @param pActionName {string} name of the action to return
             * @return {action} action or undefined if action doesn't exist
             * @memberOf actions
             */
            lookup: function( pActionName ) {
                return actionMap[pActionName];
            },

            /**
             * Return an array of actionName, label pairs for all actions.
             *
             * @return {Array} array of objects with name and label properties.
             * @memberOf actions
             */
            list: function() {
                var i, choice, actionName, action, actionLabel, label,
                    actionsList = [];

                // unfiltered loop OK because actionMap is private
                for ( actionName in actionMap ) {
                    action = actionMap[actionName];
                    actionLabel = getActionLabel( action );
                    if ( action.choices ) {
                        for ( i = 0; i < action.choices.length; i++ ) {
                            choice = action.choices[i];
                            if ( actionLabel ) {
                                label = actionLabel + ": " + choice.label;
                            } else {
                                label = choice.label;
                            }
                            actionsList.push({
                                name: actionName,
                                choice: choice.value,
                                label: label
                            });
                        }
                    } else {
                        actionsList.push({
                            name: actionName,
                            label: actionLabel
                        });
                    }
                }
                return actionsList;
            },

            /**
             * Update any UI elements associated with the action after it changes. Calling update will
             * notify any observers that the action has changed. Debug warnings will be logged and
             * the return value is false if the action has a problem with the shortcut.
             *
             * @param pActionName {string} name of the action to update
             * @return {Boolean} false if the shortcut is invalid or a duplicate and true otherwise.
             * @memberOf actions
             */
            update: function( pActionName ) {
                var i, choice, shortcutKey,
                    a = actionMap[pActionName],
                    oldShortcut = actionShortcut[pActionName],
                    status = true;

                if ( a.shortcut !== oldShortcut ) {
                    delete shortcutMap[oldShortcut];
                    if ( !a.shortcut ) {
                        // remove shortcut
                        delete actionShortcut[pActionName];
                    } else {
                        if ( !this.isValidShortcut( a.shortcut )) {
                            debug.warn( "Invalid shortcut '" + a.shortcut + "' for action '" + pActionName + "' ignored.");
                            a.shortcut = null;
                            status = false;
                        } else if ( !shortcutMap[a.shortcut] ) {
                            actionShortcut[pActionName] = a.shortcut;
                            shortcutMap[a.shortcut] = pActionName;
                        } else {
                            debug.warn( "Duplicate shortcut '" + a.shortcut + "' for action '" + pActionName + "' ignored.");
                            a.shortcut = null;
                            status = false;
                        }
                    }
                }
                translateMessages( a, ["label", "onLabel", "offLabel", "title"] );

                // go through choices to handle updates to choice shortcuts
                if ( a.choices ) {
                    for ( i = 0; i < a.choices.length; i++ ) {
                        choice = a.choices[i];
                        shortcutKey = pActionName + ACTION_SHORTCUT_SEPARATOR + choice.value;
                        oldShortcut = actionShortcut[shortcutKey];
                        if ( choice.shortcut !== oldShortcut ) {
                            delete shortcutMap[oldShortcut];
                            if ( !choice.shortcut ) {
                                // remove shortcut
                                delete actionShortcut[shortcutKey];
                            } else {
                                if ( !this.isValidShortcut( choice.shortcut )) {
                                    debug.warn( "Invalid shortcut '" + choice.shortcut + "' for action '" + pActionName + "' ignored.");
                                    choice.shortcut = null;
                                    status = false;
                                } else if ( !shortcutMap[choice.shortcut] ) {
                                    actionShortcut[shortcutKey] = choice.shortcut;
                                    shortcutMap[choice.shortcut] = { actionName: pActionName, value: choice.value };
                                } else {
                                    debug.warn( "Duplicate shortcut '" + choice.shortcut + "' for action '" + pActionName + "' ignored.");
                                    choice.shortcut = null;
                                    status = false;
                                }
                            }
                        }
                        translateMessages( choice, ["label", "group", "title"] );
                    }
                }

                notifyObservers( a, "update" );
                return status;
            },

            /**
             * Call this only if the set of choices for an action has changed. This will
             * notify any observers that the set of action choices has changed.
             *
             * @param pActionName {string} name of the action that has had its choices updated
             * @return {Boolean} false if the action has no choices and true otherwise
             * @memberOf actions
             */
            updateChoices: function( pActionName ) {
                var a = actionMap[pActionName],
                    status = false;

                if ( a.choices ) {
                    notifyObservers( a, "updateChoices" );
                    status = true;
                }
                return status;
            },

            /**
             * Enable UI elements associated with the action by setting disabled property to false.
             * This is a convenience method to enable without having to do lookup and update.
             *
             * @param pActionName {string} name of the action
             * @memberOf actions
             */
            enable: function( pActionName ) {
                var a = actionMap[pActionName];
                if ( a && a.disabled ) {
                    a.disabled = false;
                    // the shortcut didn't change so no need to call update
                    notifyObservers( a, "update" );
                }
            },

            /**
             * Disable UI elements associated with the action by setting disabled property to true.
             * This is a convenience method to disable without having to do lookup and update.
             *
             * @param pActionName {string} name of the action
             * @memberOf actions
             */
            disable: function( pActionName ) {
                var a = actionMap[pActionName];
                if ( a && !a.disabled ) {
                    a.disabled = true;
                    // the shortcut didn't change so no need to call update
                    notifyObservers( a, "update" );
                }
            },

            /**
             * Hide UI elements associated with the action by setting the hide property to true.
             * This is a convenience method to hide without having to do lookup and update.
             *
             * @param pActionName {string} name of the action
             * @memberOf actions
             */
            hide: function( pActionName ) {
                var a = actionMap[pActionName];
                if ( a && !a.hide ) {
                    a.hide = true;
                    // the shortcut didn't change so no need to call update
                    notifyObservers( a, "update" );
                }
            },

            /**
             * Show UI elements associated with the action by setting the hide property to false.
             * This is a convenience method to show without having to do lookup and update.
             *
             * @param pActionName
             * @memberOf actions
             */
            show: function( pActionName ) {
                var a = actionMap[pActionName];
                if ( a && a.hide ) {
                    a.hide = false;
                    // the shortcut didn't change so no need to call update
                    notifyObservers( a, "update" );
                }
            },

            /**
             * Invoke the named action.
             *
             * @param pActionName {string} name of the action to invoke
             * @param pEvent {Event} Optional. Browser event that caused the action to be invoked
             * @param pFocusElement {DOMElement}
             * @return {boolean} false if there is no such action or action has no action method, true if action set the focus
             * all other cases should return undefined
             * @memberOf actions
             */
            invoke: function( pActionName, pEvent, pFocusElement ) {
                var a = actionMap[pActionName];

                if ( a && ( a.action || a.href ) ) {
                    if ( !a.disabled && !a.hide ) {
                        if ( a.action ) {
                            return a.action( pEvent, pFocusElement );
                        } else {
                            // it must be href
                            apex.navigation.redirect( a.href );
                            return true;
                        }
                    }
                } else {
                    debug.error( "No such action '" + pActionName + "' or action can't be invoked." );
                    return false;
                }
                // intentionally "return" undefined here
            },

            /**
             * Toggle the named action. This should only be used for toggle actions.
             * Toggle actions have get and set methods and don't have a choices property.
             *
             * @param pActionName {string} name of the action
             * @return {Boolean} false if there is no such action or action doesn't have get/set methods
             * all other cases should return undefined
             * @memberOf actions
             */
            toggle: function( pActionName ) {
                var value,
                    a = actionMap[pActionName];

                if ( a && a.get && a.set && !a.choices ) {
                    if ( !a.disabled && !a.hide ) {
                        value = !a.get();
                        a.set( value );
                        this.update( pActionName );
                    }
                } else {
                    debug.error( "No such action '" + pActionName + "' or action cannot be toggled." );
                    return false;
                }
                // intentionally "return" undefined here
            },

            /**
             * Return the current value of a radio group action.
             *
             * @param pActionName {string} name of the action
             * @return {*} typically a string
             */
            get: function( pActionName ) {
                var a = actionMap[pActionName];

                if ( a && a.get ) {
                    return a.get();
                } else {
                    debug.error( "No such action '" + pActionName + "' or cannot get action value." );
                    return null;
                }
            },

            /**
             * Set the value of a radio group action.
             *
             * @param pActionName {string} name of the action
             * @param pValue {*} value to set typically a string
             */
            set: function( pActionName, pValue ) {
                var a = actionMap[pActionName];

                if ( a && a.get && a.set ) {
                    if ( !a.disabled && !a.hide ) {
                        // if setting a toggle action force the value to be boolean
                        if ( !a.choices ) {
                            pValue = !!pValue;
                        }
                        // otherwise allow setting any value even though it probably doesn't make much sense to
                        // set a value that isn't the value of one of the choices.
                        a.set( pValue );
                        this.update( pActionName );
                    }
                } else {
                    debug.error( "No such action '" + pActionName + "' or cannot set action value." );
                }
            },

            /**
             * Return the action that a keyboard shortcut will invoke. For actions that select
             * a radio group choice the return value is an object with properties actionName and value.
             *
             * @param pShortcutName {string} the keyboard shortcut to lookup the action
             * @return {string|Object} the action name that the shortcut invokes or an object for selecting
             * a radio group action or undefined if there is no such shortcut name or if shortcuts are disabled.
             */
            lookupShortcutAction: function( pShortcutName )  {
                if ( !shortcutsDisabled ) {
                    return shortcutMap[pShortcutName];
                }
            },

            /**
             * Add a keyboard shortcut synonym for an action. Debug warnings are logged if there are problems.
             *
             * @param pShortcutName {string} keyboard shortcut synonym to add
             * @param pActionName {string} name of the action to add a shortcut for
             * @param pChoiceValue {string} optional choice value only if the action is a radio group the shortcut
             *                              will select the given choice
             * @return {boolean} true if successful and false if there is no such action or there is a duplicate shortcut.
             */
            addShortcut: function( pShortcutName, pActionName, pChoiceValue ) {
                var i, found,
                    action = actionMap[pActionName];

                if ( !action ) {
                    debug.warn( "No such action '" + pActionName + "'.");
                    return false;
                }
                if ( !shortcutMap[pShortcutName] ) {
                    if ( action.choices ) {
                        if ( !pChoiceValue ) {
                            debug.warn( "Shortcut '" + pShortcutName + "' for radio group action '" + pActionName + "' ignored.");
                            return false;
                        }
                        // make sure choice is valid
                        found = false;
                        for ( i = 0; i < action.choices.length; i++ ) {
                            if ( action.choices[i].value === pChoiceValue ) {
                                found = true;
                                break;
                            }
                        }
                        if ( !found ) {
                            debug.warn( "No such choice '" + pChoiceValue + " for action '" + pActionName + "'.");
                            return false;
                        }
                        shortcutMap[pShortcutName] = { actionName: pActionName, value: pChoiceValue };
                    } else {
                        shortcutMap[pShortcutName] = pActionName;
                    }
                } else {
                    debug.warn( "Duplicate shortcut '" + pShortcutName + "' for action '" + pActionName + "' ignored.");
                    return false;
                }
                return true;
            },

            /**
             * Remove a keyboard shortcut synonym for an action.
             *
             * @param pShortcutName {string} keyboard shortcut synonym to remove
             * @return {boolean} true if successful false if the shortcut is the primary shortcut for an action
             */
            removeShortcut: function( pShortcutName ) {
                var pActionName;

                pActionName = shortcutMap[pShortcutName];
                if ( pActionName && typeof pActionName !== "string" ) {
                    pActionName = pActionName.actionName + ACTION_SHORTCUT_SEPARATOR + pActionName.value;
                }
                if (actionShortcut[pActionName] === pShortcutName ) {
                    debug.warn( "Can't delete primary for action.");
                    return false;
                }
                delete shortcutMap[pShortcutName];
                return true;
            },

            /**
             * Return a list of all shortcuts.
             * @return {shortcutListItem[]}
             */
            listShortcuts: function() {
                var i, shortcutName, actionName, action, choiceValue, item, label,
                    shortcuts = [];

                // unfiltered loop OK because shortcutMap is private
                for (shortcutName in shortcutMap) {
                    actionName = shortcutMap[shortcutName];
                    if ( typeof actionName !== "string" ) {
                        choiceValue = actionName.value;
                        actionName = actionName.actionName;
                    }
                    action = actionMap[actionName];

                    /**
                     * @typedef {Object} shortcutListItem
                     * @prop shortcut {string}
                     * @prop shortcutDisplay {string}
                     * @prop actionName {string}
                     * @prop actionLabel {string}
                     */
                    item = {
                        shortcut: shortcutName,
                        shortcutDisplay: this.shortcutDisplay( shortcutName ),
                        actionName: actionName
                    };

                    if ( choiceValue && action.choices ) {
                        item.choice = choiceValue;

                        for ( i = 0; i < action.choices.length; i++ ) {
                            if ( action.choices[i].value === choiceValue ) {
                                label = getActionLabel( action );
                                if ( label ) {
                                    label += ":" + action.choices[i].label;
                                } else {
                                    label = action.choices[i].label;
                                }
                                item.actionLabel = label;
                                break;
                            }
                        }
                    } else {
                        item.actionLabel = getActionLabel( action );
                    }
                    shortcuts.push(item);
                }
                return shortcuts;
            },

            /**
             * Return the friendly display string for a keyboard shortcut name.
             *
             * @param pShortcutName {string} keyboard shortcut to get the display string for
             * @return {String} friendly version of the shortcut
             */
            shortcutDisplay: function( pShortcutName ) {
                var i,
                    display = "",
                    parts = pShortcutName.split("+");

                for ( i = 0; i < parts.length; i++ ) {
                    if ( i > 0 ) {
                        display += "+";
                    }
                    if ( isMac ) {
                        if ( parts[i] === "Alt" ) {
                            display += "Option";
                        } else if ( parts[i] === "Meta" ) {
                            display += "\u2318";
                        } else {
                            display += parts[i];
                        }
                    } else {
                        display += parts[i];
                    }
                }
                return display;
            },

            /**
             * This is used to disable all shortcuts temporarily. Call at the start of a user interaction
             * that should have shortcuts disabled for example a custom popup. Call enableShortcuts when
             * finished. It is called automatically when modal dialogs or menus open.
             * Calls can be nested. For each call to disableShortcuts there should be a corresponding
             * call to enableShortcuts.
             */
            disableShortcuts: function() {
                disableDepth += 1;
                if ( disableDepth > 0 ) {
                    shortcutsDisabled = true;
                }
            },

            /**
             * This is used to enable all shortcuts after they were disabled with disableShortcuts.
             * It is called automatically when modal dialogs or menus close.
             * Calls can be nested. For each call to disableShortcuts there should be a corresponding
             * call to enableShortcuts.
             */
            enableShortcuts: function() {
                disableDepth -= 1;
                if ( disableDepth <= 0 ) {
                    shortcutsDisabled = false;
                    disableDepth = 0;
                }
            },

            /**
             * Register callback to be notified when an action changes.
             *
             * @param pCallback {function} function notifyObservers( action, operation )
             *                              operation is one of "add", "remove", "update", or "updateChoices"
             */
            observe: function( pCallback ) {
                observers.push( pCallback );
            },

            /**
             * Remove callback
             *
             * @param pCallback {function}
             */
            unobserve: function( pCallback ) {
                var i;

                for ( i = 0; i < observers.length; i++ ) {
                    if ( observers[i] === pCallback ) {
                        observers.splice(i, 1);
                        break;
                    }
                }
            },

            isValidShortcut: function( pShortcutName, pCheckIfUsed ) {
                var i, part, nextPos, key,
                    pos = 0,
                    valid = true,
                    parts = pShortcutName.split("+");

                // check all the modifiers
                for ( i = 0; i < parts.length - 1; i++ ) {
                    part = parts[i];
                    if ( part === "Ctrl" ) {
                        nextPos = 1;
                    } else if ( part === "Alt" ) {
                        nextPos = 2;
                    } else if ( part === "Meta" ) {
                        nextPos = 3;
                    } else if ( part === "Shift" ) {
                        nextPos = 4;
                    } else {
                        valid = false;
                        break;
                    }
                    if ( nextPos <= pos ) {
                        valid = false;
                        break;
                    }
                    pos = nextPos;
                }
                if ( valid ) {
                    part = parts[i]; // the key name
                    // now assume not valid
                    valid = false;
                    // find keyname in map
                    for ( key in mapKeyToName ) {
                        if ( mapKeyToName[key] === part) {
                            valid = true;
                            break;
                        }
                    }
                    // make sure keys that require a modifier have one
                    // todo this test could be a little more correct
                    if ( valid && pos === 0 && part.length === 1) {
                        valid = false;
                    }
                    if ( valid && pCheckIfUsed && shortcutMap[pShortcutName] ) {
                        valid = false;
                    }
                }
                return valid;
            }
        };
    }

    // todo allow customizing these selectors?
    var buttonSelector = ".a-Button, .t-Button", // indicates that something is styled like a button and hopefully follows
                                                 // same rules for labels and icons below.
        buttonLabelSelector = ".a-Button-label, .t-Button-label", // how to find where to update the label text
        buttonHasIconSelector = ".a-Button--withIcon, .t-Button--icon", // how to tell if a button has an icon
        buttonNoLabelSelector = ".a-Button--noLabel, .t-Button--noLabel"; // how to tell if a button has no label text

    var gActionContextTypes = {};

    /**
     * Create a new actions type object that is scoped to the given DOM element context.
     * Any action buttons must be within the given pContext. And focus must be within pContext for
     * keyboard shortcuts to work. A global context at document.body is created automatically and is accessed
     * from apex.actions. The global context type name is "global".
     *
     * The type name
     * @param pTypeName {string} type name of the actions context
     * @param pContext {DOMElement} DOM element context that actions are scoped within
     * @return {actions} The actions type object that was created
     * @memberOf apex.actions
     */
    function createContext( pTypeName, pContext ) {
        var typeList,
            context$ = $( pContext ),
            actions =  makeActionContext( pTypeName, pContext );

        // remove it first just in case
        removeContext( pTypeName, pContext );

        typeList = gActionContextTypes[ pTypeName ];
        if ( !typeList ) {
            typeList = [];
            gActionContextTypes[pTypeName] = typeList;
        }
        typeList.push(actions);

        actions.observe(function( action, op ){
            var name = action.name,
                actionElements$ = $( "[data-action="+ name +"]", context$ );

            function updateIcons( el$, iconType, newIcon ) {
                // icons don't typically have classes besides the one for icon type and icon but if they do
                // assume they all start with "t-".
                // todo consider if the prefix or set of classes to keep should be customizable
                // todo this doesn't handle the case where a button has more than one icon
                el$.find( "." + iconType ).each( function() {
                    var icon$ = $( this ),
                        newClasses = [],
                        classes = icon$.attr( "class" );

                    $.each( classes.split( " " ), function( i, c ) {
                        if ( /^t-/.test( c ) ) {
                            newClasses.push( c );
                        }
                    } );
                    newClasses.push( iconType );
                    newClasses.push( newIcon );
                    icon$.attr( "class", newClasses.join( " " ) );
                });
            }

            if ( op !== "remove") {
                // update elements associated with the action
                actionElements$.each( function() {
                    var i, out, choice, lastGroup, id, cls, iconType, label, value, title, input$, label$, textLabel$,
                        idPrefix, rbStartClasses, rbClasses, rbEndClasses, rbWrapElement, rbWrapClasses,
                        elementType, choices$,
                        isButtonLike = false,
                        actionType = "action",
                        el$ = $( this );

                    if ( action.hide ) {
                        el$.hide();
                        return;
                    } // else
                    el$.show();

                    // determine if the action is radio group or toggle and save value for later
                    if ( action.get ) {
                        if ( action.choices ) {
                            actionType = "radio";
                        } else {
                            actionType = "toggle";
                        }
                        value = action.get();
                    }
                    // determine the type of element (ui control)
                    // some combinations of actionType and elementType don't make sense but we don't check for that
                    // it is up the the caller to match up actions to appropriate markup
                    if ( el$.is( SEL_ACTION_SELECT ) ) {
                        elementType = "select";
                    } else if ( el$.is( SEL_ACTION_RADIO_GROUP ) ) {
                        elementType = "radio";
                    } else if ( el$.is( SEL_ACTION_CHECKBOX ) ) {
                        elementType = "checkbox";
                        input$ = el$;
                        if ( input$[0].nodeName !== "INPUT" ) {
                            input$ = el$.find( "input" );
                         }
                        label$ = getLabelFor( input$[0] );
                        if ( label$.is( buttonSelector ) ) {
                            isButtonLike = true;
                            textLabel$ = el$.find( buttonLabelSelector ).first();
                            if ( !textLabel$.length ) {
                                textLabel$ = label$;
                            }
                        } else {
                            textLabel$ = label$;
                        }
                    } else if ( el$.is( SEL_ACTION_BUTTON ) ) {
                        elementType = "button";
                        // if the label is in a .t/.a-Button-label span use that otherwise its the whole text content of the button
                        textLabel$ = el$.find( buttonLabelSelector ).first();
                        if ( !textLabel$.length ) {
                            textLabel$ = el$;
                        }
                    }

                    if ( op === "updateChoices" ||
                            ( op === "add" && actionType === "radio" && el$.children().length === 0 ) ) {

                        out = util.htmlBuilder();
                        if ( elementType === "select" ) {
                            lastGroup = null;
                            for ( i = 0; i < action.choices.length; i++ ) {
                                choice = action.choices[i];
                                if ( choice.hide ) {
                                    continue;
                                }
                                if ( choice.group !== undefined && choice.group !== lastGroup ) {
                                    if ( lastGroup !== null ) {
                                        out.markup("</optgroup>");
                                    }
                                    if ( choice.group !== null ) {
                                        out.markup("<optgroup")
                                            .attr("label", choice.group)
                                            .markup(">");
                                    }
                                    lastGroup = choice.group;
                                }
                                out.markup( "<option" )
                                    .attr( "value", choice.value )
                                    .markup( ">" )
                                    .content( choice.label )
                                    .markup( "</option>" );
                            }
                        } else if ( elementType === "radio" ) {
                            idPrefix = el$.closest( "[id]" )[0].id + "_" + action.name;
                            rbStartClasses = el$.attr( "data-item-start" ) || action.labelStartClasses;
                            rbClasses = el$.attr( "data-item" ) || action.labelClasses;
                            rbEndClasses = el$.attr( "data-item-end" ) || action.labelClasses;
                            rbWrapClasses = el$.attr( "data-item-wrap" ) || action.itemWrapClasses || false;
                            rbWrapElement = false;
                            if (rbWrapClasses ) {
                                rbWrapElement = rbWrapClasses ? "span" : false;
                                if ( rbWrapClasses && rbWrapClasses.indexOf( ":" ) > 0 ) {
                                    rbWrapClasses = rbWrapClasses.split( ":" );
                                    rbWrapElement = rbWrapClasses[0];
                                    rbWrapClasses = rbWrapClasses[1];
                                }
                                // the wrapping element, if there is one, must be one of p, li, span, div
                                if ( rbWrapElement && {"p":1,"li":1,"span":1,"div":1}[rbWrapElement] !== 1 ) {
                                    rbWrapElement = "span";
                                }
                            }
                            for ( i = 0; i < action.choices.length; i++ ) {
                                choice = action.choices[i];
                                if ( choice.hide ) {
                                    continue;
                                }
                                id = idPrefix + "_" + i;
                                if ( i === 0 && rbStartClasses ) {
                                    cls = rbStartClasses;
                                } else if ( i === action.choices.length - 1 && rbEndClasses ) {
                                    cls = rbEndClasses;
                                } else {
                                    cls = rbClasses || "";
                                }
                                if ( rbWrapElement ) {
                                    out.markup( "<" + rbWrapElement )
                                        .attr( "class", rbWrapClasses )
                                        .markup( ">" );
                                }
                                out.markup( "<input type='radio'" )
                                    .attr( "name", idPrefix )
                                    .attr( "id", id )
                                    .attr( "value", choice.value )
                                    .markup( "><label" )
                                    .attr( "for", id)
                                    .attr( "class", cls )
                                    .attr( "title", choice.title || choice.icon ? choice.title || choice.label : null )
                                    .markup( ">" );
                                if ( choice.icon ) {
                                    // assume if there is an icon then it should be an icon only radio button
                                    out.markup( "<span class='u-VisuallyHidden'>" )
                                        .content( choice.label )
                                        .markup( "</span><span aria-hidden='true'" )
                                        .attr( "class", (choice.iconType || "a-Icon") + " " +  choice.icon )
                                        .markup( "></span>" );
                                } else {
                                    out.content( choice.label );
                                }
                                out.markup( "</label>" );
                                if ( rbWrapElement ) {
                                    out.markup( "</" + rbWrapElement + ">" );
                                }
                            }
                        }
                        el$.html( out.toString() );

                    } else if ( op === "add" ) {
                        // initialize action from markup when first added
                        if ( action.label === null ) {
                            // initialize action text from button
                            // the label of a toggle button action can't come from the markup
                            if ( elementType === "button" && actionType === "action" ) {
                                if ( el$.attr( A_LABEL ) ) {
                                    // if there is a label for AT use that as the label
                                    action.label = el$.attr( A_LABEL );
                                } else if ( el$.is( buttonNoLabelSelector )) {
                                    // if there is no button label text use the title
                                    action.label = el$.attr( "title" );
                                } else {
                                    action.label = textLabel$.text();
                                }
                            } else if ( elementType === "checkbox" && actionType === "toggle" ) {
                                if ( isButtonLike && label$.is( buttonNoLabelSelector ) ) {
                                    // if there is no checkbox button label text look for hidden but accessible child
                                    action.label = label$.find( ".u-VisuallyHidden" ).first().text();
                                    if ( !action.label ) {
                                        // fall back to title
                                        action.label = label$.attr( "title" );
                                    }
                                } else {
                                    action.label = textLabel$.text();
                                }
                            } else {
                                if ( el$.attr( A_LABEL ) ) {
                                    // if there is a label for AT use that as the label
                                    action.label = el$.attr( A_LABEL );
                                }
                            }
                        }
                        if ( action.title === null && elementType !== "radio" ) {
                            action.title = el$.attr( "title" );
                            // if that didn't work see if there is a title on the label
                            if ( ! action.title && label$ ) {
                                action.title = label$.attr( "title" );
                            }
                            // The above code should handle most cases of finding the label but just in case fall back to the title
                            if ( !action.label && action.title ) {
                                action.label = action.title;
                            }
                        }
                        // there is no reliable way to get the icon from the markup

                        if ( action.disabled === null ) {
                            if ( el$.prop( "disabled" ) || ( input$ && input$.prop( "disabled" ) ) ) {
                                action.disabled = true;
                            }
                        }
                    }

                    // update markup with action properties
                    // note radio group labels/icons etc. are not updated - use updateChoices
                    label = action.label;
                    title = action.title;
                    if ( !label && actionType === "toggle" ) {
                        label = value ? action.onLabel : action.offLabel;
                    }

                    // update label, title, icon, disabled
                    if ( el$.attr( "data-no-update" ) === undefined ) {
                        if ( elementType === "button" && el$.is( buttonNoLabelSelector ) ) {
                            el$.attr( A_LABEL, label );
                            // on a button with no label and no title force the title to be the label
                            if ( !title ) {
                                title = label;
                            }
                        } else if ( elementType === "checkbox" && actionType === "toggle" && isButtonLike && label$.is( buttonNoLabelSelector ) ) {
                            label$.find( ".u-VisuallyHidden" ).first().text( label );
                            // on a button with no label and no title force the title to be the label
                            if ( !title ) {
                                title = label;
                            }
                        } else if ( actionType === "radio" ) {
                            if ( label ) {
                                el$.attr( A_LABEL, label );
                            } else {
                                el$.removeAttr( A_LABEL );
                            }
                            // for select lists that generally have no label use the label as the title if there isn't one
                            if ( elementType === "select" && !title ) {
                                title = label;
                            }
                        } else if ( textLabel$ ) {
                            textLabel$.text( label );
                        }

                        // radio groups don't have a title because it would get in the way of titles on each radio button
                        if ( elementType !== "radio" ) {
                            if ( title ) {
                                el$.attr( "title", title );
                            } else {
                                el$.removeAttr( "title" );
                            }
                        }

                        if ( elementType === "button" && action.icon && ( el$.is( buttonHasIconSelector ) ) ) {
                            iconType = action.iconType || "a-Icon";
                            updateIcons( el$, iconType, action.icon );
                        } else if ( isButtonLike && action.icon && ( label$.is( buttonHasIconSelector ) ) ) {
                            iconType = action.iconType || "a-Icon";
                            updateIcons( label$, iconType, action.icon );
                        }
                    }

                    // set the value/active state
                    if ( actionType !== "action" ) {
                        if ( elementType === "select" ) {
                            el$.val( value );
                        } else if ( elementType === "radio" ) {
                            el$.find( ".is-active" ).removeClass( "is-active" );
                            input$ = el$.find( "[value=" + util.escapeCSS( value ) + "]" );
                            input$.prop( "checked", true );
                            if ( input$.length ) {
                                label$ = getLabelFor( input$[0] );
                                label$.addClass( "is-active" );
                            }
                        } else if ( elementType === "checkbox" ) {
                            input$.prop( "checked", !!value );
                            label$.toggleClass( "is-active", value );
                        } else if ( el$.is( SEL_ACTION_BUTTON ) ) {
                            if ( actionType === "toggle" && !action.onLabel ) {
                                el$.toggleClass( "is-active", value );
                            }
                        }
                    }

                    if ( elementType === "checkbox" ) {
                        input$.prop( "disabled", !!action.disabled );
                    } if ( elementType === "radio" || elementType === "select") {
                        choices$ = el$.find("option,input"); // should work for radio buttons or select lists
                        for ( i = 0; i < action.choices.length; i++ ) {
                            choice = action.choices[i];
                            choices$.eq( i ).prop( "disabled", !!choice.disabled );
                        }
                        // select based radio group is special in that the whole select list can be disabled
                        if ( elementType === "select" ) {
                            el$.prop( "disabled", !!action.disabled );
                        }
                    } else {
                        el$.prop( "disabled", !!action.disabled );
                    }
                } );
            } else {
                actionElements$.hide();
            }
        });

        function doAction( actionName, event, logPrefix, value ) {
            var action, result, boolValue;

            try {
                action = actions.lookup( actionName );
                if ( !action ) {
                    debug.error( "No such action '" + actionName + "'." );
                    return;
                }

                if ( action.disabled ) {
                    return;
                }

                if ( action.action || action.href ) {
                    debug.info( logPrefix + " invoke action '" + actionName + "'");
                    result = actions.invoke( actionName, event, event.target );
                    if ( result === false ) {
                        return;
                    }
                } else if ( action.get && action.set ) {
                    if ( value !== undefined ) {
                        action.set( value );
                        debug.info( logPrefix + " set action '" + actionName + "'. Value now " + value);
                    } else {
                        boolValue = !action.get();
                        action.set( boolValue );
                        debug.info( logPrefix + " toggle action '" + actionName + "'. Value now " + boolValue);
                    }
                    actions.update( actionName );
                } else {
                    debug.error("Error action '" + actionName + "' has no href or action, get, or set methods." );
                }
            } catch ( ex ) {
                // ignore error so focus can be set
                debug.error("Error in action for '" + actionName + "'.", ex );
            }

            if ( result !== true ) {
                event.target.focus();
            }
        }

        /*
         * Handle actions buttons and keyboard shortcuts
         */
        context$.on( "click.actions", SEL_ACTION_BUTTON, function( event ) {
            var actionName = $( this ).attr( "data-action" );

            doAction( actionName, event, "Button click" );
            event.stopPropagation();
        } ).on( "click.actions", SEL_ACTION_CHECKBOX, function( event ) {
            var input, actionName = $( this ).attr( "data-action" );

            if ( event.target.nodeName !== "INPUT" ) {
                return;
            }
            input = event.target;
            doAction( actionName, event, "Checkbox click", input.checked );
            event.stopPropagation();
        } ).on( "change.actions", SEL_ACTION_RADIO_GROUP, function( event ) {
            var value, input$,
                actionName = $( this ).attr( "data-action" );

            input$ = $( this ).find( ":checked" );
            value = "";
            if ( input$.length ) {
                value = input$.val();
            }
            doAction( actionName, event, "Radio group change", value );
            event.stopPropagation();
        } ).on( "change.actions", SEL_ACTION_SELECT, function( event ) {
            var value,
                actionName = $( this ).attr( "data-action" ),
                select$ = $( this );

            value = select$.val();
            doAction( actionName, event, "Select change", value );
            event.stopPropagation();
        } ).on( "keydown.actions", function ( event ) {
            var actionName, value,
                keyName = makeKeyName( event );

            if ( keyName ) {
                if ( event.target.nodeName === "TEXTAREA" || event.target.nodeName === "INPUT" && event.target.type.toLowerCase() === "text" ) {
                    // ignore keys that are handled by the browser
                    if ( ["Ctrl+C", "Ctrl+X", "Ctrl+V", "Ctrl+A", "Ctrl+Z", "Ctrl+Y"].indexOf( keyName ) >= 0 ) {
                        return;
                    }
                }
                actionName = actions.lookupShortcutAction( keyName );
                if ( actionName ) {
                    if ( typeof actionName !== "string" ) {
                        value = actionName.value;
                        actionName = actionName.actionName;
                    }
                    // else value is undefined
                    doAction( actionName, event, "Shortcut key", value );
                    event.stopPropagation();
                }
            }
        } ).on( "menubeforeopen.actions", function() {
            actions.disableShortcuts();
        } ).on( "menuafterclose.actions", function() {
            actions.enableShortcuts();
        } ).on( "dialogopen.actions", function( event ) {
            if ( event.target !== actions.context && $( event.target ).dialog( "option", "modal" ) ) {
                actions.disableShortcuts();
            }
        } ).on( "dialogclose.actions", function( event ) {
            if ( event.target !== actions.context && $( event.target ).dialog( "option", "modal" ) ) {
                actions.enableShortcuts();
            }
        } );

        return actions;
    }

    /**
     * Remove an actions context that was created with createContext.
     *
     * @param pTypeName {string} type name of the actions context
     * @param pContext {DOMElement} DOM element context that actions are scoped within
     */
    function removeContext( pTypeName, pContext ) {
        var i, typeList, context$;

        typeList = gActionContextTypes[pTypeName];
        if ( typeList ) {
            for ( i = 0; i < typeList.length; i++ ) {
                if ( typeList[i].context === pContext ) {
                    // found something to remove
                    context$ = $( typeList[i].context );
                    context$.off(".actions");
                    typeList.splice( i, 1 );
                    return;
                }
            }
        }
    }

    apex.actions = createContext( "global", document.body );

    apex.actions.createContext = createContext;
    apex.actions.removeContext = removeContext;

    /**
     * Return an array of all the actions instances for a given type.
     *
     * @param pTypeName {string} type name of the actions context
     * @return {actions[]}
     */
    apex.actions.getContextsForType = function( pTypeName ) {
        return gActionContextTypes[pTypeName] || [];
    };

    /**
     * Return the actions object for a given context. The pTypeName is optional but if given must
     * match the type name used when the context was created. This is not often needed because the actions object
     * returned from createContext should be saved by any widget/component that created the context.
     *
     * @param [pTypeName] {string} type name of the actions context
     * @param pContext {DOMElement} DOM element context that actions are scoped within
     * @return {actions} returns the actions object or undefined if there is no actions defined for pContext
     */
    apex.actions.findContext = function( pTypeName, pContext ) {
        var tn, list, actions;

        function find( typeList ) {
            var i;
            for ( i = 0; i < typeList.length; i++ ) {
                if ( typeList[i].context === pContext ) {
                    return typeList[i];
                }
            }
            // intentionally "return" undefined here
        }

        // pTypeName is optional
        if ( !pContext ) {
            pContext = pTypeName;
            pTypeName = null;
        }
        if ( pTypeName ) {
            list = gActionContextTypes[pTypeName];
            if ( list ) {
                actions = find( list );
            }
        } else {
            for ( tn in gActionContextTypes) {
                actions = find( gActionContextTypes[tn] );
                if ( actions ) {
                    break;
                }
            }
        }
        return actions;
    };

    /**
     * This is used by UI that allows entering a shortcut name by typing that key combination
     *
     * @param pEvent {Event}
     * @return {?Object}
     */
    apex.actions.getShortcutFromEvent = function( pEvent ) {
        var keyName = makeKeyName( pEvent );
        if ( keyName ) {
            return {
                shortcut: keyName,
                shortcutDisplay: this.shortcutDisplay( keyName )
            };
        }
        return null;
    };

})( apex, apex.debug, apex.lang, apex.util, apex.jQuery );
