/*!
 item.js
 Copyright (c) 2012, 2016 Oracle and/or its affiliates. All rights reserved.
 */
/**
 * @fileOverview
 * The {@link apex.item} namespace is used for all page/column item related functions of Oracle Application Express.
 **/
/*global apex, $x_FormItems, $x_ItemRow, $x*/
(function( debug, $, undefined ) {
    "use strict";

    /**
     * The APEX page item object. This will hold all single item functions. These functions assume that these are
     * APEX generated ITEMS and will make best guess at the proper functionality to apply.
     *
     * @param {DOM-node | String} pNd
     * @param {Object}            [pCallbacks]
     *
     * @class
     **/
    function Item( pNd, pCallbacks ) {

        /* Check if the apex.item function was called within a constructor ( eg, using the
         * 'new' keyword), if not, we can construct it here. This allows object creation
         * without the 'new' keyword eg:
         *       var lItem = apex.item('P1_MYITEM');
         */
        if ( !( this instanceof Item ) ) {
            return new apex.item( pNd, pCallbacks );
        }

        var self                = this;
        this.node               = false;
        this.item_type          = false;
        this.id                 = false;
        this.getValue           = getValue;
        this.init               = init; // todo don't expose this
        this.reinit             = reinit;
        this.setValue           = setValue;
        this.isEmpty            = isEmpty;
        this.isChanged          = isChanged;
        this.enable             = enable;
        this.disable            = disable;
        this.isDisabled         = isDisabled;
        this.show               = show;
        this.hide               = hide;
        this.addValue           = addValue;
        this.removeValue        = removeValue;
        this.setFocus           = setFocus;
        this.setStyle           = setStyle;
        this.afterModify        = afterModify;
        this.loadingIndicator   = loadingIndicator;
        this.displayValueFor    = displayValueFor;
        this.getValidity        = getValidity;
        this.getValidationMessage = getValidationMessage;
        this.getPopupSelector   = getPopupSelector;

        if ( pCallbacks ) {
            this.callbacks = pCallbacks;
        } else {
            this.callbacks = {};
        }
        this.init( pNd, pCallbacks );

        /**
         * init function is all about ensuring the right _self.item_type is set for the node
         * _self.item_type is used in subsequent item methods to do things like set the value,
         * get the value, determine if it is empty and more.
         *
         * @instance
         * @memberOf apex.item
         * */
        function init( pNd, pCallbacks ) {
            var lNodeType, textType, node$;

            // determine type of pNd and assign _self.node accordingly
            try {
                switch ( typeof( pNd ) ) {
                    case 'string':
                        self.node = $( "#" + apex.util.escapeCSS( pNd ), apex.gPageContext$ )[0];
                        break;
                    case 'object':
                        self.node = pNd;
                        break;
                    default:
                        self.node = false;
                }
                if ( !( self.node && self.node.nodeType === 1 ) ) {
                    self.node = false;
                }
            } catch( e ) {
                self.node = false;
            }

            // only proceed if _self.node is not false
            if ( self.node ) {
                self.id = self.node.id;

                // If callbacks have been provided, register them in our global store so that they can later on be used
                // if a page item is looked up.
                if ( pCallbacks ) {
                    apex.page.itemCallbacks[ self.id ] = self.callbacks;
                } else if ( apex.page.itemCallbacks[ self.id ] ) {
                    self.callbacks = apex.page.itemCallbacks[ self.id ];
                }

                // continue with initialisation
                lNodeType = self.node.nodeName.toUpperCase();

                // if the node is a fieldset, assign item_type based on the class name
                if ( lNodeType === 'FIELDSET' ) {
                    node$ = $( self.node );
                    self.item_type = false; // default
                    if ( node$.hasClass("checkbox_group") ) {
                        self.item_type = "CHECKBOX_GROUP";
                    } else if ( node$.hasClass("radio_group") ) {
                        self.item_type = "RADIO_GROUP";
                    } else if ( node$.hasClass("shuttle") ) {
                        self.item_type = "SHUTTLE";
                    }

                // if node type is an input, assign item_type as the node type (CHECKBOX, RADIO, TEXT etc.)
                } else if ( lNodeType === 'INPUT' ) {
                    self.item_type = self.node.type.toUpperCase();

                    // switch on item_type to ensure item_type and display_span attributes are initialised
                    switch ( self.item_type ) {
                        case 'CHECKBOX':
                        case 'RADIO':
                            break;
                        case 'TEXT':
                            textType = self.node.parentNode.className.toUpperCase();
                            if ( textType === "LOV" ) {
                                if ( $( '#' + self.id + '_HIDDENVALUE', apex.gPageContext$ ).length > 0) {
                                    self.item_type = 'POPUP_KEY_LOV';
                                } else {
                                    self.item_type = 'POPUP_LOV';
                                }
                            }
                            break;
                        case 'HIDDEN':
                            self.display_span = $x( self.id + '_DISPLAY' );
                            if ( self.display_span ) {
                                self.item_type = 'DISPLAY_SAVES_STATE';
                            }
                            break;
                        default:
                            self.item_type = 'TEXT';
                    }

                    // if the node type is not a fieldset or an input, initialise item_type accordingly
                } else {
                    self.item_type = lNodeType;
                    switch ( self.item_type ) {
                        case 'TEXTAREA':
                            if ( self.node.parentNode.className === 'html_editor' && self.node.parentNode.tagName === 'FIELDSET' ) {
                                self.item_type = 'FCKEDITOR';
                            } else {
                                try {
                                    if (window.CKEDITOR && window.CKEDITOR.instances[ self.id ] ) {
                                        self.item_type = 'CKEDITOR3';
                                    }
                                } catch( e ) {}
                            }
                            break;
                        case 'SELECT':
                            break;
                        case 'SPAN':
                            if ( $( self.node ).hasClass( "display_only" ) ) {
                                self.item_type = 'DISPLAY_ONLY';
                            }
                            break;
                        default:
                            self.item_type = false;
                    }
                } // end if on lNodeType
            } //end if (_self.node)
        } //end init

        /**
         * Called when the item is being reused in a new context. This is called for column items when a new record
         * is being edited. This is not normally used for page items. Default behaviour does setValue.
         * Items that support cascading LOVs should implement this function to first set the item's value (which may also
         * require adding the value as an option in the item), then return a function where the cascade will take place.
         *
         * @return {function}   When an item defines reinit, it will return a function for post processing
         * @param pValue        The value to set
         * @param pDisplayValue The display value only if different from pValue
         *
         * @instance
         * @memberOf apex.item
         */
        function reinit( pValue, pDisplayValue ) {
            if ( $.isFunction( self.callbacks.reinit )) {
                return self.callbacks.reinit.call( self, pValue, pDisplayValue );
            } else {
                self.setValue( pValue, pDisplayValue, true );
            }
        }

        /**
         * Returns the current value of the item. For multi valued items it returns an array of strings.
         * @return {string|Array} the current value of the item.
         * @instance
         * @memberOf apex.item
         * */
        function getValue() {
            var oEditor, lRadio$, lArray, lReturn;

            if ( !self.node ) {
                return "";
            } // else
            if ( $.isFunction( self.callbacks.getValue ) ) {
                return self.callbacks.getValue.call( self );
            } // else
            lArray = true;
            lReturn= [];
            switch ( self.item_type ) {
                case 'RADIO_GROUP':

                    // radio group should return a single value
                    lRadio$ = $( ':checked', self.node );
                    if ( lRadio$.length === 0 ) {

                        // check if the length of the jQuery object is zero (nothing checked)
                        // if so return an empty string.
                        lReturn = "";
                    } else {

                        // otherwise return the value
                        lReturn = lRadio$.val();
                    }
                    break;
                case 'CHECKBOX_GROUP':
                    $( ':checked', self.node ).each( function() {
                        lReturn[ lReturn.length ] = this.value;
                    });
                    break;
                case 'SELECT':
                    lReturn = $( self.node ).val();
                    if ( lReturn === null || lReturn === undefined ) {
                        if ( $( self.node ).attr( "multiple" ) ) {
                            lReturn = [];
                        } else {
                            lReturn = "";
                        }
                    }
                    break;
                default:
                    lArray = false;
            }
            if ( !lArray ) {
                switch ( self.item_type ) {
                    /* check single checkbox entry */
                    case 'CHECKBOX'             :lReturn = ( self.node.checked ) ? self.node.value : ""; break;
                    /* check single radio entry */
                    case 'RADIO'                :lReturn = ( self.node.checked ) ? self.node.value : ""; break;
                    case 'POPUP_KEY_LOV'        :lReturn = $( '#' + self.node.id + "_HIDDENVALUE", apex.gPageContext$ ).val(); break;
                    case 'TEXT':
                    case 'POPUP_LOV':
                    case 'HIDDEN':
                    case 'DISPLAY_SAVES_STATE':
                    case 'TEXTAREA':
                        lReturn = self.node.value;
                        break;
                    case 'DISPLAY_ONLY'         :lReturn = self.node.innerHTML; break;
                    case 'FCKEDITOR'            :
                        oEditor = FCKeditorAPI.GetInstance( self.node.id ) ;
                        lReturn = oEditor.GetHTML();
                        break;
                    default                     :lReturn = "";
                }
            }
            return lReturn;
        } //end getValue

        /**
         * Sets the item value. It also allows for the caller to suppress the 'change' event for the item being set,
         * if desired. Normally a change event is explicitly triggered on the item node.
         *
         * @param pValue the value to set
         * @param pDisplayValue the display value only if different from pValue and can't be determined by the item itself
         * @param pSuppressChangeEvent if true the change event will be suppressed.
         *
         * @instance
         * @memberOf apex.item
         **/
        function setValue( pValue, pDisplayValue, pSuppressChangeEvent ) {
            var i, len, lCheck, lOpts, lSelf$, oEditor;

            if ( $.isFunction( self.callbacks.setValue )) {
                self.callbacks.setValue.call( self, pValue, pDisplayValue, pSuppressChangeEvent );
            } else {
                lOpts = false;
                lSelf$ = $( self.node, apex.gPageContext$ );
                if ( !self.node ) {
                    return;
                }
                switch ( self.item_type ) {
                    case 'RADIO_GROUP'      :
                        lOpts = $x_FormItems( self.node, 'RADIO' );
                        break;
                    case 'CHECKBOX_GROUP'   :
                        lOpts = $x_FormItems( self.node, 'CHECKBOX' );
                        break;
                    case 'POPUP_KEY_LOV'    :
                        // popup key lovs store there value in a hidden field
                        $( '#' + self.node.id + '_HIDDENVALUE', apex.gPageContext$ ).val( pValue );
                        lSelf$.val( pDisplayValue );
                        break;
                    case 'SELECT'           :
                        lOpts = self.node.options;
                        break;
                    default                 :
                        lOpts = false;
                }
                if ( lOpts ) {
                    for ( i = 0, len = lOpts.length; i < len; i++ ) {
                        lCheck = lOpts[i].value == pValue;
                        if ( self.item_type === 'RADIO_GROUP' || self.item_type === 'CHECKBOX_GROUP' ) {
                            lOpts[ i ].checked = lCheck;
                        } else {
                            lOpts[ i ].selected = lCheck;
                        }
                    }
                } else {
                    switch ( self.item_type ) {
                        case 'CHECKBOX':
                        case 'RADIO':
                            self.node.checked = self.node.value === pValue;
                            break;
                        case 'TEXT':
                        case 'POPUP_LOV':
                        case 'HIDDEN':
                        case 'TEXTAREA':
                            lSelf$.val( pValue );
                            break;
                        case 'DISPLAY_SAVES_STATE':
                            lSelf$.val( pValue );
                            $( self.display_span, apex.gPageContext$ ).html( pValue );
                            break;
                        case 'DISPLAY_ONLY':
                            lSelf$.html( pValue );
                            break;
                        case 'FCKEDITOR':
                            oEditor = FCKeditorAPI.GetInstance( self.node.id );
                            oEditor.SetHTML( pValue );
                            break;
                    /**
                     * must be some other tag item set it's innerHTML
                     * */
                        default:
                            self.node.innerHTML = pValue;
                    }
                }
            }
            self.afterModify();

            /* Only if pSuppressChangeEvent is set to true, do we not trigger the change event.
             * In the case where this is not passed, the change event is triggered (for backwards
             * compatibility). Or if this is explicitly set to false, then the event will also trigger.
             */
            if ( !pSuppressChangeEvent ) {
                $( self.node ).trigger( 'change' );
            }
        } //end setValue

        /**
         * Enables the form element. Checks if there is an enable callback registered for the
         * current element, if there is _self is called (setting 'this' context to current node,
         * otherwise do base enabling.
         *
         * @instance
         * @memberOf apex.item
         * */
        function enable() {
            if ( $.isFunction( self.callbacks.enable ) ) {
                self.callbacks.enable.call( self );
            } else {
                $( self.node )
                    .removeClass( "apex_disabled" )
                    .prop( "disabled", false );

                if ( $.mobile &&
                    ( self.item_type === "TEXTAREA" ||
                        $.inArray( self.node.type, [ "text", "email", "url", "tel", "search", "number", "password", "time", "date", "month", "week", "datetime", "datetime-local", "color" ]) >= 0 )
                    )
                {
                    $( self.node ).textinput( "enable" );
                }
            }
            self.afterModify();
        } // end enable

        /**
         * Disables the form element. Checks if there is an disable callback registered for the
         * current element, if there is _self is called (setting 'this' context to current node,
         * otherwise do base disabling.
         *
         * @instance
         * @memberOf apex.item
         * */
        function disable() {
            if ( $.isFunction( self.callbacks.disable )) {
                self.callbacks.disable.call( self );
            } else {
                $( self.node )
                    .addClass( "apex_disabled" )
                    .prop( "disabled", true );

                if ( $.mobile &&
                    ( self.item_type === "TEXTAREA" ||
                        $.inArray( self.node.type, [ "text", "email", "url", "tel", "search", "number", "password", "time", "date", "month", "week", "datetime", "datetime-local", "color" ]) >= 0 )
                    )
                {
                    $( self.node ).textinput( "disable" );
                }
            }
            self.afterModify();
        } // end disable

        /**
         * Return true if the item is disabled and false otherwise
         * @return {Boolean} disabled state of item
         */
        function isDisabled() {
            if ( $.isFunction( self.callbacks.isDisabled )) {
                return self.callbacks.isDisabled.call( self );
            } else {
                return !!$( self.node ).prop( "disabled" );
            }
        } // end isDisabled

        /**
         * Shows the item form element. When using grid layout the #CURRENT_ITEM_CONTAINER_ID#
         * is used in the template to identify the element to show. A grid based layout container node
         * is assumed to include both the label and the control. When using table layout the table
         * row is shown if the pShowRow is true otherwise the form element and corresponding label
         * are shown.
         *
         * A widget can provide a callback to override the logic for showing the form element.
         *
         * @instance
         * @memberOf apex.item
         * @param {boolean} pShowRow Optional. If true, shows the nearest containing table row (TR).
         * Only applicable when item is in a table layout.
         **/
        function show( pShowRow ) {
            // Note: the logic involving CONTAINER and DISPLAY suffix must be reflected in
            // $x_Toggle so that it tests the correct node for visibility.
            var lNodeId = apex.util.escapeCSS( self.node.id ),
                lNodeDisplay$ = $( '#' + lNodeId + '_CONTAINER', apex.gPageContext$ );

            if ( lNodeDisplay$.length > 0 ) {
                lNodeDisplay$.show();
            } else {
                if ( pShowRow ) {
                    $x_ItemRow( self.node, 'SHOW' );
                } else {
                    if ( $.isFunction( self.callbacks.show )) {
                        self.callbacks.show.call( self );
                    } else {
                        lNodeDisplay$ = $( '#' + lNodeId + '_DISPLAY', apex.gPageContext$ );
                        if ( lNodeDisplay$.length > 0 ) {
                            lNodeDisplay$.show();
                        } else {
                            $( self.node ).show().trigger("apexaftershow");
                        }
                    }

                    // try and show the label as well, regardless of whether callback is defined
                    if ( lNodeId ) {
                        $( 'label[for=' + lNodeId + ']', apex.gPageContext$ ).show();
                    }
                }
            }
        } // end show

        /**
         * Hides the item form element. When using grid layout the #CURRENT_ITEM_CONTAINER_ID#
         * is used in the template to identify the element to hide. A grid based layout container node
         * is assumed to include both the label and the control. When using table layout the table
         * row is hidden if the pHideRow is true otherwise the form element and corresponding label
         * are hidden.
         *
         * A widget can provide a callback to override the logic for hiding the form element.
         *
         * @instance
         * @memberOf apex.item
         * @param {boolean} pHideRow Optional. If true, hides the nearest containing table row (TR).
         * Only applicable when item is in a table layout.
         * */
        function hide( pHideRow ) {
            // Note: the logic involving CONTAINER and DISPLAY suffix must be reflected in
            // $x_Toggle so that it tests the correct node for visibility.
            var lNodeId = apex.util.escapeCSS( self.node.id ),
                lNodeDisplay$ = $( '#' + lNodeId + '_CONTAINER', apex.gPageContext$ );

            if ( lNodeDisplay$.length > 0 ) {
                lNodeDisplay$.hide();
            } else {
                if ( pHideRow ) {
                    $x_ItemRow( self.node, 'HIDE' );
                } else {
                    if ( $.isFunction( self.callbacks.hide )) {
                        self.callbacks.hide.call( self );
                    } else {
                        lNodeDisplay$ = $( '#' + lNodeId + '_DISPLAY', apex.gPageContext$ );
                        if ( lNodeDisplay$.length > 0 ) {
                            lNodeDisplay$.hide();
                        } else {
                            $( self.node ).hide().trigger( "apexafterhide" );
                        }
                    }

                    // try and hide the label as well, regardless of whether callback is defined
                    if ( lNodeId ) {
                        $( 'label[for=' + lNodeId + ']', apex.gPageContext$ ).hide();
                    }
                }
            }
        } // end hide


        /**
         * Returns true or false if a form element is empty, this will consider any whitespace including a space, a tab, a form-feed,
         * as empty.
         *
         * @return {boolean}
         *
         * @instance
         * @memberOf apex.item
         * */
        function isEmpty() {
            var lItemValue, re, lThis, lReturn,
                lNullValue = "";

            lItemValue = self.getValue(); //does the heavy lifting!

            // Make life easier and always use a string for all compare operations! $v doesn't work in this context
            if ( $.isArray( lItemValue ) ) {
                lItemValue = lItemValue.join( ':' );
            } else {
                lItemValue = "" + lItemValue;
            }

            re = /^\s{1,}$/g; //match any white space including space, tab, form-feed, etc.
            lThis = $x( self.node );

            /* Different item types will be tested for 'is empty' in different ways:
             *
             *  Case 1: text input, textareas will return true if they are null or they match any white space
             *  Case 2: multi select lists return true if they have no options selected or the current value equals the null value (not sure whether this should include null value)
             *  Case 3: all select list will ONLY return true if their current value equals the matching value in the apex.nullmap array
             *  Case 4: display only no state will return true if the span's innerHTML is empty
             *  Case 5: display only save state will return true if the relevant input's value is empty
             *  Case 6: popup lov returns true if null
             *  Case 7: popup key lov returns true if null
             *  Case 8: shuttles will return true by having no options in the right hand select element
             *  Case 9: checkboxes will return true if no checkboxes in the page item's group are checked
             * Case 10: radio groups will return true if no radio buttons in the page item's group are selected
             * Case 11: list managers will return true by having no options in the element
             * Case 12: popup color pickers will return true if no color is specified
             * Case 13: popup date pickers will return true if no date is specified
             * Case 14: FCKEditor will return null if the iFrame content is empty
             * Case 15: CKEditor will return null if the iFrame content is empty
             *
             */

            if ( 'nullValue' in self.callbacks ) {
                if ( $.isFunction( self.callbacks.nullValue )) {
                    return self.callbacks.nullValue.call( self );
                } else {
                    // basic comparison
                    return ( lItemValue.length === 0 ) || ( lItemValue === null ) || ( lItemValue === self.callbacks.nullValue ) || ( ( lItemValue.search( re ) ) > -1 );
                }
            } else {
                if ( self.item_type === 'SELECT' ) {
                    if ( apex.widget && apex.widget.report && apex.widget.report.tabular && apex.widget.report.tabular.gNullValueList ) {
                        $.each( apex.widget.report.tabular.gNullValueList, function( pId, pValue ) {
                            if ( this.name === lThis.name ) {
                                lNullValue = pValue.value;
                                return false;
                            }
                        });
                    }
                    if ( lThis.multiple ) {
                        lReturn = ( lItemValue.length===0 ) || ( lItemValue === lNullValue );   //case 2
                    } else {
                        lReturn = ( lNullValue || lNullValue === "" ) ? ( lItemValue === lNullValue ):false;           //case 3
                    }
                } else {
                    lReturn = ( lItemValue.length === 0 ) || ( lItemValue === null ) || ( ( lItemValue.search( re ) ) > -1);    //case 1,4,5,6,7,9,10,11,12,13,14,15 (exp 2 or 3)
                    //case 8 (exp 1)
                }
                return lReturn;
            }
        } // end isEmpty

        /**
         * Determine if the value of this item has changed since it was first initialized.
         *
         * @return {boolean} true if the item value has changed
         * @instance
         * @memberOf apex.item
         */
        function isChanged() {
            var i, opt, curValue, origValue, elements,
                changed = false;

            function checkMulti() {
                var changed = curValue.length !== origValue.length;
                if ( !changed ) {
                    for ( i = 0; i < origValue.length; i++ ) {
                        if ( curValue[i] !== origValue[i] ) {
                            changed = true;
                            break;
                        }
                    }
                }
                return changed;
            }

            if ( $.isFunction( self.callbacks.isChanged )) {
                return self.callbacks.isChanged.call( self );
            } else {
                switch ( self.item_type ) {
                    case 'TEXTAREA':
                    case 'TEXT':
                    case 'POPUP_LOV':
                        changed = self.node.value !== self.node.defaultValue;
                        break;
                    case 'SELECT':
                        curValue = $( self.node ).val();
                        if ( self.node.type === "select-multiple" ) {
                            if ( !curValue ) {
                                curValue = [];
                            }
                            origValue = [];
                            for ( i = 0; i < self.node.options.length; i++ ) {
                                opt = self.node.options[i];
                                if ( opt.attributes.selected !== undefined ) {
                                    origValue.push( opt.value );
                                }
                            }
                            changed = checkMulti();
                        } else {
                            origValue = "";
                            for ( i = 0; i < self.node.options.length; i++ ) {
                                opt = self.node.options[i];
                                if ( opt.attributes.selected !== undefined ) {
                                    origValue = opt.value;
                                    break;
                                }
                            }
                            changed = curValue !== origValue;
                        }
                        break;
                    case 'RADIO_GROUP':
                        curValue = self.getValue();
                        origValue = "";
                        elements = $x_FormItems( self.node, 'RADIO' );
                        for ( i = 0; i < elements.length; i++ ) {
                            if ( elements[i].defaultChecked ) {
                                origValue = elements[i].value;
                                break;
                            }
                        }
                        changed = curValue !== origValue;
                        break;
                    case 'CHECKBOX_GROUP':
                        curValue = self.getValue();
                        origValue = [];
                        elements = $x_FormItems( self.node, 'CHECKBOX' );
                        for ( i = 0; i < elements.length; i++ ) {
                            if ( elements[i].defaultChecked ) {
                                origValue.push( elements[i].value );
                            }
                        }
                        changed = checkMulti();
                        break;
                    case 'RADIO':
                    case 'CHECKBOX':
                        changed = self.node.checked !== self.node.defaultChecked;
                        break;
                    case 'POPUP_KEY_LOV':
                        // NOTE this is comparing the display value to detect a change. Comparing the hidden value
                        // would never find a change because a hidden value and defaultValue are always identical.
                        // So this will fail in the case where the current and original display values are identical
                        // but the return values are different. This seems too unlikely to worry about.
                        changed = self.node.value !== self.node.defaultValue;
                        break;
                    // because user can't change these directly they are never changed so let them default
                    //case 'DISPLAY_SAVES_STATE':
                    //case 'DISPLAY_ONLY':
                    //case 'HIDDEN':
                    // other types will need to implement their own changed logic
                }
            }
            return changed;
        }

        // todo may need a method to mark as not changed clearChange

        /**
         * @TODO Add documentation
         *
         * @instance
         * @memberOf apex.item
         **/
        function addValue( pValue ) {
            if ( $.isFunction( self.callbacks.addValue )) {
                self.callbacks.addValue.call( self, pValue );
            } else {
                debug.error( "No default handling defined for addValue" );
            }
            self.afterModify();
        } // end addValue

        /**
         * @TODO Add documentation
         *
         * @instance
         * @memberOf apex.item
         **/
        function removeValue() {
            if ( $.isFunction( self.callbacks.removeValue )) {
                self.callbacks.removeValue.call( self );
            } else {
                debug.error( "No default handling defined for removeValue" );
            }
            self.afterModify();
        } // end removeValue

        /**
         * Sets focus to this item.
         *
         * @instance
         * @memberOf apex.item
         **/
        function setFocus() {
            var lSetFocusTo$;

            if ( 'setFocusTo' in self.callbacks ) {
                if ( $.isFunction( self.callbacks.setFocusTo ) ) {

                    // setFocusTo can be a function
                    lSetFocusTo$ = self.callbacks.setFocusTo.call ( self );
                } else {

                    // If not a function, setFocusTo can be either a DOM object, jQuery selector or jQuery object
                    lSetFocusTo$ = $( self.callbacks.setFocusTo );
                }
            } else {

                // Default handling is to use the element with the ID of the page item
                lSetFocusTo$ = $( "#" + apex.util.escapeCSS( self.id ), apex.gPageContext$ );
            }
            lSetFocusTo$.focus();
        } // end setFocus


        /**
         * @TODO Add documentation
         *
         * @instance
         * @memberOf apex.item
         * */
        function setStyle( pPropertyName, pPropertyValue ) {
            var lSetStyleTo$;

            if ( 'setStyleTo' in self.callbacks ) {
                if ( $.isFunction( self.callbacks.setStyleTo ) ) {

                    // setStyleTo can be a function
                    lSetStyleTo$ = self.callbacks.setStyleTo.call ( self );
                } else {

                    // If not a function, setStyleTo can be either a DOM object, jQuery selector or jQuery object
                    lSetStyleTo$ = $( self.callbacks.setStyleTo );
                }
            } else {

                // Default handling is to use the element with the ID of the page item
                lSetStyleTo$ = $( "#" + apex.util.escapeCSS( self.id ), apex.gPageContext$ );
            }
            lSetStyleTo$.css( pPropertyName, pPropertyValue );
            self.afterModify();
        } // end setStyle


        /**
         * @TODO Add documentation
         *
         * @instance
         * @memberOf apex.item
         * */
        function afterModify() {
            /* Some frameworks need to get notified if widgets are modified */
            if ( $.isFunction( self.callbacks.afterModify )) {
                self.callbacks.afterModify.call( self );
            }
        } // end afterModify

        /**
         * @TODO Add documentation
         *
         * @instance
         * @memberOf apex.item
         * */
        function loadingIndicator( pLoadingIndicator$ ) {
            var lLoadingIndicator$;

            if ( 'loadingIndicator' in self.callbacks ) {
                if ( $.isFunction( self.callbacks.loadingIndicator )) {

                    // loadingIndicator currently just supports a function currently.
                    // The function receives the loading indicator span as 1st argument
                    // and must return the created jQuery object.
                    lLoadingIndicator$ = self.callbacks.loadingIndicator.call( self, pLoadingIndicator$ );
                }
            }
            return lLoadingIndicator$;
        } // loadingIndicator

        /**
         * Given a value return the corresponding display value
         * @param pValue
         */
        function displayValueFor( pValue ) {
            var display = pValue;

            if ( display === undefined || display === null ) {
                display = "";
            }
            if ( $.isFunction( self.callbacks.displayValueFor ) ) {
                display = self.callbacks.displayValueFor.call( self, pValue );
            } else {
                if ( self.node.type === "password" ) {
                    return "******";
                }
                switch ( self.item_type ) {
                case 'POPUP_KEY_LOV':
                    if ( pValue === self.getValue() ) {
                        display = $( self.node ).val();
                    }
                    break;
                case 'SELECT':
                    if ( pValue !== undefined && pValue !== null ) {
                        display = $( self.node ).find( "[value='" + apex.util.escapeCSS( pValue + "" ) + "']" ).html();
                        if ( display === undefined || display === null ) {
                            display = pValue;
                        }
                    }
                    break;
                }
            }
            return display;
        }

        /**
         * Return a validity object as defined by the HTML5 constraint validation API.
         * If a custom item implements its own validation then the object may not contain all the fields defined
         * by HTML5. At a minimum it should have the "valid" property.
         * If the item doesn't support HTML5 validation then { valid: true } is returned.
         * @return {object} validity object
         */
        function getValidity() {
            var validity = null;

            if ( $.isFunction( self.callbacks.getValidity ) ) {
                validity = self.callbacks.getValidity.call( self );
            } else {
                validity = self.node.validity || { valid:true };
            }
            return validity;
        }

        /**
         * Return a validation error message if the item is not valid and empty string otherwise.
         * By default, if the item has the "data-valid-message" attribute then it is returned if the item is not valid.
         * @return {string} validation error message or empty string if item is valid
         */
        function getValidationMessage() {
            var validMessage,
                message = "";

            if ( $.isFunction( self.callbacks.getValidationMessage ) ) {
                message = self.callbacks.getValidationMessage.call( self );
            } else {
                validMessage = $( self.node ).attr( "data-valid-message" );
                if ( !self.getValidity().valid && validMessage  ) {
                    message = validMessage;
                } else {
                    message = self.node.validationMessage || "";
                }
            }
            return message;
        }

        /**
         * Any item type that uses a popup (a div added near the end of the document that is positioned near the input
         * and floating above) needs to provide a selector that locates the top level element of the popup.
         * This allows the item type to be used in the grid widget.
         * In addition the popup element must be focusable (tabindex = -1).
         * For best behavior of a popup in the grid. The popup should
         * - have a way of taking focus
         * - return focus to the grid when it closes
         * - close on escape when it has focus and
         * - close when the element it is attached to looses focus
         * - manage its tab stops so they cycle in the popup or return to the grid at the ends
         *
         * @return {String} selector that identifies a popup associated with this item or null if there is no popup
         */
        function getPopupSelector() {
            if ( $.isFunction( self.callbacks.getPopupSelector ) ) {
                return self.callbacks.getPopupSelector.call( self );
            } // else
            return null;
        }
    }

    apex.item = Item;

    /**
     * Create an apex item for an instance given by pNd with implementation overrides in pItemImpl.
     *
     * @param pNd page/column item name or DOM node.
     * @param pItemImpl object with any of these properties:
     * - reinit: function()
     * - getValue: function()
     * - setValue: function( value, display, suppressChangeEvent )
     * - enable: function()
     * - disable: function()
     * - isDisabled: function()
     * - show: function()
     * - hide: function()
     * - addValue: function( value )
     * - removeValue: function ( ? ) todo consider correct behavior
     * - nullValue: {string}
     * - setFocusTo: function(), DOM object, jQuery selector or jQuery object
     * - setStyleTo: function(), DOM object, jQuery selector or jQuery object
     * - afterModify: function()
     * - loadingIndicator: function( loadingIndicator$ )
     * - isChanged: function()
     * - displayValueFor: function()
     * - getValidity: function()
     * - getValidationMessage: function()
     */
    apex.item.create = function ( pNd, pItemImpl ) {
        apex.item( pNd, pItemImpl );
    };

    // todo consider if there is a need for apex.item.destroy

})( apex.debug, apex.jQuery );
