/*!
 region.js
 Copyright (c) 2016, Oracle and/or its affiliates. All rights reserved.
 */
/**
 * @fileOverview
 * The {@link apex.region} namespace is used for all general region related functions of Oracle Application Express.
 */
/*global apex*/

(function( debug, $ ) {
    "use strict";

    var REGION_KEY = "apexregion",
        C_REGION = "js-apex-region";

    /**
     * @type {Object} region
     */
    var regionPrototype = {
        type: "generic",
        /**
         * Cause the region to get new data from the server. Not all regions support refresh.
         */
        refresh: function() {
            this.element.trigger( "apexrefresh" );
        },
        /**
         * Cause the region to take focus. Not all regions will take focus. It is up to the region
         * to determine where focus will go. Some regions manage focus such that there is a single
         * tab stop (or limited number).
         */
        focus: function() {
            this.element.find( ":tabbable" ).first().focus();
        },
        /**
         * If the region implementation uses a widget such as a jQuery UI widget this method will
         * return that widget. Otherwise it returns null; Typically this will be a jQuery object
         * that you then call widget methods on.
         * @return {*}
         */
        widget: function() {
            return null;
        },

        /**
         * For regions that use an apex model or column items this provides a way for the apex.message facility
         * to link error messages to the source of the data entry error.
         * @param {Object} pErrorContext an object with properties that identify the source of the error.
         */
        gotoError: function( pErrorContext ) {
        },

        /**
         * Return an alternative loading indicator for the given element. For regions that support column items
         * and when the column items may not be visible on the screen at all times this allows the region to
         * place the loading indicator in an appropriate location. This can return the loading indicator passed in
         * or return a completely new loading indicator.
         *
         * @param {DOM Element} pElement DOM element that may represent a column item.
         * @param {jQuery} pLoadingIndicator$ A loading indicator that can be inserted in to the DOM where desired and returned or ignored.
         * @return {Object} loadingIndicator jQuery object or null if the region has no alternative for given element
         */
        /* there is no default implementation
         alternateLoadingIndicator: function( pElement, pLoadingIndicator$ ) {
         return {};
         }
         */

        /**
         * Returns an object containing these properties:
         *
         * region: The region property contains any region specific data that needs to be sent to the server.
         * The region object must include these properties:
         *  - id: this is the region id (not the region static id)
         *  - ajaxColumns: TODO think will all regions have this?
         *  - ajaxIdentifier: ajax identifier for the region plugin as returned by PL/SQL API apex_plugin.get_ajax_identifier
         *  - setSessionState: and object containing properties
         *      - values: this is an array of {n: NAME, v: VALUE, ck: CHECKSUM} objects for each column item in pItemsToSubmit
         *      - protected: this is the protected property of the model record metadata for the active record
         *      - salt: this is the salt property of the model record metadata for the active record
         *  TODO think what about master data model parentItems?
         *
         * pageItems: The pageItems property is an array of page items. It contains all the items from pItemsToSubmit except
         * column items that belong to this region are removed.
         *
         * beforeAsync: Optional no argument function that must be called before any async operation begins
         * that could update the value of any column items related to this region.
         *
         * afterAsync: Optional no argument function that must be called after any async operation ends
         * that could update the value of any column items related to this region. There can be multiple calls to
         * beforeAsync and for each one afterAsync must be called.
         *
         * @param {array} pItemsToSubmit An array of item names. It includes both page items and column items.
         * @return {Object} region context object or null if the region doesn't support getting session state
         */
        getSessionState: function( pItemsToSubmit ) {
            return null;
        }

        // TODO future consider hide, show methods
    };

    function getElement( pRegionId ) {
        return $( "#" + apex.util.escapeCSS( pRegionId ), apex.gPageContext$ );
    }

    /**
     * Return a region object for the given region id.
     *
     * @param pRegionId the region's DOM element id. This is the region id or region static id.
     * @return {Region | null} region object or null if there is no element with the given pRegionId
     */
    apex.region = function( pRegionId ) {
        var region  = null,
            element$ = getElement( pRegionId );

        if ( element$.length ) {
            region = getElement( pRegionId ).data( REGION_KEY );
            // it is expected (and more efficient) that the region has been initialized but if not return a generic region type
            if ( !region ) {
                region = $.extend( {}, regionPrototype );
                region.element = element$;
            }
        }
        return region;
    };

    /**
     * Create a region element with a set of behaviors given by pRegionImpl.
     * The pRegionImpl can provide its own implementation for standard methods (refresh, focus, widget) or omit
     * them to get the default implementation. It can add its own methods as well. It should include
     * a type string property that specifies the type of region.
     *
     * @param pRegionId the region's DOM element id. This is the region id or region static id.
     * @param pRegionImpl a region type object that provides the implementation of standard and custom behaviors for the region
     */
    apex.region.create = function ( pRegionId, pRegionImpl ) {
        var region,
            element$ = getElement( pRegionId );

        if ( element$.length ) {
            region = $.extend( {}, regionPrototype, pRegionImpl );
            region.element = element$;
            element$.addClass( C_REGION );
            element$.data( REGION_KEY, region );
        } else {
            throw new Error( "Region element not found " + pRegionId );
        }
    };

    /**
     * Destroy and remove the behaviors associated with a region element.
     * @param pRegionId the region's DOM element id. This is the region id or region static id.
     */
    apex.region.destroy = function ( pRegionId ) {
        var element$ = getElement( pRegionId );

        if ( element$.length ) {
            element$.removeData( REGION_KEY );
        } else {
            throw new Error( "Region element not found " + pRegionId );
        }
    };

    /**
     * Returns true if and only if there is an element with id equal pRegionId that has had a region created for it
     * with apex.region.create.
     *
     * @param pRegionId an element id to check if the element is an apex region
     * @return {Boolean} true if there is an element with the given id that supports the region interface
     */
    apex.region.isRegion = function( pRegionId ) {
        var element$ = getElement( pRegionId );

        if ( element$.length ) {
            return !!element$.data( REGION_KEY );
        }
        return false;
    };

    /**
     * Returns the region that contains the pTarget element. Returns null if there is no pTarget element or if it is
     * not in a region that has been initialized with a call to apex.region.create.
     *
     * @param pTarget A DOM element or CSS selector suitable as the first argument to the jQuery function.
     * @return {Region | null}
     */
    apex.region.findClosest = function( pTarget ) {
        var id = $( pTarget ).parent().closest( "." + C_REGION ).prop( "id" );

        if ( id ) {
            return apex.region( id );
        }
        return null;
    };

})( apex.debug, apex.jQuery );
