/*!
 Grid - a jQuery UI based widget for displaying and editing a table grid.
 Copyright (c) 2015, 2017, Oracle and/or its affiliates. All rights reserved.
 */
/**
 * @fileOverview
 * Grid is a navigable data table that supports selection and editing.
 * It follows the DHTML Style Guide and WAI-ARIA design pattern for a grid box with these differences:
 * - Only row selection is supported so the Shift and Ctrl modifiers work like a list control and Shift-8 is not supported
 * - In edit/actionable mode you can tab out of the grid at the beginning or end
 *
 * The markup expected by this widget is simply an empty div. The grid displays and optionally edits data
 * stored in an apex data model (see {@link apex.model} namespace). If the grid is editable then the div must
 * be proceeded by a div with class "u-vh" (to visually hide the contents)that contains each of the rendered column items.
 * Each column item needs to be wrapped in a div with class a-GV-columnItem. The markup looks like this:
 *    <div class="a-GV-columnItemContainer u-vh" aria-hidden="true">
 *        <div class="a-GV-columnItem">column item markup goes here</div>
 *        ...
 *    </div>
 * todo consider changing a-GV-columnItemContainer to a-ColumnItem[s|Container]
 * Grid widget functional css takes care of hiding the column items off screen.
 *
 * The grid widget has a number of configuration options. See the comments below on the options property of the
 * widget initialization. Some of the configuration options are described in more detail in the following topics.
 *
 * Editing:
 * The grid can be editable or not editable. This is controlled by the editable option. If not editable then no UI
 * is provided to do any editing, however it will still respond to any changes to the model data. When the grid
 * is editable it has two modes; navigation mode and editing mode. The distinction is mainly for the purpose of keyboard
 * behavior. In navigation mode keyboard keys move among the grid cells. In editing mode most keys are passed through
 * to the edit controls. Edit mode pertains to cell editing only. Other kinds of edits such as deleting rows is
 * possible as long as the grid is editable. The editable property can be changed after the grid is created provided
 * the necessary column items are available on the page. See the model documentation for how it can be used to
 * provide fine grained control over what kinds of edits are allowed. The column definition can specify columns that
 * are read-only. For a cell to be editable the grid must be editable, the row must be editable (as determined by the
 * model), the column configuration must specify elementId and readonly must not be true and the cell field metadata
 * must not have a checksum (ck) property.
 *
 * Column Edit Items:
 * When the grid is editable and a column can be edited, it is a column item that does the editing. Column items are
 * essentially the same as page items except they edit a column value rather than a page item. See elementId in
 * list of column properties below.
 *
 * Column configuration:
 * Columns are configured with the columns option. The value is an object that maps a column name to a column
 * definition object. The column name is a friendly name (likely related to a database column ex: FIRST_NAME) that may be
 * used in template substitutions. The properties are the column/property names. The property value is a column definition.
 * This column structure is likely shared with the model and upper layer widgets or controller code as well.
 * The column definition properties are:
 *    elementId: {string} Id of DOM element used to edit cells in this column or null if not editable
 *    heading: {string} The text of the column header. Allows markup
 *    label: {string} The text of the column. Does not allow markup. Used by control break headers.
 *    headingAlignment: {string} One of "start", "end", "center", "left", "right". Determines how the heading text is aligned
 *                      default is "center".
 *    alignment: {string} Determines how the cell content is aligned. Takes same values headingAlignment. Optional with no default.
 *    headingCssClasses: {string} CSS classes applied to the column heading cell
 *    columnCssClasses: {string} CSS classes applied to each cell in this column
 *    cellTemplate: {string} Optional. An HTML template that defines the cell content. The template syntax is the same
 *                      as APEX server side templates. See apex.util.applyTemplate for template syntax.
 *                      The substitutions are column names from this column configuration and they are replaced with
 *                      data values from the current row.
 *    escape: {boolean} If false the column value may contain trusted markup otherwise the column value is escaped.
 *    seq: {number} Determines the order of the column among all the others. Lower numbers come first.
 *    width: {number}, optional column width
 *    noStretch: {boolean} optional. If true the column width will not stretch to fill available space as the grid resizes
 *                         The user can still adjust the width. By default columns will stretch.
 *    groupName: {string} optional name of column group
 *    useGroupFor: {string} Optional. If not present or if the string contains "heading" then the group given in groupName will be used
 *    usedAsRowHeader   If true the column is an accessible row header. It will be a th rather than a td and have scope="row".
 *    noHeaderActivate: {boolean} if true the column header does not support activation. default is false
 *    sortDirection: {string} "asc" if the data is sorted by this column ascending,
 *                            "desc" if the data is sorted by this column descending
 *                            null or not present or undefined if the data is not sorted by this column
 *    sortIndex: {number} Optional. Only applies if sortDirection is given.
 *                        The order in which this column is sorted 1 = first, 2 = second, etc.
 *    canSort: {boolean} if true the column can be sorted by the user. This controls if the header sort buttons are
 *                      shown or if the keyboard sort keys work. The grid does not actually do the sorting.
 *                      The sortChange event/callback is triggered to let the controller or parent widget know to do sorting.
 *    controlBreakIndex: {number} Optional. The order in which this column is sorted for the purpose of control breaks. Starting at 1.
 *    controlBreakDirection: {string} "asc" if the control break data is sorted by this column ascending,
 *                            "desc" if the control break data is sorted by this column descending
 *                            null or not present or undefined if the column is not a control break
 *    hidden: {boolean} If true the column is not rendered
 *    frozen: {boolean} If true the column is frozen (does not h-scroll). Only the start most columns can be frozen.
 *                      The last column cannot be frozen.
 *    readonly: {boolean} If true the column cannot be edited
 *    linkText: {string} Optional. The anchor content. Allows markup. Allows substitutions just like the cellTemplate property.
 *                       If not given the rendered display value of this column cell is used as the content. If the display
 *                       value of the column is empty then the link URL is used.
 *                       To display a link, at least one of linkTargetColumn or the cell field metadata url property must must be given.
 *                       Note: if the cell is editable it is always the data value of the cell that is edited.
 *                       So if you want to edit the link text is it best to omit linkText and use linkTargetColumn.
 *    linkTargetColumn: {string} Optional. The name of the column that contains the anchor href.
 *                               If not given the href comes from the cell field metadata url property. If there is
 *                               no url property then this column/cell is not a link.
 *    linkAttributes: {string} Optional. Additional anchor attributes. Allows substitutions just like the cellTemplate property.
 *                             Ignored unless a link is present.
 *    isRequired {boolean}  If true the column is indicated as required by adding is-required class to the column header
 *                          This should correspond with the required setting of the underlying editable column item.
 *    helpid {string} Help id for the column. If present pressing the help key Alt+F1 will display the help text for the field.
 *
 *    The "property" property is added to the definition. Its value is the column name (property from the columns map.
 *
 * Column methods:
 *    getColumns
 *    refreshColumns
 *    setColumnWidth
 *    hideColumn
 *    showColumn
 *    freezeColumn
 *    unfreezeColumn
 *    moveColumn
 *
 * Column Group definitions:
 *    heading: {string} The text of the column header. Allows markup.
 *    label: {string} [not used] Same as heading but markup not allowed.
 *    headingAlignment: {string} One of "start", "end", "center", "left", "right". Determines how the heading text is aligned
 *                      default is "center".
 *    parentGroupName: {string} Optional name of the parent column group.
 *
 * Selection:
 * The grid supports either single or multiple selection. Rows can be selected even for grids that are not editable
 * Only row selection is supported. (i.e. no column or cell selection). For multiple selection standard keyboard
 * modifiers Shift and Ctrl are always supported. In addition the rowHeaderCheckbox option allows for checkbox style
 * selection behavior.
 *
 * Context Menus:
 * The contextMenuAction option allows you to respond to mouse or keyboard interactions that typically result in a
 * context menu. Specifically right mouse click (via contextmenu event), Shift-F10 key (via keydown event) and the
 * Windows context menu key (via contextmenu event). The original event is passed to the contextMenuAction function.
 * The event object can be used to position the menu. If you implement your own menu it is best if you put focus
 * back on the grid using the grid focus method when the menu closes (unless the menu action directs focus
 * elsewhere). A simpler alternative that uses the APEX menu widget is to pass in a menu widget options object as the
 * contextMenu option. When the contextMenu option is used the beforeOpen menu callback ui argument has these
 * additional properties:
 *  - menuElement the menu jQuery object
 *  - grid this grid jQuery object
 *  - selection a jQuery object with the grid selection at the time the menu was opened
 * Also the afterClose callback will automatically focus the grid if the menu action didn't take the focus and
 * the ui argument has these additional properties:
 *  - menuElement the menu jQuery object
 *  - grid this grid jQuery object
 * Only one of contextMenuAction and contextMenu should be specified. The contextMenu option can only be set when the
 * grid is initialized and it can't be changed. The contextMenuAction cannot be set if the contextMenu option was
 * given when the grid was created.
 * If using contextMenu option the contextMenuId option can be used to give the menu element an ID. This is useful
 * if other code must refer to the menu element or widget.
 * Note: If using contextMenu option make sure necessary menu and jQuery UI css and js resources are loaded on the page.
 * The context menu only applies to grid cells - not to header cells.
 *
 * todo
 * - Currently not visible Required columns or columns which are needed as parent for Cascading LOVs and which don't have
 *   a default value specified will automatically be displayed when switching into Edit Mode.
 * - "If a row is a parent row in a parent – child relationship, all child rows will automatically be marked for deletion too.
 *    Reverting the delete will also automatically revert the deletion of the child rows. In the child Interactive Grid, it's
 *    not possible to revert single deleted rows if the parent row is marked for deletion." Multiple models and grids are involved how is this coordinated?
 * - for inline edit items how to for example check the checkbox on click, drop down the select list on click, open the
 *   date picker/color picker on click.
 * - what if total width of all frozen columns [nearly] exceeds the client area width? This can happen because of
 *    browser window or container resize. Should columns automatically be unfrozen until there is enough room to scroll?
 * - scroll pagination; with no total, interaction with control breaks
 *
 * Future possibilities:
 * - support type to go into edit mode
 * - support double click on the resize border to adjust the width to accommodate the widest value (perhaps with limits)
 * - use a class rather than show hide for buttons, drag thumb
 * - for popup edit columns consider if the resize and move should be keyboard accessible i.e. use Shift+arrow to move and Ctrl+arrow to resize
 *
 * This widget requires some functional CSS.
 *
 * Depends:
 *    jquery.ui.core.js
 *    jquery.ui.widget.js
 *    jquery.ui.draggable.js (only when resizeColumns option is true)
 *    jquery.ui.tooltip.js (for showing error messages)
 *    apex/tooltipManager.js (for tooltips)
 *    apex/widget.tableModelViewBase.js
 *    apex/util.js
 *    apex/lang.js
 *    apex/debug.js
 *    apex/model.js
 *    apex/item.js
 *    apex/widget.stickyWidget.js (only if stickyTop is not false)
 *    (the following are for context menu integration)
 *    jquery.ui.position.js
 *    apex/widget.menu.js
 */
/*global apex,$v*/
(function ( util, model, debug, lang, item, $ ) {
    "use strict";

    var C_GRID = "a-GV",
        C_GRID_HEADER = "a-GV-hdr",
        SEL_GRID_HEADER = "." + C_GRID_HEADER,
        C_GRID_BODY = "a-GV-bdy",
        SEL_GRID_BODY = "." + C_GRID_BODY,
        C_GRID_WRAP_FROZEN = "a-GV-w-frozen",
        C_GRID_WRAP_HEADER = "a-GV-w-hdr",
        C_GRID_WRAP_SCROLL = "a-GV-w-scroll",
        C_GRID_TABLE = "a-GV-table",
        C_GRID_ROW = "a-GV-row",
        SEL_ROW = "tr",
        C_GRID_CELL = "a-GV-cell",
        SEL_GRID_CELL = "." + C_GRID_CELL,
        C_GRID_SCROLL_FILLER = "a-GV-scrollFiller",
        C_GRID_COL_HEADER = "a-GV-header",
        SEL_GRID_COL_HEADER = "." + C_GRID_COL_HEADER,
        C_GRID_COL_GROUP = "a-GV-headerGroup",
        SEL_GRID_COL_GROUP = "." + C_GRID_COL_GROUP,
        C_GRID_ROW_SELECT = "a-GV-rowSelect",
        SEL_GRID_ROW_SELECT = "." + C_GRID_ROW_SELECT,
        C_GRID_HEADER_LABEL = "a-GV-headerLabel",
        C_GRID_COL_CONTROLS = "a-GV-columnControls",
        SEL_GRID_COL_CONTROLS = "." + C_GRID_COL_CONTROLS,
        C_GRID_COL_HANDLE = "a-GV-columnHandle",
        C_GRID_ROW_HEADER = "a-GV-rowHeader",
        SEL_GRID_ROW_HEADER = "." + C_GRID_ROW_HEADER,
        SEL_GRID_SELECT_ALL = SEL_GRID_ROW_HEADER + SEL_GRID_COL_HEADER,
        SEL_GRID_CELL_OR_HEADER = ".a-GV-header,.a-GV-headerGroup,.a-GV-cell,.a-GV-controlBreakHeader",
        C_CONTROL_BREAK = "a-GV-controlBreak",
        C_GRID_BREAK_HEADER = "a-GV-controlBreakHeader",
        SEL_GRID_BREAK_HEADER = "." + C_GRID_BREAK_HEADER,
        SEL_GRID_CELL_OR_BREAK = "." + C_GRID_CELL + ",." + C_GRID_BREAK_HEADER,
        C_TAB_TRAP = "js-tabTrap",
        SEL_TAB_TRAP = "." + C_TAB_TRAP,
        C_FLOATING_ITEM_CONT = "a-GV-floatingItemContent",
        SEL_FLOATING_ITEM_CONT = "." + C_FLOATING_ITEM_CONT,
        C_SELECTED = "is-selected",
        SEL_SELECTED = "." + C_SELECTED,
        C_FOCUSED = "is-focused",
        C_HOVER = "is-hover",
        SEL_HOVER = "." + C_HOVER,
        C_DISABLED = "is-disabled",
        C_EXPANDED = "is-expanded",
        C_ACTIVE = "is-active",
        SEL_ACTIVE = "." + C_ACTIVE,
        SEL_ACTIVE_CELL = SEL_ACTIVE + SEL_GRID_CELL,
        C_DELETED = "is-deleted",
        SEL_DELETED = "tr." + C_DELETED,
        ARIA_SELECTED = "aria-selected",
        ARIA_MULTI = "aria-multiselectable",
        ARIA_RO = "aria-readonly",
        ARIA_LABEL = "aria-label",
        ARIA_EXPANDED = "aria-expanded",
        ARIA_LBL_BY = "aria-labelledby",
        C_RTL = "u-RTL",
        SEL_VISIBLE = ":visible",
        SEL_TABBABLE = ":tabbable";

    var NONE = "none",
        SEQUENCE = "sequence",
        LABEL = "label";

    var EVENT_SELECTION_CHANGE = "selectionChange",
        EVENT_MODE_CHANGE = "modeChange",
        EVENT_SORT_CHANGE = "sortChange",
        EVENT_COLUMN_REORDER = "columnReorder",
        EVENT_COLUMN_RESIZE = "columnResize",
        EVENT_ACTIVATE_COLUMN_HEADER = "activateColumnHeader",
        EVENT_CANCEL_COLUMN_HEADER = "cancelColumnHeader",
        EVENT_PAGE_CHANGE = "pageChange";

    var COL_WIDTH_INC = 10,
        MIN_COL_WIDTH = 40, // THINK consider putting this back to 32 once there is a better way to set width of actions column
        FROZEN_LABEL_WIDTH = 200, // if frozen area is wider than this then it will show labels such as load more button and column breaks
        HEADER_SCROLL_SENSITIVITY = 10,
        HEADER_SCROLL_SPEED = 10,
        ATTR_TITLE = "title",
        ATTR_COLSPAN = "colspan",
        ATTR_DATA_IDX = "data-idx",
        ATTR_DATA_ID = "data-id";

    var keys = $.ui.keyCode;

    function alignmentClass( alignment, centerIsDefault ) {
        var cls = "";
        if ( alignment === "start" ) {
            cls = " u-tS";
        } else if ( alignment === "end" ) {
            cls = " u-tE";
        } else if ( alignment === "left" ) {
            cls = " u-tL";
        } else if ( alignment === "right" ) {
            cls = " u-tR";
        } else if ( alignment === "center" && !centerIsDefault ) {
            cls = " u-tC";
        }
        return cls;
    }

    function domIndex( el$ ) {
        return el$.parent().children().index( el$ );
    }

    function getRows$( out ) {
        var rows$ = $( out.toString() );
        rows$.find( "a,button,input,select,textarea" ).prop( "tabIndex", -1 );
        return rows$;
    }

    function getContainingTableCell$( el ) {
        return $( el ).closest( SEL_GRID_CELL_OR_HEADER );
    }

    function getContainingGridCell$( el ) {
        return $( el ).closest( SEL_GRID_CELL_OR_BREAK );
    }

    function getContainingDataCell$( el ) {
        return $( el ).closest( SEL_GRID_CELL );
    }

    function cellFromColumnIndex( row$, column ) {
        var cell$, index,
            cells$ = row$.children();

        if ( cells$.last().hasClass( C_GRID_COL_GROUP ) ) {
            index = 0;
            cells$.each(function() {
                index += parseInt( $(this).attr(ATTR_COLSPAN), 10 ) || 1;
                if ( index > column ) {
                    cell$ = $( this );
                    return false;
                }
            });
        } else {
            cell$ = cells$.eq( column );
        }
        return cell$;
    }

    function columnIndexFromCell( cell$ ) {
        var index;
        if ( cell$.hasClass( C_GRID_COL_GROUP ) ) {
            index = 0;
            cell$.parent().children().each(function() {
                if ( this === cell$[0] ) {
                    return false;
                }
                index += parseInt( $(this).attr(ATTR_COLSPAN), 10 ) || 1;
            });
        } else {
            index = domIndex( cell$ );
        }
        return index;
    }

    var E_KEYDOWN_ESCAPE = "keydown.gridGlobalEscape";
    function escapeOn( f ) {
        $( "body" ).on( E_KEYDOWN_ESCAPE, function( event ) {
            if ( event.which === keys.ESCAPE ) {
                $( "body" ).off( E_KEYDOWN_ESCAPE );
                f();
            }
        });
    }

    function escapeOff() {
        $( "body" ).off( E_KEYDOWN_ESCAPE );
    }

    function effectiveGroupName( col ) {
        var group = col.groupName || null;
        if ( group && col.useGroupFor && col.useGroupFor.indexOf( "heading" ) < 0 ) {
            group = null;
        }
        return group;
    }

    $.widget( "apex.grid", $.apex.tableModelViewBase, {
        version: "5.1",
        widgetEventPrefix: "grid",
        options: {
            // see also options from tableModelViewBase
            showNullAs: "-", // what to display when a value is null
            stickyTop: false,         // only applies if hasSize is false. If false the column header will not stick to the page.
                                      // if true or a function the column header will stick to the top of the page using the stickyWidget
                                      // if a function then it is passed to the stickyWidget as the top option.
            columns: null,      // Required. An array containing a single element that is a map of column definitions
                                // The array wrapper is so that it is not copied. This helps sharing the column definitions
                                // between the model, this widget and other widgets.
                                // todo THINK about if this is necessary. How is column metadata related to options needed when the model requests data from server?
            columnGroups: null, // optional map of column groups
            columnSort: true,   // if true the mouse and keyboard can be used in column headings to adjust the sort order
                                // Note the data can still be sorted external to the grid
            columnSortMultiple: true, // if true multiple columns can be sorted. This affects the sort control header buttons.
                                // The caller is still responsible for actually doing/causing the sorting.
            resizeColumns: true,  // if true the mouse and keyboard can be used in column headings to adjust the width of columns
                                // Note the column widths can still be changed external to the grid
            reorderColumns: true,  // if true the mouse and keyboard can be used in column headings to reorder the columns or column groups
                                // Note the column order can still be changed external to the grid
            aggregateLabels: {},// map of <aggregate name> to aggregate label and overallLabel Example: { "sum": { label: "Sum", overallLabel: "Total Sum" }, ... }
            aggregateTooltips: {}, // map of <aggregate name> + "|" + <aggregate column> to tooltip text. Example { { "AVG_SALARY": "Average Salary", .... }
            tabbableCellContent: "a,button", // selector that identifies cell content that can be a tab stop in navigation mode
            editable: false,    // if true the grid data can be edited according to what the model allows
            allowEditMode: true, // only applies if editable is true. If false then can't go in or out of edit mode using mouse or keyboard
            allowDelete: true,  // only applies if editable is true. If false then can't use delete key to delete a row
            allowInsert: true,  // only applies if editable is true. If false then can't use Insert key to add a row
            highlights: {},     // mapping of highlight id to { seq: {int}, row: <bool>, color: <color>, background: <color>, cssClass: <string> }
                                // for all highlights that may be used in the grid.
            // When the grid is editable or when multiple selection is allowed it is a good idea to have a rowHeader but it is not enforced
            rowHeader: NONE,  // One of: none, plain, sequence, label.
                                // a value other than none is required for editable grids in order to see visual indicators such as row level errors
            rowHeaderCheckbox: false, // only applies if rowHeader is not none
            rowHeaderLabelColumn: null, // only applies if rowHeader is label. This is a model column to take the row header label value from. Can include markup
            rowHeaderWidth: null, // only applies if rowHeader is not none. The default chooses the width based on the kind of rowHeader.
            multiple: false, // when true multiple rows can be selected.
            selectAll: true, // only applies when multiple is true. If true then the current page/all visible rows can
                             // be selected with Ctrl+A and/or a select all checkbox if rowHeaderCheckbox is true.
            persistSelection: false, // if true the selection state for each row will be saved as record metadata in the model
            tooltip: {          // tooltip object suitable for jQuery UI tooltip widget except that items is not needed.
                                // It is supplied by the grid and cannot be overridden. It matches td,th. Also the open option is not available.
                                // Default values are provided for tooltipClass, show, and todo position but can be overridden.
                                // The content callback function receives extra arguments: model, recordMeta, colMeta, columnDef.
                                // (recordMeta.record provides access to the record data.)
                                // tooltips are used to show errors and warnings at the row and cell level. The content function is not
                                // called if there is an error or warning message to display. If tooltip is null the
                                // error or warning message is added as a title attribute.
                tooltipClass: "a-GV-tooltip"
            },

            contextMenuAction: null, // optional. function( event ) called when a context menu should be displayed
            contextMenu: null, // optional. A menu options object suitable for the APEX menu widget. Only specify one of contextMenu and contextMenuAction.
            contextMenuId: null,    // optional. Element id to give the internal context menu. Or if contextMenu is not given
                                    // this is the id of a menu that already exists.

            //
            // events:
            //
            selectionChange: null,  // This event is fired when the selection changes. It has no additional data
                                    // function( event )
            activateColumnHeader: null, // This event fires when a column header is activated (Enter/Space key or click)
                                        // function( event, { header$: <column-element>, column: <column-definition> } )
            cancelColumnHeader: null,   // This event fires when whatever popup is opened in response to activateColumnHeader
                                        // should be closed/canceled. This happens because of another operation on the
                                        // column such as dragging.
                                        // function( event, { } )
            sortChange: null,       // This event fires when the sort direction changes. This does not actually
                                    // sort the data or ask the model to sort or fetch new data. It is expected that
                                    // a handler will call refreshColumns and then take action that causes the data to be sorted.
                                    // function( event, { header$: <column-element>, column: <column-definition>, direction: <asc-or-desc>, action: "add"|"remove"|"change" } )
            columnReorder: null,    // This event fires when the columns have been reordered.
                                    // function( event, { header$: <column-element>, column: <column-definition>, newPosition: <n>, oldPosition: <n> } )
            columnResize: null,     // This event fires when a column width has been changed.
                                    // function( event, { header$: <column-element>, column: <column-definition>, width: <n> } )
            modeChange: null,       // This event fires when the edit mode changes
                                    // function( event, { editMode: boolean } )
            pageChange: null       // This event fires when there is a pagination event that results in new records
                                    // being displayed. function( event, { offset: number, count: number } )
        },
        forwardKey: keys.RIGHT,
        backwardKey: keys.LEFT,

        _create: function () {
            var wheelEvent,
                self = this,
                o = this.options,
                ctrl$ = this.element;

            debug.info("Grid '" + ctrl$[0].id + "' created. Model: " + o.modelName );

            ctrl$.addClass( C_GRID )
                .attr( "role", "grid" );

            if ( o.multiple ) {
                ctrl$.attr( ARIA_MULTI, "true" );
            }
            if ( !o.editable ) {
                ctrl$.attr( ARIA_RO, "true" );
            }

            if ( ctrl$.css("direction") === "rtl" ) {
                ctrl$.addClass( C_RTL );
                this.forwardKey = keys.LEFT;
                this.backwardKey = keys.RIGHT;
            }

            this._enforceOptionConstraints();

            this._super(); // init table model view base

            this.tableFrozenHeader$ = null;  // table > tbody element of the frozen header table if present
            this.tableHeader$ = null;        // table > tbody element of the header table
            this.tableFrozenData$ = null;    // table > tbody element of the frozen data column table if present
            this.tableData$ = null;          // table > tbody element of the data table
            this.lastEditModeCell = null;

            this.editMode = false;
            this.columns = null;              // sorted array of column metadata from options.columns
            this.columnGroups = null;         // array of sorted arrays of column groups
            this.lastFrozenColIndex = -1;     // index of last frozen column or -1 if no frozen columns
            this.lastFocused = null;          // last cell that had focus or has focus could be button/link inside a cell
            this.lastRow$ = null;             // All the columns (td) in the row that contains lastFocused. This will be all
            // the columns from one or two tr elements depending on if there are frozen columns
            this.focusInHeader = false;       // if true the lastFocused cell is in the table header
            this.lastHoverCell = null;        // used to track mouse hover
            this.lastHoverRow = null;         // used to track mouse hover
            this.columnControls$ = null;
            this.columnHandle$ = null;
            this.scrollTimerId = null;
            this.deactivateDelayTimer = null;

            this.gotoCellPending = null;

            if ( o.contextMenu || o.contextMenuId ) {
                if ( $.apex.menu ) {
                    if ( o.contextMenu ) {
                        if ( o.contextMenu.menubar ) {
                            throw new Error( "Grid contextMenu must not be a menubar" );
                        }
                        // augment the menu
                        o.contextMenu._originalBeforeOpen = o.contextMenu.beforeOpen;
                        o.contextMenu.beforeOpen = function( event, ui ) {
                            if ( o.contextMenu._originalBeforeOpen ) {
                                ui.menuElement = self.contextMenu$;
                                ui.grid = ctrl$;
                                ui.selection = self.getSelection();
                                o.contextMenu._originalBeforeOpen( event, ui );
                            }
                        };
                        o.contextMenu._originalAfterClose = o.contextMenu.afterClose;
                        o.contextMenu.afterClose = function( event, ui ) {
                            if ( o.contextMenu._originalAfterClose ) {
                                ui.menuElement = self.contextMenu$;
                                ui.grid = ctrl$;
                                o.contextMenu._originalAfterClose( event, ui );
                            }
                            if ( !ui.actionTookFocus ) {
                                self.focus();
                            }
                        };
                        this.contextMenu$ = $( "<div style='display:none'></div>" ).appendTo( "body" );
                        if ( o.contextMenuId ) {
                            this.contextMenu$[0].id = o.contextMenuId;
                        }
                        this.contextMenu$.menu( o.contextMenu );
                    } else {
                        // must have only a contextMenuId so use that externally defined menu
                        this.contextMenu$ = $( "#" + o.contextMenuId );
                        if ( this.contextMenu$.length === 0 || !this.contextMenu$.is( ":apex-menu" ) ) {
                            throw new Error( "Grid contextMenuId not found" );
                        }
                    }
                    if ( o.contextMenuAction ) {
                        debug.warn("Grid contextMenuAction option ignored when contextMenu option present");
                    }
                    o.contextMenuAction = function( event ) {
                        var target$, pos;

                        if ( event.type === "contextmenu" && event.button !== 0 ) {
                            self.contextMenu$.menu( "toggle", event.pageX, event.pageY );
                        } else {
                            target$ = $( event.target );
                            pos = target$.offset();
                            self.contextMenu$.menu( "toggle", pos.left, pos.top + target$.closest( "tr" ).outerHeight() );
                        }
                    };
                } else {
                    debug.warn("Grid contextMenu option ignored because menu widget not preset");
                }
            }

            // get the model
            this._initModel( o.modelName );

            // Modern browsers support "wheel", check wheel because only need to do this once for all instances
            if ( !("onwheel" in ctrl$[0]) && this._eventHandlers.wheel) {
                if ( document.onmousewheel !== undefined ) {
                    wheelEvent = "mousewheel"; // Webkit and IE support at least "mousewheel"
                } else {
                    wheelEvent = "DOMMouseScroll"; // otherwise assume that remaining browsers are older Firefox
                }
                // move the event handler
                this._eventHandlers[wheelEvent] = this._eventHandlers.wheel;
                delete this._eventHandlers.wheel;
            }

            this._on( this._eventHandlers );

            this._updateHighlights();
            this.refresh();

            if ( o.disabled ) {
                this._setOption( "disabled", o.disabled );
            }
        },

        _eventHandlers: {
            resize: function( event ) {
                if (event.target !== this.element[0]) {
                    return;
                }
                this.resize();
                event.stopPropagation();
            },
            click: function ( event ) {
                var cell$, allChecked,
                    self = this,
                    o = this.options,
                    target$ = $( event.target );

                // ignore click, shift and ctrl click on anchors and buttons and anything in the footer to let the browser do its thing
                if ( target$.closest( "a,button,.a-GV-footer" ).length > 0 ) {
                    return;
                }

                // ignore clicks in the the floating (popup) edit column item
                if ( target$.closest( ".a-GV-floatingItem" ).length > 0 ) {
                    return;
                }

                // Check if click on the select all header
                cell$ = target$.closest( SEL_GRID_SELECT_ALL );
                if ( cell$.length ) {
                    this.focusInHeader = true;
                    self.lastColIndex = null;
                    if ( o.rowHeaderCheckbox && o.multiple && o.selectAll ) {
                        allChecked = cell$.find( SEL_GRID_ROW_SELECT )[0].checked;
                        if ( target$.hasClass( C_GRID_ROW_SELECT ) ) {
                            allChecked = !allChecked; // if the click was on the checkbox then it had already changed state
                        }
                        // change the selection after a short delay to avoid changing the check state while the
                        // checkbox is being clicked on
                        setTimeout(function() {
                            // if currently all selected then select none otherwise select all.
                            if ( allChecked ) {
                                self.setSelection( $() );
                            } else {
                                self.selectAll();
                            }
                            self._select( cell$, event, true, false );
                            self._setFocusable( cell$[0] ); // need to force this because of setting selection twice
                        }, 10);
                        event.preventDefault();
                    } else {
                        self._select( cell$, event, true, false );
                    }
                    return;
                }

                // Check if click on a column group header. They don't do much but need to take focus
                cell$ = target$.closest( SEL_GRID_COL_GROUP );
                if ( cell$.length ) {
                    this.focusInHeader = true;
                    self.lastColIndex = null;
                    self._select( cell$, event, true, false );
                    event.preventDefault();
                    return;
                }

                // Check if click on any other header cell and not on a button
                cell$ = target$.closest( SEL_GRID_COL_HEADER );
                if ( cell$.length && target$.closest( "button" ).length === 0 ) {
                    this.focusInHeader = true;
                    self.lastColIndex = null;
                    self._select( cell$, event, true, false );
                    this._activateColumnHeader( event, cell$ );
                    event.preventDefault();
                    return;
                }

                // Check if click on a row header checkbox
                if ( target$.hasClass( C_GRID_ROW_SELECT ) ) {
                    // delay this until after the checkbox is clicked so _select can set the check state
                    cell$ = target$.closest( SEL_GRID_ROW_HEADER );
                    setTimeout(function() {
                        event.ctrlKey = true;
                        event.shiftKey = false;
                        self.focusInHeader = false;
                        self.lastColIndex = null;
                        self._select( cell$, event, true, false );
                    }, 10);
                    return;
                }
                // Check if click on a row header
                cell$ = target$.closest( SEL_GRID_ROW_HEADER );
                if ( cell$.length ) {
                    if ( o.rowHeaderCheckbox ) {
                        // always behave like toggle
                        event.ctrlKey = true;
                        event.shiftKey = false;
                    }
                    this.focusInHeader = false;
                    self.lastColIndex = null;
                    this._select( cell$, event, true, false );
                    event.preventDefault();
                    return;
                }

                // Check if click on any data cell
                cell$ = target$.closest( SEL_GRID_CELL );
                if ( cell$.length ) {
                    this.focusInHeader = false;
                    self.lastColIndex = null;
                    if ( !cell$.hasClass( C_ACTIVE ) ) {
                        this._select( cell$, event, true, false );
                        event.preventDefault();
                    }
                }
                cell$ = target$.closest( SEL_GRID_BREAK_HEADER );
                if ( cell$.length ) {
                    this.focusInHeader = false;
                    self.lastColIndex = null;
                    this._select( cell$, event, true, false );
                    event.preventDefault();
                }
            },
            dblclick: function ( event ) {
                var o = this.options,
                    cell$ = $( event.target ).closest( SEL_GRID_CELL ).not( SEL_GRID_ROW_HEADER );

                if ( cell$.length && !this.editMode && o.editable && o.allowEditMode ) {
                    this.setEditMode( true );
                }
            },
            mousedown: function ( event ) {
                var target$ = $( event.target );

                // ignore this event for anything in the footer
                if ( target$.closest( ".a-GV-footer" ).length > 0 ) {
                    return;
                }

                // preventing default for mouse down prevents some cases of text selection and also
                // makes it so that the focus happens after the click. This includes preventing mouse down on the
                // expand/collapse button so it can be clicked without deactivating the cell.
                // But if the click is on the active editing cell don't prevent it so the click gets through to the editing controls
                if ( target$.closest( ".a-GV-expandCollapse" ).length > 0 ||
                        !( this.editMode && (
                            target$.closest( SEL_GRID_CELL ).hasClass( C_ACTIVE ) ||
                            target$.closest( ".a-GV-floatingItem" ) ) )) {
                    event.preventDefault();
                }
            },
            keydown: function ( event ) {
                var i, colIndex, cur$, column, pos, page, groupLevel, sel, w,
                    self = this,
                    o = this.options,
                    next$ = null,
                    kc = event.which;

                function activateHeader() {
                    if ( cur$.hasClass( C_GRID_ROW_HEADER ) ) {
                        // simulate a click on select all checkbox if there is one
                        cur$.find( SEL_GRID_ROW_SELECT ).click();
                    } else {
                        self._activateColumnHeader( event, cur$ );
                        event.preventDefault();
                    }
                }

                function findNextColumnPosition( pos, inc ) {
                    var i,
                        nextPos = null;

                    pos += inc;
                    for ( i = pos; i >= 0 && i < self.columns.length; i += inc ) {
                        if ( !self.columns[i].hidden ) {
                            nextPos = i;
                            break;
                        }
                    }
                    return nextPos;
                }

                // ignore key events for anything in the footer
                if ( $( event.target ).closest( ".a-GV-footer" ).length > 0 ) {
                    return;
                }

                if ( kc === keys.PAGE_UP || kc === keys.PAGE_DOWN ) {
                    page = this.displayPageSize;
                }

                if ( kc === keys.HOME ) {
                    if ( !this.editMode && this.lastFocused ) {
                        this.lastColIndex = null;
                        next$ = this.lastRow$.first();
                    }
                } else if ( kc === keys.END ) {
                    if ( !this.editMode && this.lastFocused ) {
                        this.lastColIndex = null;
                        next$ = this.lastRow$.last();
                    }
                } else if ( kc === keys.PAGE_DOWN ) {
                    if ( !this.focusInHeader ) {
                        next$ = this._getNextCellDown( getContainingGridCell$( this.lastFocused ), page );
                    }
                } else if ( kc === keys.PAGE_UP ) {
                    if ( !this.focusInHeader ) {
                        next$ = this._getPrevCellUp( getContainingGridCell$( this.lastFocused ), page, true );
                    }
                } else if ( kc === keys.DOWN ) {
                    cur$ = getContainingTableCell$( this.lastFocused );
                    if ( o.columnSort && event.altKey && cur$.hasClass( C_GRID_COL_HEADER )) {
                        self._sortChange( event, cur$, "desc" );
                        return;
                    }
                    if ( (!this.editMode || this.focusInHeader) && this.lastFocused ) {
                        next$ = this._getNextCellDown( cur$, 1 );
                    }
                    // don't let a menu button do it's down arrow thing
                    if ( $( event.target ).closest( ".js-menuButton" ).length ) {
                        event.stopPropagation();
                    }
                    // don't let selection happen
                    if ( event.shiftKey ) {
                        event.preventDefault();
                    }
                } else if ( kc === keys.UP ) {
                    cur$ = getContainingTableCell$( this.lastFocused );
                    if ( o.columnSort && event.altKey && cur$.hasClass( C_GRID_COL_HEADER )) {
                        self._sortChange( event, cur$, "asc" );
                        return;
                    }
                    if ( (!this.editMode || this.focusInHeader) && this.lastFocused ) {
                        next$ = this._getPrevCellUp( cur$, 1 );
                        event.stopPropagation(); // Don't let a containing tab or accordion act on Ctrl+Up
                    }
                } else if ( kc === this.backwardKey ) {
                    if ( event.altKey ) {
                        return; // let the browser have this key
                    }
                    cur$ = getContainingTableCell$( this.lastFocused );
                    if ( ( event.ctrlKey || event.shiftKey ) && (
                                    cur$.hasClass( C_GRID_COL_HEADER ) || cur$.hasClass( C_GRID_COL_GROUP )
                                    ) && ( !cur$.hasClass ( C_GRID_ROW_HEADER ) || o.rowHeader === SEQUENCE || o.rowHeader === LABEL ) ) {
                        pos = parseInt( cur$.attr( ATTR_DATA_IDX ), 10 );
                        if ( !( isNaN( pos ) && !event.ctrlKey && o.rowHeader !== SEQUENCE && o.rowHeader !== LABEL ) ) {
                            // must not be a row header, which can't be reordered and only resized when rowHeader is sequence or text
                            if ( pos >= 0 ) {
                                column = this.columns[pos];
                            }
                            if ( event.shiftKey && o.reorderColumns ) {
                                event.preventDefault(); // do this before moveColumn redraws the grid
                                // reorder column
                                if ( cur$.hasClass( C_GRID_COL_GROUP ) ) {
                                    groupLevel = domIndex( cur$.parent() );
                                    if ( pos > 0 ) {
                                        this.moveColumnGroup( groupLevel, pos, pos - 1 );
                                        cur$ = $( this.lastFocused ).closest( SEL_GRID_COL_GROUP );
                                    }
                                } else {
                                    pos = findNextColumnPosition( pos, -1 );
                                    if ( pos ) {
                                        this.moveColumn( column, pos );
                                        // the grid has been re-rendered but lastFocused should have been maintained
                                        // set cur$ again so navigate to previous column can happen bellow
                                        // xxx sometimes (around a hidden col) lastFocus is not maintained and focus returns to the start of the header row
                                        cur$ = $( this.lastFocused ).closest( SEL_GRID_COL_HEADER );
                                    }
                                }
                            } else if ( event.ctrlKey && !cur$.hasClass( C_GRID_COL_GROUP ) ) { // else must be resize but that doesn't apply to group headers
                                // make column narrower
                                if ( column ) {
                                    if ( cur$.outerWidth() > MIN_COL_WIDTH ) {
                                        w = cur$.outerWidth() - COL_WIDTH_INC;
                                        if ( w < MIN_COL_WIDTH ) {
                                            w = MIN_COL_WIDTH;
                                        }
                                        this.setColumnWidth( column, w );
                                    }
                                } else {
                                    if ( self.rowHeaderWidth > MIN_COL_WIDTH ) {
                                        w = self.rowHeaderWidth - COL_WIDTH_INC;
                                        if ( w < MIN_COL_WIDTH ) {
                                            w = MIN_COL_WIDTH;
                                        }
                                        self.rowHeaderWidth = w;
                                        this._calculateColumnWidths( cur$, true );
                                    }
                                }
                                return;
                            }
                        }
                    }
                    if ( ( !this.editMode || this.focusInHeader ) && this.lastFocused ) {
                        this.lastColIndex = null;
                        colIndex = this.lastRow$.index( cur$ );
                        if ( colIndex > 0) {
                            colIndex -= 1;
                            next$ = this.lastRow$.eq( colIndex );
                        }
                        event.preventDefault();
                    }
                } else if ( kc === this.forwardKey ) {
                    if ( event.altKey ) {
                        return; // let the browser have this key
                    }
                    cur$ = getContainingTableCell$( this.lastFocused );
                    if ( ( event.ctrlKey || event.shiftKey ) && (
                                    cur$.hasClass( C_GRID_COL_HEADER ) || cur$.hasClass( C_GRID_COL_GROUP )
                                    ) && ( !cur$.hasClass ( C_GRID_ROW_HEADER ) || o.rowHeader === SEQUENCE || o.rowHeader === LABEL ) ) {
                        pos = parseInt( cur$.attr( ATTR_DATA_IDX ), 10 );
                        if ( !( isNaN( pos ) && !event.ctrlKey && o.rowHeader !== SEQUENCE && o.rowHeader !== LABEL ) ) {
                            // must not be a row header, which can't be reordered and only resized when rowHeader is sequence or text
                            if ( pos >= 0 ) {
                                column = this.columns[pos];
                            }
                            if ( event.shiftKey && o.reorderColumns ) {
                                event.preventDefault(); // do this before moveColumn redraws the grid
                                // reorder column
                                if ( cur$.hasClass( C_GRID_COL_GROUP ) ) {
                                    groupLevel = domIndex( cur$.parent() );
                                    if ( pos < this.columnGroups[groupLevel].length - 1 ) {
                                        this.moveColumnGroup( groupLevel, pos, pos + 1 );
                                        cur$ = $( this.lastFocused ).closest( SEL_GRID_COL_GROUP );
                                    }
                                } else {
                                    pos = findNextColumnPosition( pos, 1 );
                                    if ( pos ) {
                                        this.moveColumn( column, pos );
                                        // the grid has been re-rendered but lastFocused should have been maintained
                                        // set cur$ again so navigate to next column can happen bellow
                                        // xxx sometimes (around a hidden col) lastFocus is not maintained and focus returns to the start of the header row
                                        cur$ = $( this.lastFocused ).closest( SEL_GRID_COL_HEADER );
                                    }
                                }
                            } else if ( event.ctrlKey && !cur$.hasClass( C_GRID_COL_GROUP ) ) { // else must be resize but that doesn't apply to group headers
                                // make column wider
                                if ( column ) {
                                    this.setColumnWidth( column, cur$.outerWidth() + COL_WIDTH_INC );
                                } else {
                                    self.rowHeaderWidth += COL_WIDTH_INC;
                                    this._calculateColumnWidths( cur$, true );
                                }
                                return;
                            }
                        }
                    }
                    if ( ( !this.editMode || this.focusInHeader ) && this.lastFocused ) {
                        this.lastColIndex = null;
                        colIndex = this.lastRow$.index( cur$ );
                        if ( colIndex < this.lastRow$.length - 1 ) {
                            colIndex += 1;
                            next$ = this.lastRow$.eq( colIndex );
                        }
                        event.preventDefault();
                    }
                } else if ( kc === 65 && event.ctrlKey && !this.editMode ) { // Ctrl+A
                    this.selectAll();
                    event.preventDefault();
                } else if ( kc === keys.SPACE ) {
                    // ignore if on a button
                    if ( event.target.nodeName === "BUTTON" ) {
                        return;
                    }
                    cur$ = $( event.target ).find( SEL_GRID_ROW_SELECT );
                    if ( cur$.length ) {
                        cur$.click();
                        event.preventDefault();
                        return;
                    }
                    cur$ = getContainingTableCell$( this.lastFocused );
                    if ( cur$.hasClass( C_GRID_COL_HEADER ) ) {
                        activateHeader();
                    } else if ( cur$.hasClass( C_GRID_CELL ) && !this.editMode ) {
                        // a grid cell not in edit mode and not on the row header checkbox
                        this._select( cur$, event, true, true );
                        event.preventDefault();
                    }
                } else if ( kc === keys.ENTER ) {
                    // ignore when target is an anchor or button or in a popup edit field unless it is on the expand/collapse button
                    if ( !$( event.target ).hasClass( "a-GV-expandCollapse" ) && $( event.target ).closest( "a,button," + SEL_FLOATING_ITEM_CONT ).length > 0 ) {
                        return;
                    }
                    cur$ = $( event.target ).find( SEL_GRID_ROW_SELECT );
                    if ( cur$.length ) {
                        cur$.click();
                        event.preventDefault();
                        return;
                    }
                    cur$ = getContainingTableCell$( this.lastFocused );
                    if ( this.editMode && ( cur$.hasClass( C_GRID_CELL ) || cur$.hasClass( C_GRID_BREAK_HEADER ) ) ) {
                        // if in edit mode go to next cell in column
                        if ( event.shiftKey ) {
                            next$ = this._getPrevCellUp( cur$, 1 );
                            // don't move up to header
                            if ( next$ && next$.hasClass( C_GRID_COL_HEADER ) ) {
                                next$ = null;
                            }
                            // don't let the selection code act on the shift modifier
                            event.shiftKey = false;
                        } else {
                            next$ = this._getNextCellDown( cur$, 1 );
                        }
                    } else {
                        event.preventDefault();
                        if ( o.editable && o.allowEditMode && cur$.hasClass( C_GRID_CELL ) && !cur$.hasClass( C_GRID_ROW_HEADER ) ) {
                            this.setEditMode( true );
                        } else if ( cur$.hasClass( C_GRID_COL_HEADER ) ) {
                            activateHeader();
                        }
                    }
                } else if ( kc === 113 ) { // F2
                    cur$ = getContainingDataCell$( this.lastFocused );
                    if ( cur$.length && o.editable && o.allowEditMode && !this.editMode ) {
                        this.setEditMode( true );
                    }
                } else if ( kc === keys.ESCAPE ) {
                    if ( o.editable && o.allowEditMode && this.editMode ) {
                        cur$ = getContainingDataCell$( this.lastFocused );
                        if ( cur$.length ) {
                            this._setFocusable( cur$[0] );
                        }
                        // move focus so that the change event happens
                        cur$.focus();
                        this.setEditMode( false );
                        if ( cur$.length ) {
                            this._setFocus( this._getCellFocusable( cur$ ) );
                        }
                        // if we handle the escape key don't let anyone else. this for example keeps dialogs from closing
                        event.stopPropagation();
                    }
                } else if ( !this.editMode && o.contextMenuAction && (event.shiftKey && kc === 121) ) { // Shift+F10
                    // ignore when target is an anchor
                    if ( $( event.target ).closest( "a" ).length > 0 ) {
                        return;
                    }
                    // context menu only applies to grid cells and when not in edit mode
                    cur$ = getContainingDataCell$( this.lastFocused );
                    if ( cur$.length ) {
                        // if target row not selected then select it
                        if ( !cur$.parent().hasClass( C_SELECTED ) ) {
                            this._select( cur$, {}, true, false ); // empty event so that selection will be set
                        }
                        o.contextMenuAction( event );
                        event.preventDefault();
                    }
                } else if ( kc === keys.DELETE ) {
                    if ( o.editable && o.allowDelete && !this.editMode && !this.focusInHeader ) {
                        sel = this.getSelectedRecords();

                        for ( i = 0; i < sel.length; i++ ) {
                            if ( !this.model.allowDelete( sel[i] ) ) {
                                sel.splice( i, 1 );
                                i -= 1;  // i stays the same for next iteration
                            }
                        }
                        if ( sel.length > 0 ) {
                            this.model.deleteRecords( sel );
                        }
                    }
                } else if ( kc === 45 ) { // INSERT
                    if ( o.editable && o.allowInsert && !this.editMode ) {
                        if ( this.model.allowAdd() ) {
                            sel = this.getSelectedRecords();
                            if ( sel.length > 0 ) {
                                this.model.insertNewRecord( null, sel[sel.length - 1] );
                            } else {
                                this.model.insertNewRecord( );
                            }
                        }
                    }
                } else if ( kc === 112  && event.altKey ) { // Alt+F1 get column help
                    cur$ = getContainingTableCell$( this.lastFocused );
                    if ( !cur$.hasClass( C_GRID_ROW_HEADER ) && !cur$.hasClass( C_GRID_COL_GROUP ) && !cur$.hasClass( C_GRID_BREAK_HEADER ) ) {
                        column = this._getColumnMetadata( cur$ );
                        if ( column && column.helpid ) {
                            // todo if in edit mode Alt+F6 or close dialog doesn't refocus the active cell.
                            apex.theme.popupFieldHelp( column.helpid, $v( "pInstance" ) );
                        }
                    }
                }
                if ( next$ ) {
                    this._select( next$, event, true, true );
                    event.preventDefault();
                }
            },
            focusin: function ( event ) {
                var cell$, prevActive$,
                    target$ = $( event.target );

                // ignore focus changes within a floating edit popup
                if ( target$.closest( ".a-GV-floatingItem" ).length ) {
                    this._clearDeactivate();
                    return;
                }

                cell$ = target$.closest( SEL_GRID_CELL_OR_HEADER );
                if ( cell$.length === 0 ) {
                    // focus not in a cell could be in a tab trap which will soon put focus in a cell
                    // remember the last edit cell across the tab trap bump
                    if ( !target$.hasClass( C_TAB_TRAP ) ) {
                        this.lastEditModeCell = null;
                    }
                    return;
                }
                cell$.addClass( C_FOCUSED );
                if ( this.lastEditModeCell === cell$[0] ) {
                    // ignore focus changes within an editing cell
                    this._clearDeactivate();
                    return;
                }
                if ( this.editMode && ( cell$.hasClass( C_GRID_CELL ) || cell$.hasClass( C_GRID_BREAK_HEADER ) ) ) {
                    // the change in focus that will happen due to activate keeps the click event from happening
                    // that is why focusInHeader set to false here
                    this.focusInHeader = false;
                    prevActive$ = $( this.lastEditModeCell );
                    this._clearDeactivate();
                    if ( prevActive$.length ) {
                        this._deactivateCell( prevActive$ );
                    }
                    this._activateCell( cell$ );
                } else {
                    this._setFocusable( event.target );
                }
            },
            focusout: function ( event ) {
                var target$ = $( event.target ),
                    cell$ = target$.closest( SEL_GRID_CELL_OR_HEADER );

                // if target is in a floating edit we want the underlying table cell
                if ( !cell$.length && this.lastEditModeCell && target$.closest( ".a-GV-floatingItem" ).length ) {
                    cell$ = $( this.lastEditModeCell );
                }
                cell$.removeClass( C_FOCUSED );
                if ( this.editMode && cell$.hasClass( C_GRID_CELL ) ) {
                    this._beginDeactivate( cell$ );
                }
            },
            contextmenu: function( event ) {
                var cell$,
                    action = this.options.contextMenuAction;

                // ignore when target is an anchor
                if ( $( event.target ).closest( "a" ).length > 0 ) {
                    return;
                }
                if ( !this.editMode && action ) {
                    // context menu only applies to grid cells and when not in edit mode
                    cell$ = $( event.target ).closest( SEL_GRID_CELL );
                    if ( cell$.length ) {
                        // if target row not selected then select it
                        if ( !cell$.parent().hasClass( C_SELECTED ) ) {
                            this._select( cell$, {}, true, false ); // force set selection
                        }
                        action( event );
                        event.preventDefault();
                    }
                }
            },
            mousemove: function( event ) {
                var row$, container, x, column, colIndex,
                    target$ = $(event.target ),
                    o = this.options,
                    cell$ = target$.closest( SEL_GRID_CELL_OR_HEADER );

                if ( cell$.length ) {
                    if ( cell$[0] !== this.lastHoverCell ) {
                        row$ = cell$.parent();
                        if ( row$[0] !== this.lastHoverRow ) {
                            this._clearHoverStates();
                            this.lastHoverRow = row$[0];
                            if ( ( row$.hasClass( C_GRID_ROW ) || row$.hasClass( C_CONTROL_BREAK ) ) &&
                                    !this.columnReordering && !this.columnWidthDragging ) {
                                this._setRowHoverState( row$ );
                            }
                        }
                        this.lastHoverCell = cell$[0];
                        // start by assuming these are hidden because each is conditionally shown
                        this.columnControls$.hide();
                        this.columnHandle$.hide();
                        if ( cell$[0].nodeName === "TH" && !cell$.hasClass( C_GRID_CELL ) && !row$.hasClass( C_CONTROL_BREAK ) ) {
                            column = this.columns[cell$.attr( ATTR_DATA_IDX )];
                            // if column sorting is enabled and over a column header and the column can be sorted
                            if ( o.columnSort && !this.columnReordering && !this.columnWidthDragging &&
                                    !cell$.hasClass( C_GRID_ROW_HEADER ) && !cell$.hasClass( C_GRID_COL_GROUP ) &&
                                    column.canSort ) {
                                cell$.append( this.columnControls$.show() );
                                this._updateColumnControls( column );
                            }
                            // if column ordering is enabled and over a column header
                            if ( o.reorderColumns && !this.columnReordering && !this.columnWidthDragging &&
                                !cell$.hasClass( C_GRID_ROW_HEADER ) ) {
                                this.columnReorderCell$ = cell$;
                                cell$.append( this.columnHandle$.show().css( "left", "" ) );
                            }
                            if ( o.resizeColumns && !this.columnReordering && !this.columnWidthDragging &&
                                    ( !cell$.hasClass( C_GRID_ROW_HEADER ) || o.rowHeader === SEQUENCE || o.rowHeader === LABEL ) ) {
                                // position the column width drag handle and associate it with its column
                                container = cell$.closest( "div" )[0];
                                x = cell$[0].offsetLeft + container.offsetLeft - container.scrollLeft;
                                if ( !this.element.hasClass( C_RTL ) ) {
                                    x += cell$.outerWidth();
                                }
                                this.columnWidthHandle$.show().css( "left", x - ( this.columnWidthHandle$.width() / 2 ) );
                                colIndex = cell$.attr( ATTR_DATA_IDX );
                                if ( colIndex ) {
                                    this.columnWidthCell$ = this._columnFromIndex( colIndex ); // do this to get both header cells - don't use cell$
                                } else {
                                    this.columnWidthCell$ = this.element.find( ".a-GV-header.a-GV-rowHeader" );
                                }
                            }
                        }
                    }
                } else {
                    if ( this.lastHoverCell !== null ) {
                        this.lastHoverCell = null;
                        this.lastHoverRow = null;
                        this._clearHoverStates();
                    }
                }
            },
            mouseleave: function() {
                if ( this.lastHoverCell !== null ) {
                    this.lastHoverCell = null;
                    this.lastHoverRow = null;
                    this._clearHoverStates();
                }
            },
            wheel: function( event ) {
                var sp, deltaY, frozen$;

                if ( this.options.hasSize ) {
                    frozen$ = $( event.target ).closest( "." + C_GRID_WRAP_FROZEN );

                    if ( frozen$.length && frozen$.parent().hasClass( C_GRID_BODY ) ) {
                        event.preventDefault();
                        sp = frozen$.parent().children( "." + C_GRID_WRAP_SCROLL )[0];
                        deltaY = event.originalEvent.deltaY || event.originalEvent.detail || ( - 1/40 * event.originalEvent.wheelDelta);
                        if ( event.originalEvent.deltaMode === undefined || event.originalEvent.deltaMode === 1 ) {
                            deltaY = deltaY * 30;
                        }
                        sp.scrollTop += deltaY;
                    }
                }
            }
        },

        _initTooltips: function( options ) {
            var ttOptions,
                self = this;

            if ( !$.ui.tooltip ) {
                debug.warn( "tooltip option ignored because missing tooltip widget dependency" );
                return;
            }
            if ( this.tooltipOptions ) {
                // tooltip widget already initialized so destroy it
                this.element.find( SEL_GRID_BODY ).filter( ":data('ui-tooltip')" ).tooltip( "destroy" );
                this.tooltipOptions = null;
            }
            if ( options ) {
                ttOptions = this.tooltipOptions = $.extend( true, {
                                show: apex.tooltipManager.defaultShowOption(),
                                tooltipClass: "a-GV-tooltip"
                                // todo position
                            }, options ); // deep copy
                ttOptions.items = "td,th"; // force this
                ttOptions._originalContent = ttOptions.content;
                ttOptions.content = function( callback ) {
                    var recId, cls, text,
                        cell$ = $( this ),
                        row$ = cell$.parent(),
                        recordMeta = null,
                        colMeta = null,
                        column = self._getColumnMetadata( cell$ );

                    recId = row$.attr( ATTR_DATA_ID );
                    if ( recId ) {
                        recordMeta = self.model.getRecordMetadata( recId );
                    }
                    if ( column && recordMeta && recordMeta.fields ) {
                        colMeta = recordMeta.fields[ column.property ];
                    }
                    if ( recordMeta && !recordMeta.deleted ) {
                        if ( colMeta && ( colMeta.error || colMeta.warning ) ) {
                            cls = colMeta.error ? "is-error" : "is-warning";
                            return "<span class='" + cls + "'>" + util.escapeHTML( colMeta.message ) + "</span>";
                        }
                        // TODO think where to show the message if there is no row header
                        if ( ( recordMeta.error || recordMeta.warning ) && cell$.hasClass( C_GRID_ROW_HEADER ) ) {
                            cls = recordMeta.error ? "is-error" : "is-warning";
                            return "<span class='" + cls + "'>" + util.escapeHTML( recordMeta.message ) + "</span>";
                        }
                        if ( recordMeta.agg && column ) {
                            text = self.options.aggregateTooltips[ recordMeta.agg + "|" + column.property ];
                            if ( text ) {
                                return "<span class='" + cls + "'>" + util.escapeHTML( text ) + "</span>";
                            }
                        }
                    }
                    if ( ttOptions._originalContent ) {
                        return ttOptions._originalContent.call( this, callback, self.model, recordMeta, colMeta, column );
                    }
                    return null;
                };
                ttOptions.open = function( event, ui ) {
                    // hoist up the error indicator class
                    if ( ui.tooltip.find( ".is-error" ).length ) {
                        ui.tooltip.addClass( "is-error" );
                    }
                    if ( ui.tooltip.find( ".is-warning" ).length ) {
                        ui.tooltip.addClass( "is-warning" );
                    }
                };
                this.element.find( SEL_GRID_BODY ).tooltip( ttOptions );
            }
        },

        _destroy: function() {
            this._tableModelViewDestroy();
            this.element.removeClass( C_GRID + " " + C_DISABLED + " " + C_RTL )
                .removeAttr( "role" )
                .removeAttr( ARIA_MULTI )
                .removeAttr( ARIA_RO )
                .empty();

            debug.info("Grid '" + this.element[0].id + "' destroyed. Model: " + this.options.modelName );

            this.tableData$ = null;
            this.tableFrozenData$ = null;

            // disconnect from the model
            this._initModel( null ); // this will cleanup change listener

            if ( this.gridStyles$ ) {
                this.gridStyles$.remove();
            }

            // remove the menu only if we created it
            if ( this.options.contextMenu && this.contextMenu$ ) {
                this.contextMenu$.remove();
            }
        },

        _setOption: function ( key, value ) {
            var o = this.options;

            debug.info("Grid '" + this.element[0].id + "' set option '" + key + "' to: ", value );

            if ( key === "contextMenu" || key === "contextMenuId" ) {
                throw new Error( "Grid " + key + " cannot be set" );
            } else if ( key === "contextMenuAction" && ( o.contextMenu || o.contextMenuId ) ) {
                throw new Error( "Grid contextMenuAction cannot be set when the contextMenu option is used" );
            } else if ( key === "editable" && !value && this.editMode ) {
                // if changing from editable to not editable make sure to leave edit mode
                this.setEditMode( false );
            } else if ( key === "multiple" ) {
                this.setSelection( $() ); // clear the selection mainly for going from multiple to single
            }

            this._super( key, value );

            this._enforceOptionConstraints();

            if ( key === "disabled" ) {
                this.element.toggleClass( C_DISABLED, value );
                if ( this.lastFocused ) {
                    if ( value ) {
                        this.lastFocused.tabIndex = -1;
                    } else {
                        this._setFocusable( this.lastFocused );
                    }
                }
                if ( value ) {
                    // when enabling do this just in case it was resized while disabled
                    this.resize();
                }
            } else if ( key === "multiple" ) {
                if ( value ) {
                    this.element.attr( ARIA_MULTI, "true" );
                } else {
                    this.element.removeAttr( ARIA_MULTI );
                }
                if ( o.rowHeader !== NONE && o.rowHeaderCheckbox ) {
                    this.refreshColumns(); // need to have the input change to checkbox or radio
                }
                this.refresh();
            } else if ( key === "modelName" ) {
                // leave edit mode if the model changes
                if ( this.editMode ) {
                    this.setEditMode( false );
                }
                this._initModel( value );
                this.refresh( false, true );
            } else if ( key === "highlights" ) {
                this._updateHighlights();
            } else if ( key === "tooltip" ) {
                this._initTooltips( value );
            } else if ( key === "rowHeader" || key === "rowHeaderCheckbox" || key === "rowHeaderLabelColumn" || key === "columnSort" ||
                        key === "resizeColumns" || key === "reorderColumns" || key === "editable" ||
                        key === "pagination" || key === "selectAll" || key === "footer" || key === "autoAddRecord") {

                if ( key === "editable" ) {
                    if ( !value ) {
                        this.element.attr( ARIA_RO, "true" );
                    } else {
                        this.element.removeAttr( ARIA_RO );
                    }
                }
                this._refreshGrid();
            }

        },

        _enforceOptionConstraints: function() {
            var o = this.options;

            if ( o.rowHeaderCheckbox && o.rowHeader === NONE ) {
                apex.debug.warn("Option rowHeaderCheckbox cannot be true when rowHeader is 'none'.");
                o.rowHeaderCheckbox = false;
            }

            if ( !o.editable || !o.allowEditMode ) {
                o.autoAddRecord = false;
            }

            if ( o.rowHeader !== NONE ) {
                if ( o.rowHeaderWidth ) {
                    this.rowHeaderWidth = o.rowHeaderWidth;
                } else {
                    if ( o.rowHeaderCheckbox ) {
                        this.rowHeaderWidth = 32;
                    } else {
                        this.rowHeaderWidth = 16;
                    }
                    if ( o.rowHeader === SEQUENCE || o.rowHeader === LABEL ) {
                        this.rowHeaderWidth += 38;
                    }
                }
            }
        },

        /**
         * Refreshes the grid with data from the model.
         * This method is rarely needed because it is called automatically when the model changes,
         * widget options change, or when pagination or column related methods are called.
         *
         * @param {boolean} [pFocus] if true put focus in the grid
         */
        refresh: function( pFocus, pModelChanged ) {
            var headerColumn, header$,
                cell$ = null,
                selection = [],
                headerRow = null,
                self = this,
                o = this.options,
                ctrl$ = this.element;

            debug.info("Grid '" + ctrl$[0].id + "' refreshed. Model: " + o.modelName );

            // this offsetParent method of checking visibility is OK because grid widget should never be position fixed
            if ( ctrl$[0].offsetParent === null ) {
                // Grid is invisible so don't bother rendering anything expect a resize or refresh later
                this.pendingRefresh = true;
                return;
            }
            this.pendingRefresh = false;

            // if there is an active cell make sure it is deactivated before re-rendering
            cell$ = ctrl$.find( SEL_ACTIVE_CELL );
            if ( cell$.length ) {
                this._deactivateCell( cell$ );
            }

            // if the last focused was in the header try to restore it after refresh
            if ( this.focusInHeader ) {
                // if the grid and columns are not being rendered the focus will not be lost
                if ( !this.columns ) {
                    // otherwise the last focused is lost but keep the focus in the header and
                    // try to stay in the same cell
                    headerRow = domIndex( $( this.lastFocused ).closest( "tr" ) );
                    headerColumn = $( this.lastFocused ).closest( "th" ).attr( ATTR_DATA_IDX );
                    this.lastFocused = null;
                }
            } else {
                this.focusInHeader = null;
                this.lastFocused = null;
            }

            // preserve current selection if not kept in model
            if ( this.tableData$ && !o.persistSelection ) {
                selection = this.getSelectedRecords();
            }

            this._refreshPagination( ctrl$.find( "." + C_GRID_WRAP_SCROLL ) );

            if ( !this.columns ) {
                this._initGrid();
            } else {
                if ( this.tableFrozenData$ ) {
                    this.tableFrozenData$.empty();
                }
                this.tableData$.empty();
            }

            this._updateStatus();
            // if there is no data then the add page of records call will do a resize so don't do it here also
            if ( !this.noData ) {
                this.resize();
            }
            this._addPageOfRecords();

            header$ = ctrl$.find( SEL_GRID_HEADER );

            ctrl$.find( SEL_GRID_ROW_SELECT ).prop( "disabled", this.noData );

            this.lastHoverCell = null;
            this.lastHoverRow = null;

            // if the last focused element was lost (or never established)
            if ( !this.lastFocused || !$( this.lastFocused ).is( SEL_VISIBLE ) || headerRow !== null ) {
                if ( headerRow !== null ) {
                    header$.find( "thead" ).each(function() {
                        cell$ = $(this ).children().eq( headerRow ).find( "th[data-idx='" + headerColumn + "']" );
                        if ( cell$.length ) {
                            self._select( cell$, {}, true, false, true );
                            return false;
                        }
                    });
                    if ( !cell$ || !cell$.length ) {
                        // the previous header column couldn't be found so pick first column
                        cell$ = header$.find( "thead" ).first()
                            .find( "tr" ).eq( headerRow ).children().first();
                        if ( cell$ ) {
                            self._select( cell$, {}, true, false, true );
                        }
                    }
                }
                // try to focus the first data cell
                if ( !this.lastFocused ) {
                    this.lastRow$ = null;
                    this.lastFocused = this._getCellFocusable( ctrl$.children( SEL_GRID_BODY )
                                        .find( "tbody tr" ).first().children().first() );
                    if ( this.lastFocused ) {
                        this.selectAnchor = this.lastFocused;
                        this.focusInHeader = false;
                    } else {
                        // if there is no data then focus a header
                        this.lastFocused = this._getCellFocusable( header$.find( "thead" ).first().find( "tr" ).last().children().first() );
                        this.focusInHeader = true;
                    }
                    if ( this.lastFocused && !o.disabled ) {
                        this._setFocusable( this.lastFocused );
                        this._setCurrentRow( this.lastFocused );
                    }
                }
            }
            // restore selection if any
            if ( selection.length > 0 ) {
                this.setSelectedRecords( selection, pFocus );
            } else {
                if ( pModelChanged ) {
                    self._trigger( EVENT_SELECTION_CHANGE, {} );
                }
                if ( pFocus ) {
                    this.focus();
                }
            }
        },

        /**
         * Give focus to the grid. Focus is given to the last element that had focus.
         */
        focus: function() {
            if ( this.lastFocused ) {
                this.lastFocused.focus();
            }
        },

        /**
         * Put focus in the cell given by pRecordId and pColumn. This is used to focus rows or cells that have
         * errors. This will scroll or change pages as needed to focus the cell. The record must be in the model.
         * The row containing the cell is selected. If pColumn is null then the first column is focused which
         * may be the row selection header.
         *
         * @param {String} pRecordId the record id of the row to go to.
         * @param {String} [pColumn] column in the record row to go to.
         */
        gotoCell: function( pRecordId, pColumn ) {
            var record, index, sp$, st,
                o = this.options,
                needRefresh = false,
                column = null,
                cell$ = null,
                row$ = this.element.find( "[data-id='" + pRecordId + "']" );

            this.gotoCellPending = null;

            if ( !this.tableData$ ) {
                return; // grid is not yet initialized
            }
            if ( pColumn ) {
                column = o.columns[0][pColumn];
                if ( !column ) {
                    debug.warn( "Warning column not found " + pColumn );
                } else {
                    // make sure column is visible
                    if ( column.hidden && column.canHide ) {
                        // show it and setup for a callback
                        column.hidden = false;
                        needRefresh = true;
                        this.refreshColumns();
                        this.gotoCellPending = { record: pRecordId, column: pColumn };
                    }
                }
            }

            if ( !row$.length ) {
                record = this.model.getRecord( pRecordId );
                if ( !record ) {
                    // if the record is not in the model then can' go to it
                    debug.warn( "Warning record not found " + pRecordId );
                    return;
                }
                index = this.model.indexOf( record );
                // todo this scrolling/pageging may not be accurate if the model has aggregate records and/or the view has control breaks
                // may need to hunt for the record but make sure no infinite loop
                this.gotoCellPending = { record: pRecordId, column: pColumn };
                if ( o.pagination.scroll ) {
                    sp$ = this.scrollParent$;
                    st = Math.floor( index / this.displayPageSize ) * this.displayPageHeight;
                    sp$.scrollTop( st ); // this will trigger a refresh as needed
                } else {
                    this.pageOffset = Math.floor( index / this.pageSize ) * this.pageSize;
                    needRefresh = true;
                }
            }

            if ( needRefresh ) {
                this.refresh( false );
            }
            if ( this.gotoCellPending ) {
                // we will be called back after the view has been refreshed
                return;
            }

            if ( pColumn && column ) {
                cell$ = this._cellFromColumn( row$, pColumn );
            }
            if ( !cell$ ) {
                cell$ = row$.children().first();
            }
            this._select( cell$, null, true, false );
        },

        /**
         * Given a column item return the cell for that column item in the current active row.
         * If pItem is not a column item or if there is no active row return null.
         * @param {DOM Element} pItem
         * @return {jQuery} cell corresponding to pItem
         */
        getActiveCellFromColumnItem: function( pItem ) {
            var row$,
                self = this,
                cell$ = null;

            if ( this.activeRecordId ) {
                row$ = this.element.find( "[data-id='" + this.activeRecordId + "']" );

                $.each( this.columnItems, function( name, value ) {
                    if ( pItem === value.item.node ) {
                        cell$ = self._cellFromColumn( row$, name );
                    }
                } );
            }
            return cell$;
        },

        /**
         * Determine if grid is in edit mode rather than navigation mode.
         *
         * @return {Boolean} true if in edit mode and false otherwise.
         */
        inEditMode: function() {
            return this.editMode;
        },

        /**
         * Set the current edit mode. Should only be used if the grid is editable.
         * Triggers modeChange event.
         *
         * @param pEditMode {boolean} if true enter edit mode if false enter navigation mode
         */
        setEditMode: function( pEditMode ) {
            var changed, cell$,
                delayDeactivate = false,
                self = this;

            function completeDeactivate() {
                self.columnItemsContainer$.off( "change.gridedit" );
                self.activeRecord = null;
                self.activeRecordId = null;
            }

            if ( !this.tableData$ ) {
                return; // not yet initialized
            }
            if ( this.options.editable ) {
                pEditMode = !!pEditMode;
                changed = pEditMode !== this.editMode;
                this.editMode = pEditMode;  // todo consider if this and the event should be delayed if deactivate delayed due to lock
                this.element.toggleClass( "a-GV--editMode", pEditMode );
                cell$ = $( this.lastFocused ).closest( SEL_GRID_CELL );
                if ( cell$.length ) {
                    if ( pEditMode ) {
                        this._activateCell( cell$ );
                    } else {
                        this.lastEditModeCell = null;
                        if ( this.deactivateDelayTimer ) {
                            clearTimeout( this.deactivateDelayTimer );
                            this.deactivateDelayTimer = null;
                        }
                        this._deactivateCell( cell$ );
                        if ( this.activeRecord ) {
                            this._deactivateRow( completeDeactivate );
                            delayDeactivate = true;
                        }
                    }
                }

                if ( changed ) {
                    // todo if changed and now editing may need to force some columns to be visible and this will cause a refresh of the grid

                    if ( pEditMode ) {
                        // handle item changes that may happen from code such as a DA.
                        this.columnItemsContainer$.on( "change.gridedit", function( event ) {
                            var i, column, prop;

                            if ( self.activeRecord ) {
                                for ( i = 0; i < self.columns.length; i++ ) {
                                    column = self.columns[i];
                                    if ( column.elementId === event.target.id ) {
                                        prop = column.property;
                                        self.setActiveRecordValue( prop );
                                        break;
                                    }
                                }
                            }

                        } );
                    } else if ( !delayDeactivate ) {
                        completeDeactivate();
                    }

                    this._trigger( EVENT_MODE_CHANGE, null, {
                        editMode: pEditMode
                    });
                }
            } else if ( pEditMode ) {
                debug.warn( "Grid is not editable." );
            }
        },

        /**
         * Return the currently selected rows as an array of jQuery objects each item in the array is a row.
         * When there are frozen columns (or even just a frozen row header) each jQuery item in the array
         * contains two tr elements. One for the frozen columns and one for the non-frozen columns.
         *
         * @example
         *    var i, rows = $( "#mygrid" ).grid( "getSelection" );
         *    for ( i = 0; i < rows.length; i++ ) {
         *        // note rows[i].length === 2 if there are frozen columns and 1 otherwise
         *        rows[i].addClass("foo"); // this adds a class to both tr elements in the case where there are frozen columns
         *        rows[i].children().each(function() {
         *            // do something with each column
         *        }
         *    }
         * @return {Array} array of jQuery row objects.
         */
        getSelection: function() {
            var sel1$, sel2$,
                result = [];

            if ( this.tableData$ ) {
                sel2$ = this.tableData$.find( SEL_SELECTED );
                if ( this.tableFrozenData$ ) {
                    sel1$ = this.tableFrozenData$.find( SEL_SELECTED );
                    sel1$.each( function( i ) {
                        result.push( $( [ this, sel2$[i] ] ) );
                    });
                } else {
                    sel2$.each(function() {
                        result.push( $( this ) );
                    });
                }
            }
            return result;
        },

        /**
         * Set the selected rows of the grid.
         *
         * @param pRows an array of jQuery row objects such as the return value of getSelection or a jQuery object
         *                  containing one or more rows (tr elements) or columns (td elements) from this grid.
         * @param {boolean} [pFocus] if true the first cell (in pRows) of the selection is given focus.
         * @param {boolean} [pNoNotify] if true the selection change event will be suppressed
         */
        setSelection: function( pRows, pFocus, pNoNotify ) {
            var i, cell,
                saveLastFocused = this.lastFocused,
                save = this.focusInHeader,
                cells = [];

            if ( !this.tableData$ ) {
                return; // grid is not yet rendered
            }

            if ( $.isArray( pRows ) ) {
                for ( i = 0; i < pRows.length; i++ ) {
                    cell = pRows[i].children( "td,th" )[0];
                    if ( cell ) {
                        cells.push( cell );
                    }
                }
            } else {
                pRows.each(function() {
                    if ( this.nodeName === "TR" ) {
                        cell = $( this ).children( "td,th" )[0];
                    } else {
                        cell = $( this ).closest( "td,th" )[0];
                    }
                    if ( cell ) {
                        cells.push( cell );
                    }
                });
            }

            pFocus = !!pFocus;
            this.focusInHeader = false;
            this._select( $( cells ), null, pFocus, false, pNoNotify );
            if ( !pFocus ) {
                this.focusInHeader = save;
                // if not setting focus and focus was in the header put it back where it was
                if ( save ) {
                    this._setFocusable( saveLastFocused );
                    this._setCurrentRow( saveLastFocused );
                }
            }
        },

        /**
         * Select all rows. This has no effect if the multiple or selectAll options are not true.
         * This only applies to the current page xxx or what has been rendered so far????
         * xxx if pagination is not scroll then it applies to the whole page otherwise it applies to what is in the scroll viewport
         * @param {boolean} [pFocus] if true the first cell (in pRows) of the selection is given focus.
         * @param {boolean} [pNoNotify] if true the selection change event will be suppressed
         */
        selectAll: function( pFocus, pNoNotify ) {
            var cells = [],
                table$ = this.tableFrozenData$ || this.tableData$;

            if ( table$ && this.options.multiple && this.options.selectAll ) {
                table$.children().each(function() {
                    cells.push($( this ).children()[0] );
                });
                this.setSelection( $( cells ), pFocus, pNoNotify );
            }
        },

        /**
         * Given an array of jQuery row objects return the underlying data model records.
         * The return value of getSelection is an appropriate value for pRows. However it
         * need not contain tr elements from both the frozen and un-frozen areas.
         *
         * @param pRows array of jQuery objects representing grid rows.
         * @return {Array} array of records from the model corresponding to the grid rows
         */
        getRecords: function( pRows ) {
            var i, id, value,
                values = [];

            for ( i = 0; i < pRows.length; i++ ) {
                id = pRows[i].first().attr( ATTR_DATA_ID );
                if ( id !== undefined ) {
                    value = this.model.getRecord( id );
                    if ( value ) {
                        values.push( value );
                    }
                }
            }

            return values;
        },

        /**
         * Return the underlying data model records corresponding to the current selection.
         *
         * Note: If option persistSelection is true then the selected records could span multiple pages
         * and getSelectedRecords returns a different selection from getSelection which can only return
         * elements from the current page. To get just the records that correspond to getSelection when
         * this option is true use: $(<grid>).grid("getRecords", $(<grid>).grid("getSelection"))
         *
         * @return {Array} array of records from the model corresponding to the selected grid rows
         */
        getSelectedRecords: function() {
            if ( this.options.persistSelection ) {
                // if selection is persisted in the model the ask the model for the count
                // in this case the selection could span multiple pages
                return this.model.getSelectedRecords();
            } // else
            return this.getRecords( this.getSelection() );
        },

        /**
         * Select the grid rows that correspond to the given data model records.
         * @param pRecords {Array} array of data model records to select
         * @param pFocus {Boolean} if true the first cell of the selection is given focus
         */
        setSelectedRecords: function( pRecords, pFocus ) {
            var i,
                rows = [],
                keys = {},
                table$ = this.tableFrozenData$ || this.tableData$;

            if ( !table$ ) {
                return;
            }
            for ( i = 0; i < pRecords.length; i++ ) {
                keys[this.model.getRecordId( pRecords[i] )] = true;
            }
            table$.children().each(function() {
                var row$ = $( this ),
                    id = row$.attr( ATTR_DATA_ID );

                if ( keys[id] ) {
                    rows.push( row$ );
                }
            });
            this.setSelection( rows, pFocus );
        },

        /**
         * Call this method anytime the container that the grid is in changes its size.
         * For performance it is best to call this after the size has changed not continuously while it is changing.
         */
        resize: function() {
            var i, w, h, fw, headerHeight, col,
                ctrl$ = this.element,
                o = this.options,
                ctrlH = ctrl$.height(),
                ctrlW = ctrl$.width(),
                header$ = ctrl$.children( SEL_GRID_HEADER ),
                body$ = ctrl$.children( SEL_GRID_BODY );

            if ( ctrl$[0].offsetParent === null ) {
                // Grid is invisible so nothing to resize. Expect a resize or refresh later when made visible
                return;
            }
            if ( !body$.length || this.pendingRefresh ) {
                // grid was never initialized probably because it was initially invisible
                // or was refreshed while invisible. So do that now
                this.refresh();
                return; // because refresh calls resize
            }

            headerHeight = this.tableData$.prev().height();
            ctrl$.find( SEL_GRID_BODY + " .a-GV-table" ).css( "margin-top",
                -( headerHeight + parseInt( this.tableHeader$.find( "th" ).first().css( "border-bottom-width" ), 10) )  );

            w = ctrlW;
            if ( this.tableFrozenData$ ) {
                // need to set widths so the frozen and normal tables are side by side
                fw = this.tableFrozenData$.closest( ".a-GV-w-frozen" ).width();
                ctrl$.find( ".a-GV-w-frozen" ).width( fw ).toggleClass( "a-GV-w-frozen--showLabels", fw > FROZEN_LABEL_WIDTH );
                // when showing load more button in the frozen area don't show it in scrolling area
                this.tableData$.parent().parent().find( ".a-GV-loadMoreButton" ).css( "visibility", fw > FROZEN_LABEL_WIDTH ? "hidden" : "visible" );
                w = w - fw;
            }
            ctrl$.find( ".a-GV-w-scroll,.a-GV-w-hdr" ).width( w );

            // if size shrinks reset widths
            if ( this.tableData$.parent().width() > w ) {
                for ( i = 0; i < this.columns.length; i++ ) {
                    col = this.columns[i];
                    if ( col.width ) {
                        col.curWidth = col.width;
                    }
                }
            }
            this._calculateColumnWidths( null, true );

            // refresh any sticky widgets that may be confused by the resize
            if ( $.apex.stickyWidget ) {
                ctrl$.find(".js-stickyWidget-toggle").stickyWidget("refresh");
            }

            if ( o.hasSize ) {
                h = ctrlH - header$.height() - this._footerHeight();
                body$.children( "." + C_GRID_WRAP_SCROLL ).addBack().height( h );
                this._adjustSizeForScrollBars( h, w, headerHeight );
            }

            this._initPageSize();
            // todo if pagination is page and rowsPerPage is null (auto) consider refreshing so view reflects the correct number of rows. Beware infinite recursion
        },

        //
        // Control break methods
        //


        expandControlBreak: function( pRows ) {
            this._expandCollapseControlBreak( pRows, true );
        },

        collapseControlBreak: function( pRows ) {
            this._expandCollapseControlBreak( pRows, false );
        },

        //
        // Column methods
        //

        /**
         * Get the column definitions in order.
         * @return {Array} Array of column definition objects.
         */
        getColumns: function() {
            return this.columns;
        },

        /**
         * Let the grid know that column metadata has changed so that the next time it is refreshed columns will
         * be rendered. Call this method after any column metadata has changed external to this widget.
         * Refresh must be called after this but typically this happens due to the model refresh notification.
         */
        refreshColumns: function() {
            this.columns = null; // clear out existing column configuration so that refresh will recreate
        },

        /**
         * Sets the width of the given column.
         * Triggers columnResize event.
         *
         * @param pColumn column name or column definition object to set the width on.
         * @param pWidth {number} width in pixels to set the column too.
         */
        setColumnWidth: function( pColumn, pWidth ) {
            var colInfo = this._updateColumnWidth( pColumn, pWidth, true );

            this._trigger( EVENT_COLUMN_RESIZE, null, {
                header$: colInfo[1], // col$
                column: colInfo[0], // col
                width: pWidth
            });
        },

        /**
         * Hide the given column. Also calls refreshColumns and will render the whole grid.
         * @param pColumn column name or column definition object to hide.
         */
        hideColumn: function( pColumn ) {
            var columnName = this._columnName( pColumn );
            this.options.columns[0][columnName].hidden = true;
            this._refreshGrid();
        },

        /**
         * Show the given column. Also calls refreshColumns and will render the whole grid.
         * @param pColumn column name or column definition object to show.
         */
        showColumn: function( pColumn ) {
            var columnName = this._columnName( pColumn );
            this.options.columns[0][columnName].hidden = false;
            this._refreshGrid();
        },

        /**
         * Freeze the given column. Also calls refreshColumns and will render the whole grid.
         * @param pColumn column name or column definition object to freeze.
         */
        freezeColumn: function( pColumn ) {
            var i, col,
                cols = this.columns,
                columnName = this._columnName( pColumn );

            for ( i = 0; i < cols.length; i++ ) {
                col = cols[i];
                // can never hide last column
                if ( i < cols.length - 1 ) {
                    col.frozen = true;
                }
                if ( col.property === columnName ) {
                    break;
                }
            }
            this._refreshGrid();
        },

        /**
         * Unfreeze the given column. Also calls refreshColumns and will render the whole grid.
         * @param pColumn column name or column definition object to unfreeze.
         */
        unfreezeColumn: function( pColumn ) {
            var i, col,
                columnName = this._columnName( pColumn );

            for ( i = this.columns.length - 1; i >= 0; i-- ) {
                col = this.columns[i];

                col.frozen = false;
                if ( col.property === columnName ) {
                    break;
                }
            }
            this._refreshGrid();
        },

        /**
         * Move the given column to the new position. Also calls refreshColumns and will render the whole grid.
         * Triggers columnReorder event.
         *
         * @param pColumn column name or column definition object to move.
         * @param pNewPosition {number} index into the array returned by getColumns where the column will be moved to.
         */
        moveColumn: function( pColumn, pNewPosition ) {
            var ui = this._moveColumn( pColumn, pNewPosition );

            this._refreshGrid();
            ui.header$ = this._columnFromIndex( pNewPosition );
            this._trigger( EVENT_COLUMN_REORDER, null, ui );
        },

        /**
         * Move a column group from one position to another. This moves all the columns associated with
         * the group to the new position.
         * Triggers columnReorder event for each column moved.
         *
         * @param pLevel {number} the level (starting with zero as the top most level) of the group
         * @param pOriginalPosition {number} original position of the column group
         * @param pNewPosition {number} new position of the column group
         */
        moveColumnGroup: function( pLevel, pOriginalPosition, pNewPosition ) {
            var i, colGroup, groups, start, end, dest, ui,
                uis = [];

            groups = this.columnGroups[pLevel];
            if ( !groups ) {
                throw new Error( "Invalid column group level: " + pLevel );
            }
            start = -1;
            for ( i = 0; i < groups.length; i++ ) {
                colGroup = groups[i];
                if ( i === pOriginalPosition ) {
                    start = colGroup.start;
                    end = colGroup.end;
                }
                if ( i === pNewPosition ) {
                    if ( start === -1 ) {
                        // the new position is before the original position
                        dest = colGroup.start;
                    } else {
                        // the new position is after the original position
                        dest = colGroup.end;
                    }
                }
            }
            // move each column
            if ( pNewPosition > pOriginalPosition ) {
                for ( i = start; i <= end; i++ ) {
                    ui = this._moveColumn( this.columns[start], dest );
                    ui.newPosition = dest - ( end - i );
                    ui.oldPosition = i;
                    uis.push( ui );
                    this._updateColumns();
                }
            } else {
                for ( i = end; i >= start; i-- ) {
                    ui = this._moveColumn( this.columns[end], dest );
                    ui.newPosition = dest + ( i - start );
                    ui.oldPosition = i;
                    uis.push( ui );
                    this._updateColumns();
                }
            }

            this._refreshGrid();
            for ( i = 0; i < uis.length; i++ ) {
                ui = uis[i];
                ui.header$ = this._columnFromIndex( ui.newPosition );
                this._trigger( EVENT_COLUMN_REORDER, null, ui );
            }
        },

        //
        // Internal methods
        //

        _refreshGrid: function() {
            var cell$;

            // if there is an active cell make sure it is deactivated before re-rendering
            // must do this before the column info is destroyed
            cell$ = this.element.find( SEL_ACTIVE_CELL );
            if ( cell$.length ) {
                this._deactivateCell( cell$ );
            }

            this.refreshColumns();
            this.refresh();
        },

        _cellFromColumn: function( row$, columnName ) {
            var i, column, col,
                cols = this.columns,
                o = this.options,
                cell$ = null,
                colIndex = 0;

            column = o.columns[0][columnName];
            if ( !column ) {
                debug.warn( "Warning column not found " + columnName );
                return null;
            } // else find the cell$

            if ( o.rowHeader !== NONE ) {
                colIndex = 1;
            }
            for ( i = 0; i < cols.length; i++ ) {
                col = cols[i];
                if ( col.hidden ) {
                    continue;
                }
                if ( col.property === columnName ) {
                    cell$ = row$.children().eq( colIndex );
                    break;
                }
                colIndex += 1;
            }
            if ( i >= cols.length ) {
                debug.warn( "Warning column is hidden " + columnName );
            }
            return cell$;
        },

        _columnFromIndex: function( columnIndex ) {
            // Note that because of the hidden but accessible column header on the main data table this returns
            // two elements!
            return this.element.find( ".a-GV-header[data-idx=" + columnIndex + "]" );
        },

        _columnName: function( column ) {
            var columnName;

            if ( typeof column === "object" ) {
                columnName = column.property;
            } else {
                columnName = column;
            }
            if ( !this.options.columns[0].hasOwnProperty( columnName ) ) {
                throw new Error( "Unknown column: " + columnName );
            }
            return columnName;
        },

        // returns notification data except for header$ which can't be known until after the grid is refreshed
        _moveColumn: function( pColumn, pNewPosition ) {
            var i, col, originalPos, seq,
                cols = this.columns,
                columnName = this._columnName( pColumn );

            originalPos = -1;
            for ( i = 0; i < cols.length; i++ ) {
                col = cols[i];
                if ( col.property === columnName ) {
                    originalPos = i;
                    break;
                }
            }

            seq = cols[pNewPosition].seq;
            if ( originalPos === pNewPosition ) {
                return; // nothing to do
            } else if ( originalPos < pNewPosition ) {
                // moving up - adjust column sequences
                for ( i = pNewPosition; i > originalPos; i-- ) {
                    cols[i].seq = cols[i - 1].seq;
                }
            } else {
                // moving down - adjust column sequences
                for ( i = pNewPosition; i < originalPos; i++ ) {
                    cols[i].seq = cols[i + 1].seq;
                }
            }
            cols[originalPos].frozen = ( pNewPosition <= this.lastFrozenColIndex );
            cols[originalPos].seq = seq;
            return {
                column: col,
                newPosition: pNewPosition,
                oldPosition: originalPos
            };
        },

        _updateColumns: function() {
            var i, lastIndex, col, colName, lastGroupRow, curGroupRow, curGroupIndex, curGroupName, curGroup, group,
                colIndex, noHeading, children,
                o = this.options,
                optColumns = o.columns[0];

            this.columns = [];
            this.columnGroups = [];
            this.breakColumns = [];
            this.fieldToColumnIndex = {};

            // put the column definitions in an array
            for ( colName in optColumns) {
                if ( optColumns.hasOwnProperty( colName )) {
                    optColumns[colName].property = colName;
                    // make sure there is a sequence value to sort by
                    optColumns[colName].seq = optColumns[colName].seq || "0";
                    this.columns.push( optColumns[colName] );
                }
            }
            // and sort them by the seq[uence] property
            this.columns.sort(function(a, b) {
                return a.seq - b.seq;
            });

            // now count the frozen ones
            this.lastFrozenColIndex = -1;
            for ( i = 0; i < this.columns.length; i++ ) {
                col = this.columns[i];
                if ( col.hidden ) {
                    continue;
                }
                if ( col.frozen ) {
                    this.lastFrozenColIndex = i;
                } else {
                    break;
                }
            }

            // gather any control break columns and index by field/property name
            colIndex = o.rowHeader !== NONE ? 1 : 0; // account for row header which is a column in the DOM but not a column in the model
            for ( i = 0; i < this.columns.length; i++ ) {
                col = this.columns[i];
                if ( col.controlBreakIndex > 0 ) {
                    this.breakColumns.push( col );
                }
                if ( !col.hidden ) {
                    this.fieldToColumnIndex[col.property] = colIndex;
                    colIndex += 1;
                }
            }
            this.breakColumns.sort(function(a, b) {
                return a.controlBreakIndex - b.controlBreakIndex;
            });

            // add column groups if there are any
            curGroupRow = null;
            curGroupName = null;
            curGroup = null;
            noHeading = true;
            colIndex = 0;
            curGroupIndex = 0;
            lastIndex = 0;
            children = [];
            for ( i = 0; i < this.columns.length; i++ ) {
                col = this.columns[i];
                if ( col.hidden ) {
                    continue;
                }
                if ( curGroupRow === null ) {
                    curGroupRow = [];
                    this.columnGroups.push( curGroupRow );
                }
                group = effectiveGroupName( col );
                if ( group !== curGroupName || i === this.lastFrozenColIndex + 1 ) {
                    curGroupName = group;
                    if ( curGroup ) {
                        curGroup.span = colIndex - curGroupIndex;
                        curGroup.end = lastIndex;
                        curGroup.children = children;
                        children = [];
                    }
                    curGroupIndex = colIndex;
                    if ( group ) {
                        curGroup = $.extend({}, o.columnGroups[ group ]);
                        if ( curGroup ) {
                            if ( curGroup.heading ) {
                                noHeading = false;
                            }
                            curGroupRow.push( curGroup );
                        } else {
                            throw new Error( "Unknown column group name: " + group );
                        }
                    } else {
                        curGroup = {};
                        curGroupRow.push( curGroup );
                    }
                    curGroup.start = i;
                    if ( i <= this.lastFrozenColIndex ) {
                        curGroup.frozen = true;
                    }
                }
                children.push(col);
                colIndex += 1;
                lastIndex = i;
                if ( curGroup === null ) {
                    curGroup = {
                        start: i
                    };
                    if ( this.lastFrozenColIndex >= 0 ) {
                        curGroup.frozen = true;
                    }
                    curGroupRow.push( curGroup );
                }
            }
            if ( curGroup ) {
                curGroup.span = colIndex - curGroupIndex;
                curGroup.end = lastIndex;
                curGroup.children = children;
            }
            if ( noHeading ) {
                this.columnGroups = [];
                curGroupRow = null;
            }

            // column groups can have parent groups so add those
            while ( curGroupRow ) {
                lastGroupRow = curGroupRow;
                curGroupRow = null;
                curGroupName = null;
                curGroup = null;
                noHeading = true;
                colIndex = 0;
                children = [];
                for ( i = 0; i < lastGroupRow.length; i++ ) {
                    col = lastGroupRow[i];
                    if ( ( col.parentGroupName || null ) !== curGroupName || col.frozen !== ( curGroup ? curGroup.frozen: false ) ) {
                        curGroupName = col.parentGroupName || null;
                        if ( curGroupRow === null ) {
                            curGroupRow = [];
                            this.columnGroups.unshift( curGroupRow );
                            if ( colIndex > 0 ) {
                                curGroup = {};
                                if ( col.frozen ) {
                                    curGroup.frozen = true;
                                }
                                curGroupRow.push( curGroup );
                            }
                        }
                        if ( curGroup ) {
                            curGroup.span = colIndex - curGroupIndex;
                            curGroup.children = children;
                            children = [];
                        }

                        curGroupIndex = colIndex;
                        if ( col.parentGroupName ) {
                            curGroup = $.extend({}, o.columnGroups[col.parentGroupName]);
                            if ( curGroup ) {
                                if ( curGroup.heading ) {
                                    noHeading = false;
                                }
                                curGroupRow.push( curGroup );
                            } else {
                                throw new Error( "Unknown column group name: " + col.parentGroupName );
                            }
                        } else {
                            curGroup = {};
                            curGroupRow.push( curGroup );
                        }
                        curGroup.start = col.start;
                        if ( col.frozen ) {
                            curGroup.frozen = true;
                        }
                    }
                    children.push(col);
                    curGroup.end = col.end;
                    colIndex += col.span;
                }
                if ( curGroup ) {
                    curGroup.span = colIndex - curGroupIndex;
                    curGroup.end = col.end;
                    curGroup.children = children;
                }
                if ( noHeading ) {
                    this.columnGroups.shift();
                    curGroupRow = null;
                }
            }
        },

        _initGrid: function() {
            var hdr$, main$, frzn$, body$, row$, traps$, floatingTrap, endTrap, dataCols$, dataCellCols$, swOpt,
                self = this,
                scrolling = false,
                o = this.options,
                ctrl$ = this.element,
                checkboxClass = "a-GV-table--checkbox";

            debug.info("Grid render and initialize.");

            this._updateColumns();
            if ( !this.columnItems ) {
                this._initColumnItems( this.columns ); // just do this once
            }
            this._renderGrid();
            // after the rendering any anchors or buttons (any tab stop) in column headings need to have the tabstop disabled
            ctrl$.find( SEL_TABBABLE ).prop( "tabIndex", -1 );

            if ( o.tooltip ) {
                this._initTooltips( o.tooltip );
            }

            if ( o.rowHeader !== NONE && o.rowHeaderCheckbox ) {
                this.tableFrozenHeader$.parent().addClass( checkboxClass );
                this.tableFrozenData$.parent().addClass( checkboxClass );
            }

            // Some columns have an explicit width and other do not. For columns without a width use what the initial
            // auto width table assigns and apply that to the column
            body$ = ctrl$.children( SEL_GRID_BODY );
            dataCols$ = body$.find( "col" );
            dataCellCols$ = body$.find( "thead" ).find( ".a-GV-header" );

            dataCols$.each(function( i ) {
                var col$ = $( this ),
                    col = self.columns[ col$.attr(ATTR_DATA_IDX) ];

                if ( col ) {
                    if ( col.width ) {
                        col.curWidth = col.width;
                    } else {
                        // xxx the problem with doing this here before the frozen/non-frozen widths are set and before data
                        // rendered is that the browsers default width is based on the full width excluding the vert scroll bar
                        // and seems to not take into consideration the frozen width.
                        col.curWidth = dataCellCols$.eq( i ).outerWidth();
                    }
                    if ( col.curWidth < MIN_COL_WIDTH ) {
                        col.curWidth = MIN_COL_WIDTH;
                    }
                }
            });
            // The column widths will be calculated and set when the widget is resized which happens soon

            // coordinate the scrolling of the various areas
            hdr$ = ctrl$.find( "." + C_GRID_WRAP_HEADER );
            main$ = ctrl$.find( "." + C_GRID_WRAP_SCROLL );
            frzn$ = ctrl$.find( "." + C_GRID_WRAP_FROZEN ).last();
            main$.scroll(function() {
                scrolling = true;
                hdr$.scrollLeft( this.scrollLeft );
                frzn$.scrollTop( this.scrollTop );
                self.tableData$.parent().parent().find( ".a-GV-loadMoreButton" ).css( "left", this.scrollLeft );

                setTimeout(function() {
                    scrolling = false;
                }, 60);
            });
            hdr$.scroll(function() {
                if ( !scrolling ) {
                    main$.scrollLeft( this.scrollLeft );
                }
                scrolling = false;
            });
            frzn$.scroll(function() {
                // xxx when using keyboard in the frozen area auto repeat up/down the scroll offset can get out of sync!
                if ( !scrolling ) {
                    main$.scrollTop( this.scrollTop );
                }
                scrolling = false;
            });

            traps$ = this.element.find( SEL_TAB_TRAP );
            this.tabTraps$ = traps$;
            // The expected visual tabbing order between the frozen and non frozen cells does not match the DOM tabbing
            // order so these invisible tab traps capture the focus and redirect focus to the correct place.
            // There are either 2 or 6 traps. If there are frozen columns then 6 else 2. The floating edit traps are
            // always last.
            if ( traps$.length === 2 ) {
                floatingTrap = 0; // this is the case where autoAddRecord is false and there are no frozen columns
            } else {
                if ( traps$.length === 3 ) {
                    floatingTrap = 1; // this is the case where autoAddRecord is true and there are no frozen columns
                    endTrap = 0;
                } else {
                    // else there are frozen columns so 6 traps
                    floatingTrap = 4;
                    endTrap = 3;
                    traps$.eq( 0 ).on( "focus", function() {
                        self.tableData$.find( SEL_TABBABLE ).first().focus();
                    });
                    traps$.eq( 1 ).on( "focus", function() {
                        self.tableData$.find( SEL_TABBABLE ).last().focus();
                    });
                    traps$.eq( 2 ).on( "focus", function() {
                        self.tableFrozenData$.find( SEL_TABBABLE ).first().focus();
                    });
                }
                traps$.eq( endTrap ).on( "focus", function() {
                    if ( $(this ).hasClass( "js-autoAdd" ) ) {
                        self._autoAddRecord( self.activeRecord );
                    } else {
                        self.tableFrozenData$.find( SEL_TABBABLE ).last().focus();
                    }
                });
            }
            traps$.eq( floatingTrap ).on( "focus", function() {
                var allTabs$, gridTab$,
                    tableData$ = $( self.lastEditModeCell ).closest( "tbody" ),
                    ts$ = tableData$.find( SEL_TABBABLE ).add( self.lastEditModeCell ),
                    index = ts$.index( self.lastEditModeCell );

                if ( index > 0 ) {
                    ts$.first().focus();
                } else if ( self.tableFrozenData$ && traps$.eq( 0 ).prop( "tabIndex" ) === 0 ) {
                    // need to focus in the other table
                    tableData$ = tableData$[0] === self.tableFrozenData$[0] ? self.tableData$ : self.tableFrozenData$;
                    tableData$.find( SEL_TABBABLE ).first().focus();
                } else {
                    // in this case need to focus the tabbable element just before the grid
                    allTabs$ = $( SEL_TABBABLE );
                    gridTab$ = self.element.find( SEL_TABBABLE ).first();
                    index = allTabs$.index( gridTab$ );
                    allTabs$.eq( index - 1 ).focus();
                }
            });
            traps$.eq( floatingTrap + 1 ).on( "focus", function() {
                var tableData$ = $( self.lastEditModeCell ).closest( "tbody" ),
                    ts$ = tableData$.find( SEL_TABBABLE ).add( self.lastEditModeCell ),
                    index = ts$.index( self.lastEditModeCell );

                if ( $(this ).hasClass( "js-autoAdd" ) ) {
                    self._autoAddRecord( self.activeRecord );
                } else if ( index < ts$.length - 1 ) {
                    ts$.last().focus();
                } else if ( self.tableFrozenData$ ) {
                    // need to focus in the other table
                    tableData$ = tableData$[0] === self.tableFrozenData$[0] ? self.tableData$ : self.tableFrozenData$;
                    tableData$.find( SEL_TABBABLE ).last().focus();
                }
            });
            // traps are not used until a cell is activated
            traps$.prop( "tabIndex", -1 );

            if ( o.reorderColumns ) {
                this._initColumnReorder();
            }

            if ( o.resizeColumns ) {
                this._initColumnResize();
            }

            this.columnControls$.on( "click", "button", function( event ) {
                var target$ = $( event.target );

                self._sortChange( event, target$.closest( SEL_GRID_COL_HEADER ),
                    target$.closest( "button" ).hasClass("js-asc") ? "asc" : "desc");
            } );

            function expandCollapse() {
                var cell$ = self.element.find( SEL_ACTIVE_CELL );

                if ( self.floatingItem$.hasClass( C_EXPANDED )) {
                    self._collapseFloatingItem( cell$ );
                } else {
                    self._expandFloatingItem( cell$ );
                }
            }

            this.floatingItem$.draggable( {
                handle: ".a-GV-cellMoveHandle",
                containment: "document"
            } ).on( "keydown", function( event ) {
                if ( event.ctrlKey && event.which === 117 ) {// Ctrl+F6
                    expandCollapse();
                    return false;
                }
            }).children( ".a-GV-expandCollapse" ).click( expandCollapse );

            this._initPaginationFooter( main$ );

            // if the region has no defined height and the stickyWidget is available then stick the
            // column headers to the top of the page and/or the footer to the bottom of the page
            if ( !o.hasSize && o.stickyTop && $.apex.stickyWidget ) {
                swOpt = {
                    stickToEnd: true,
                    toggleWidth: true, // xxx needed?
                    bottom: function() {
                        return ctrl$.offset().top + ctrl$.outerHeight() - self._footerHeight();
                    }
                };
                if ( $.isFunction( o.stickyTop ) ) {
                    swOpt.top = o.stickyTop;
                }
                ctrl$.find( SEL_GRID_HEADER ).stickyWidget( swOpt );
            }

            if ( this.breakColumns.length > 0 ) {
                ctrl$.find( SEL_GRID_BODY ).on( "click", ".js-toggleBreak", function() {
                    row$ = $( this ).closest( "." + C_CONTROL_BREAK );
                    self._expandCollapseControlBreak( [row$], !row$.hasClass( C_EXPANDED ) );
                });
            }
        },

        _initColumnReorder: function() {
            var mTop, hTop, columnOffsets, row$, lastColumnIndex, headerPos, zIndex, position, offsetAdjust, markerAdjust,
                oldPos, helperWidth, scrollOffset, handleMargin,
                rtl = this.element.hasClass( C_RTL ),
                header$ = this.element.find( SEL_GRID_HEADER ),
                self = this;

            header$.append( this.columnMoveMarker$ );
            handleMargin = parseInt( self.columnHandle$.css( "margin-left" ), 10 );
            if ( rtl ) {
                handleMargin = -handleMargin;
            }

            // enable the column and column group headings to be draggable
            self.columnHandle$.draggable({
                appendTo: this.element,
                containment: header$,
                cursor: "move",
                cursorAt: rtl ? { right: -2 * handleMargin } : { left: 0 },
                axis: "x",
                delay: 100,
                distance: 4,
                helper: function() {
                    var helper$ = $( "<div class='a-GV-header-dragHelper'></div>" ),
                        cell$ = self.columnReorderCell$;
                    // add copy of handel and label to drag helper
                    helper$.append( cell$.find( ".a-GV-headerLabel,." + C_GRID_COL_HANDLE ).clone() );
                    helper$.width( cell$.outerWidth() ); // height is set with css
                    return helper$;
                },
                start: function( event, ui ) {
                    self.columnControls$.hide();
                    // send this event in the hopes that any popup related to the column header will be closed
                    self._trigger( EVENT_CANCEL_COLUMN_HEADER, event, {});
                    // figuring out how to position the helper and new column marker is tricky because of
                    // the different ways the header is positioned
                    mTop = self.columnReorderCell$.parent().offset().top - self.element.offset().top;
                    hTop = mTop + self.element[0].offsetTop;
                    headerPos = header$.css( "position" );
                    position = "absolute";
                    zIndex = 2;
                    offsetAdjust = -header$.offset().left;
                    if ( headerPos !== "relative" ) {
                        if ( headerPos === "fixed" ) {
                            offsetAdjust = 0;
                            mTop = hTop = self.columnReorderCell$.parent()[0].offsetTop + header$[0].offsetTop;
                            position = "fixed";
                            ui.helper.css( "margin-left", self.element.offset().left + "px" );
                        }
                        zIndex = parseInt( header$.css("z-index"), 10 ) + 1;
                    }
                    markerAdjust = -self.columnMoveMarker$.width() / 2;
                    ui.helper.css( {
                        "top": hTop + "px",
                        "position": position,
                        "z-index": zIndex
                    } );
                    helperWidth = ui.helper.outerWidth();
                    self.columnMoveMarker$.show().css( {
                        "top": mTop + "px",
                        "left": ( self.columnReorderCell$.offset().left + offsetAdjust + ( rtl ? self.columnReorderCell$.width() : 0 ) ) + "px",
                        "position": position,
                        "z-index": zIndex
                    } );
                    // hide the cell content
                    self.columnReorderCell$.find( ".a-GV-headerLabel,.a-GV-header-sort" ).hide();
                    self.columnHandle$.hide();
                    self.columnWidthHandle$.hide();
                    escapeOn(function() {
                        self.columnReorderCancel = true;
                    });

                    // remember which column is moving
                    oldPos = parseInt( self.columnReorderCell$.attr( ATTR_DATA_IDX ), 10 );

                    // setup to scroll as needed when drag near boundaries
                    self.hdrScrollParent = self.element.find( ".a-GV-w-hdr" )[0];
                    scrollOffset = self.hdrScrollParent.scrollLeft;

                    // gather column offsets
                    columnOffsets = [];
                    lastColumnIndex = null;
                    row$ = self.columnReorderCell$.parent();
                    if ( self.tableFrozenHeader$ ) {
                        row$ = row$.add( self._getOtherTable( row$ ).children().eq( domIndex( row$ ) ) );
                    }
                    row$.children().each( function() {
                        var col$ = $( this ),
                            frozen = true,
                            offset = col$.offset().left + offsetAdjust;

                        if ( col$.parent().parent().parent().parent()[0] === self.hdrScrollParent ) {
                            offset += scrollOffset;
                            frozen = false;
                        }
                        columnOffsets.push( {
                            colIdx: parseInt( col$.attr( ATTR_DATA_IDX ), 10),
                            start:  offset,
                            end: offset + col$.outerWidth(),
                            frozen: frozen
                        } );
                    } );

                    self.columnReordering = true;
                    self.columnReorderCancel = false;
                },
                drag: function( event, ui ) {
                    var sl = self.hdrScrollParent.scrollLeft,
                        origLeft = parseInt( self.columnMoveMarker$.css( "left" ), 10);

                    function updateMarker() {
                        var i, col, newOffset,
                            moved = false,
                            // during drag we use event.pageX - handleMargin rather than ui.offset.left so that when the
                            // helper bumps up against the container the point to check against the columns changes
                            // this allows for example a wide column to be moved to the very end.
                            cur = ( event.pageX - handleMargin ) + offsetAdjust + self.hdrScrollParent.scrollLeft;

                        // check the positions and move the insertion marker
                        for ( i = 0; i < columnOffsets.length; i++ ) {
                            col = columnOffsets[i];
                            if ( cur >= col.start && cur < col.end ) {
                                if ( i !== lastColumnIndex ) {
                                    lastColumnIndex = i;
                                    if ( rtl ) {
                                        newOffset = ( col.colIdx < oldPos ? col.end : col.start ) + markerAdjust;
                                    } else {
                                        newOffset = ( col.colIdx < oldPos ? col.start : col.end ) + markerAdjust;
                                    }
                                    if ( !col.frozen ) {
                                        newOffset -= self.hdrScrollParent.scrollLeft;
                                    }
                                    origLeft = newOffset;
                                    sl = self.hdrScrollParent.scrollLeft;
                                    self.columnMoveMarker$.css( "left", newOffset + "px" );
                                    moved = true;
                                }
                                break;
                            }
                        }
                        // if the header has scrolled need to move the marker to match
                        if ( !moved && sl !== self.hdrScrollParent.scrollLeft ) {
                            newOffset = origLeft - ( self.hdrScrollParent.scrollLeft - sl );
                            self.columnMoveMarker$.css( "left", newOffset + "px" );
                        }
                    }

                    if ( self.columnReorderCancel ) {
                        event.preventDefault();
                        return;
                    }

                    updateMarker();
                    // Need to manually scroll. Draggable scrolling doesn't work because the header
                    // is overflow hidden and scrolls because of the table body scroll bar
                    // start or stop scrolling as needed. The actual scrolling happens from a timer
                    if ( self._headerScrollCheck( event.pageX - handleMargin, helperWidth ) ) {
                        if ( !self.scrollTimerId ) {
                            self._headerScrollStart( event.pageX - handleMargin, helperWidth, updateMarker );
                        }
                    } else {
                        if ( self.scrollTimerId ) {
                            self._headerScrollStop();
                        }
                    }
                },
                stop: function( event, ui ) {
                    var newPos = null;

                    // check the positions and move the insertion marker and do the move
                    if ( lastColumnIndex !== null ) {
                        newPos = columnOffsets[lastColumnIndex].colIdx;
                    }
                    if ( isNaN( newPos ) && lastColumnIndex + 1 < columnOffsets.length ) {
                        // must be over the row selector header skip to the next column
                        newPos = columnOffsets[lastColumnIndex + 1].colIdx;
                    }
                    if ( isNaN( newPos ) ) {
                        return;
                    }
                    self.columnReordering = false;
                    self.columnReorderCell$.find( ".a-GV-headerLabel,.a-GV-header-sort" ).show();
                    if ( newPos === oldPos ) {
                        self.columnHandle$.show();
                        self.columnWidthHandle$.show();
                        self.columnControls$.show();
                    }
                    self.columnMoveMarker$.hide();
                    self.lastHoverCell = null; // force recalculating what to show on next mouse move
                    escapeOff();
                    if ( self.scrollTimerId ) {
                        self._headerScrollStop();
                    }
                    if ( self.columnReorderCancel || newPos === null || newPos === oldPos ) {
                        return;
                    }

                    if ( self.columnReorderCell$.hasClass( C_GRID_COL_GROUP ) ) {
                        self.moveColumnGroup( domIndex( row$ ), oldPos, newPos );
                    } else {
                        self.moveColumn( self.columns[oldPos], newPos );
                    }
                    self.focus();
                }
            });
        },

        _initColumnResize: function() {
            var helperWidth, startScrollOffset, isFrozen,
                self = this,
                dir = this.element.hasClass( C_RTL ) ? -1 : 1,
                header$ = this.element.find( SEL_GRID_HEADER ),
                row$ = header$.find( "tr" ).last();

            this.columnWidthHandle$ = $( "<div class='a-GV-columnWidthHandle'></div>" )
                .appendTo( header$ )
                .css( "top", row$[0].offsetTop )
                .height( row$.height() );
            this.columnWidthHandle$.draggable({
                axis: "x",
                containment: header$,
                start: function( event, ui ) {
                    self.columnWidthDragging = true;
                    self.columnWidthCancel = false;
                    self.columnWidthStartPos = ui.position.left;
                    self.columnWidthStartWidth = self.columnWidthCell$.first().outerWidth();
                    helperWidth = ui.helper.outerWidth();
                    // setup to scroll as needed when drag near boundaries
                    self.hdrScrollParent = self.element.find( ".a-GV-w-hdr" )[0];
                    isFrozen = self.columnWidthCell$.parent().parent().parent().parent()[0] !== self.hdrScrollParent;
                    startScrollOffset = self.hdrScrollParent.scrollLeft;
                    escapeOn(function() {
                        self.columnWidthCancel = true;
                    });
                },
                drag: function( event, ui ) {

                    function updateWidth( scroll ) {
                        var spWidth, deltaW,
                            width = self.columnWidthStartWidth +
                                dir * ( ui.position.left - self.columnWidthStartPos );

                        spWidth = self.hdrScrollParent.scrollWidth;
                        if ( !isFrozen ) {
                            width += ( self.hdrScrollParent.scrollLeft - startScrollOffset );
                        }

                        if ( width > MIN_COL_WIDTH ) {
                            if ( self.columnWidthCell$.hasClass( C_GRID_ROW_HEADER ) ) {
                                self.rowHeaderWidth = width;
                                self._calculateColumnWidths();
                            } else {
                                // update the width but without notification
                                self._updateColumnWidth( self._getColumnMetadata( self.columnWidthCell$.first() ), width );
                            }
                        }
                        // this is an attempt to make shrinking the width of a column that is at the right edge of the
                        // view port resize better
                        deltaW = self.hdrScrollParent.scrollWidth - spWidth;
                        if ( deltaW < 0 && !scroll && self.hdrScrollParent.scrollLeft !== startScrollOffset) {
                            self.hdrScrollParent.scrollLeft += deltaW;
                            startScrollOffset += deltaW;
                        }
                    }

                    if ( self.columnWidthCancel ) {
                        event.preventDefault();
                        return;
                    }

                    // Need to manually scroll. Draggable scrolling doesn't work because the header
                    // is overflow hidden and scrolls because of the table body scroll bar
                    // start or stop scrolling as needed. The actual scrolling happens from a timer
                    updateWidth();
                    if ( self._headerScrollCheck( ui.offset.left, helperWidth )) {
                        if ( !self.scrollTimerId ) {
                            self._headerScrollStart( ui.offset.left, helperWidth, function() { updateWidth( true ); } );
                        }
                    } else {
                        if ( self.scrollTimerId ) {
                            self._headerScrollStop();
                        }
                    }
                },
                stop: function( event, ui ) {
                    var width = self.columnWidthStartWidth +
                            dir * ( ui.position.left - self.columnWidthStartPos );

                    if ( !isFrozen ) {
                        width += ( self.hdrScrollParent.scrollLeft - startScrollOffset );
                    }

                    self.columnWidthDragging = false;
                    escapeOff();
                    if ( self.scrollTimerId ) {
                        self._headerScrollStop();
                    }
                    if ( self.columnWidthCancel ) {
                        width = self.columnWidthStartWidth;
                    }
                    if ( width < MIN_COL_WIDTH ) {
                        width = MIN_COL_WIDTH;
                    }

                    if ( self.columnWidthCell$.hasClass( C_GRID_ROW_HEADER ) ) {
                        self.rowHeaderWidth = width;
                        self._calculateColumnWidths( null, true );
                    } else {
                        // set width one more time with notification
                        self.setColumnWidth( self._getColumnMetadata( self.columnWidthCell$.first() ), width );
                    }
                }
            });
        },

        _updateColumnWidth: function( pColumn, pWidth, finalSet ) {
            var col, col$, index,
                columnName = this._columnName( pColumn );

            col = this.options.columns[0][columnName];
            col.width = pWidth;
            col.curWidth = pWidth;
            index = this.columns.indexOf( col );
            col$ = this._columnFromIndex( index );
            this._calculateColumnWidths( col$, finalSet );

            return [col, col$];
        },

        // col$ is optional
        _calculateColumnWidths: function( col$, finalSet ) {
            var i, col, w, fw, availWidth, affectedTable, body$, extra, remainingWidth, setColumn,
                ctrl$ = this.element,
                self = this;

            function lastVisibleColumn( start ) {
                var i, col;

                for ( i = start + 1; i < self.columns.length; i++ ) {
                    col = self.columns[i];
                    if ( !col.hidden && !col.noStretch && i !== setColumn ) {
                        return false;
                    }
                }
                return true;
            }

            function adjustTableColumnWidths( data$, header$ ) {
                var tableWidth,
                    headerCols$ = header$.parent().find( "col" ),
                    dataCols$ = data$.parent().find( "col" );

                tableWidth = 0;
                dataCols$.each(function( i ) {
                    var w, col, col$ = $( this ),
                        colIndex = col$.attr( ATTR_DATA_IDX );

                    if ( colIndex ) {
                        col = self.columns[ colIndex ];
                        w = col.curWidth;
                    } else {
                        w = self.rowHeaderWidth;
                    }
                    col$[0].width = w;
                    headerCols$[i].width = w;
                    tableWidth += w;
                });
                header$.parent().width( tableWidth );
                data$.parent().width( tableWidth );
                return tableWidth;
            }

            if ( col$ ) {
                affectedTable = col$.parent().parent()[0];
                setColumn = parseInt( col$.attr( ATTR_DATA_IDX ), 10 );
            }

            // if there are frozen columns and no specific column or if calculation is due to a particular column and that column is frozen
            if ( this.tableFrozenData$ && ( !affectedTable || affectedTable === this.tableFrozenHeader$[0] ) ) {
                adjustTableColumnWidths( this.tableFrozenData$, this.tableFrozenHeader$ );
                // need to set widths so the frozen and normal tables are side by side
                // this is also done in resize
                fw = this.tableFrozenData$.parent().outerWidth();
                ctrl$.find( ".a-GV-w-frozen" ).width( fw );
                w = ctrl$.width() - fw;
                ctrl$.find( ".a-GV-w-scroll,.a-GV-w-hdr" ).width( w );
            }

            // if no specific column or calculation is due to a particular column and that column is not frozen
            if ( !affectedTable || affectedTable === this.tableHeader$[0] ) {
                w = adjustTableColumnWidths( this.tableData$, this.tableHeader$ );
            } else {
                w = this.tableData$.parent().width();
            }

            if ( finalSet ) {
                availWidth = ctrl$.find( ".a-GV-w-scroll" )[0].clientWidth; // doesn't include scroll bar
                if ( w < availWidth ) {
                    availWidth = availWidth - w - 1;
                    if ( availWidth > 10 ) {
                        // there is enough available width to distribute it among all the columns
                        remainingWidth = availWidth;
                        for ( i = 0; i < this.columns.length; i++ ) {
                            col = this.columns[i];
                            if ( col.hidden || col.frozen || col.noStretch || i === setColumn ) {
                                continue;
                            }
                            if ( lastVisibleColumn( i ) ) {
                                // use remainingWidth so there are no rounding errors to cause a gap
                                col.curWidth += remainingWidth;
                            } else {
                                extra = Math.floor( availWidth * ( col.curWidth / w ) );
                                col.curWidth += extra;
                                remainingWidth -= extra;
                            }
                        }
                    } else {
                        // just give it to the last visible column that can be stretched
                        for ( i = this.columns.length - 1; i > 0; i-- ) {
                            col = this.columns[i];
                            if ( col.hidden || col.frozen || col.noStretch || i === setColumn ) {
                                continue;
                            }
                            col.curWidth = col.curWidth + availWidth;
                            break;
                        }
                    }
                    adjustTableColumnWidths( this.tableData$, this.tableHeader$ );
                }
                // xxx the above stretches the widths but if there are any columns without explicit widths they
                // could need to be contracted especially given the rough initial estimate of curWidth.
            }

            if ( this.options.hasSize ) {
                body$ = this.element.children( SEL_GRID_BODY );
                this._adjustSizeForScrollBars( body$.height(),
                    body$.find( ".a-GV-w-scroll" ).width(),
                    this.tableData$.prev().height() );
            }
        },

        _renderGrid: function() {
            var i, tb$, hasFrozenColumns, numSortColumns,
                self = this,
                o = this.options,
                out = util.htmlBuilder(),
                BEGIN_TABLE = "<table role='presentation' class='" + C_GRID_TABLE + "'>",
                BEGIN_THEAD = "<thead role='rowgroup'>",
                // the tab traps help implement tabbing between rows when there is frozen columns
                TAB_TRAP = "<input type='text' class='" + C_TAB_TRAP + " u-vh is-focusable' tabindex='-1'>";

            function renderColumns( out, frozen ) {
                var i, col;

                if ( frozen && o.rowHeader !== NONE ) {
                    out.markup( "<col" )
                        .optionalAttr( "width", self.rowHeaderWidth )
                        .markup( "/>" );
                }
                for ( i = 0; i < self.columns.length; i++) {
                    col = self.columns[i];

                    if ( ( col.frozen || false ) !== frozen || col.hidden ) {
                        continue;
                    }
                    out.markup( "<col" )
                        .optionalAttr( "width", col.width )
                        .attr( ATTR_DATA_IDX, i )
                        .markup( "/>" );
                }
            }

            function renderColumnHeaders( out, frozen, hidden ) {
                var i, g, col, cls, sortTitle, group, ariaSort, style,
                    columns = self.columns,
                    start = self.lastFrozenColIndex + 1,
                    end = columns.length;

                // first do the column groups if any
                for ( g = 0; g < self.columnGroups.length; g++) {
                    group = self.columnGroups[g];
                    out.markup( "<tr class='" + C_GRID_ROW + "' role='row'>" );
                    for ( i = 0; i < group.length; i++ ) {
                        if ( frozen && o.rowHeader !== NONE ) {
                            out.markup( "<th scope='colgroup' role='columnheader' class='" + C_GRID_COL_GROUP + " " + C_GRID_ROW_HEADER + "'" );
                            if ( hidden ) {
                                out.markup( " style='visibility: hidden;'" );
                            }
                            out.markup( ">" );
                        }
                        for ( i = 0; i < group.length; i++) {
                            col = group[i];
                            if ( ( col.frozen || false ) !== frozen ) {
                                continue;
                            }
                            cls = C_GRID_COL_GROUP;
                            cls += alignmentClass( col.headingAlignment || "center", true );

                            out.markup( "<th scope='colgroup' role='columnheader' " )
                                // xxx want an id here
                                .attr( "class", cls )
                                .attr( ATTR_COLSPAN, col.span )
                                .attr( ATTR_DATA_IDX, i );
                            if ( hidden ) {
                                out.markup( " style='visibility: hidden;'" );
                            }
                            out.markup( " tabindex='-1'><span class='" + C_GRID_HEADER_LABEL + "'>" )
                                .markup( col.heading || "" )
                                .markup( "</span></th>");
                        }
                    }
                    out.markup( "</tr>" );
                }

                // next do the column headings
                out.markup( "<tr class='" + C_GRID_ROW + "' role='row'>" );

                if ( frozen && o.rowHeader !== NONE ) {
                    style = null;
                    if ( hidden ) {
                        style += "visibility: hidden;";
                    }

                    out.markup( "<th scope='col' role='columnheader' class='" + C_GRID_COL_HEADER + " " + C_GRID_ROW_HEADER + "'" )
                        .optionalAttr( "style", style )
                        .markup( ">" );
                    // xxx accessible label for row header heading?
                    if ( o.rowHeaderCheckbox && o.multiple && o.selectAll ) {
                        out.markup( "<span class='a-GV-rowSelector'><input class='" + C_GRID_ROW_SELECT + "' tabindex='-1' type='checkbox'/></span>" );
                    }
                    if ( o.rowHeader === SEQUENCE ) {
                        out.markup( "<span class='a-GV-rownum'></span>" );
                    } else if ( o.rowHeader === LABEL ) {
                        out.markup( "<span class='a-GV-rowLabel'></span>" );
                    }
                    out.markup( "</th>" );
                }
                if ( frozen ) {
                    start = 0;
                    end = self.lastFrozenColIndex + 1;
                }
                for ( i = start; i < end; i++) {
                    col = columns[i];
                    if ( !col.hidden ) {
                        cls = C_GRID_COL_HEADER;
                        cls += alignmentClass( col.headingAlignment || "center", true );

                        if ( col.isRequired ) {
                            cls += " is-required";
                        }
                        if ( !col.elementId || col.readonly ) {
                            cls += " is-readonly";
                        }
                        if ( col.headingCssClasses ) {
                            cls += " " + col.headingCssClasses;
                        }

                        // wai-aria spec says aria-sort should only apply to one header at a time so pick the primary sort column
                        ariaSort = null;
                        if ( col.sortIndex === 1 ) {
                            if ( col.sortDirection === "asc" ) {
                                ariaSort = "ascending";
                            } else if ( col.sortDirection === "desc" ) {
                                ariaSort = "descending";
                            }
                        }

                        style = null;
                        if ( hidden ) {
                            style = "visibility: hidden;";
                        }
//                        if ( col.width ) {
//                            style += "width: " + col.width + "px;";
//                        }
                        // todo could the column metadata include a class?
                        out.markup( "<th scope='col' role='columnheader' " )
                            .attr( "class", cls )
                            .attr( ATTR_DATA_IDX, i )
                            .optionalAttr( "style", style )
                            .optionalAttr( "aria-sort", ariaSort )
                            .optionalAttr( "aria-haspopup", o.activateColumnHeader ? "true" : null ) // if there is a handler on activate then assume there is a popup
                            .markup( " tabindex='-1'><span class='" + C_GRID_HEADER_LABEL + "'" )
                            .optionalAttr( "id", col.elementId ? col.elementId + "_HDR" : null )
                            .markup( ">" )
                            .markup( col.heading )
                            .markup( "</span>");
                        if ( col.sortDirection ) {
                            if ( numSortColumns > 1 ) {
                                sortTitle = lang.formatMessage( col.sortDirection === "asc" ? "APEX.GV.SORT_ASCENDING_ORDER" : "APEX.GV.SORT_DESCENDING_ORDER", col.sortIndex );
                            } else {
                                sortTitle = lang.getMessage( col.sortDirection === "asc" ? "APEX.GV.SORT_ASCENDING" : "APEX.GV.SORT_DESCENDING" );
                            }
                            out.markup( "<span class='a-GV-header-sort'><span" )
                                .attr( "class", "a-Icon " +
                                        ( col.sortDirection === "asc" ? "icon-irr-sort-asc" : "icon-irr-sort-desc" ) )
                                .attr( ATTR_TITLE, sortTitle )
                                .attr( ARIA_LABEL, sortTitle )
                                .markup( "></span>" );
                            if ( numSortColumns > 1 ) {
                                out.content( col.sortIndex ); // todo acc this gets read twice
                            }
                            out.markup( "</span>" );
                        }
                        out.markup( "</th>" );
                    }
                }

                out.markup( "</tr>" );
            }

            hasFrozenColumns = this.lastFrozenColIndex >= 0 || o.rowHeader !== NONE;

            numSortColumns = 0;
            for ( i = 0; i < this.columns.length; i++ ) {
                if ( this.columns[i].sortDirection ) {
                    numSortColumns += 1;
                }
            }

            out.markup( "<div class='" + C_GRID_HEADER + "'>" );
            if ( hasFrozenColumns ) {
                out.markup( "<div class='" + C_GRID_WRAP_FROZEN + "'>" + BEGIN_TABLE );
                renderColumns( out, true );
                out.markup( BEGIN_THEAD );
                renderColumnHeaders( out, true, false );
                out.markup( "</thead></table></div>" );
            }
            out.markup( "<div class='" + C_GRID_WRAP_HEADER + "'>" + BEGIN_TABLE );
            renderColumns( out, false );
            out.markup( BEGIN_THEAD );
            renderColumnHeaders( out, false, false );
            out.markup( "</thead></table></div></div><div class='" + C_GRID_BODY + "'>" );
            this._renderAltDataMessages( out );
            if ( hasFrozenColumns ) {
                out.markup( TAB_TRAP + "<div class='" + C_GRID_WRAP_FROZEN + "'>" + BEGIN_TABLE );
                renderColumns( out, true );
                out.markup( BEGIN_THEAD );
                renderColumnHeaders( out, true, true );
                out.markup( "</thead><tbody></tbody></table>" );
                this._renderLoadMore( out );
                out.markup( "</div>" + TAB_TRAP + TAB_TRAP );
            }

            // use tabindex=-1 because Firefox lets scrollable things take focus and that messes up navigation especially during editing
            out.markup( "<div class='" + C_GRID_WRAP_SCROLL + "' tabindex='-1'>" + BEGIN_TABLE );
            renderColumns( out, false );
            out.markup( BEGIN_THEAD );
            renderColumnHeaders( out, false, true );
            out.markup( "</thead><tbody></tbody></table>" );
            this._renderLoadMore( out );
            out.markup( "</div>" );

            if ( hasFrozenColumns || o.autoAddRecord ) {
                out.markup( TAB_TRAP );
            }

            out.markup( "</div>" );
            if ( o.editable ) {
                out.markup( "<div class='a-GV-floatingItem' tabindex='-1' style='display: none;'>" )
                    .markup( "<div class='a-GV-cellMoveHandle'></div>" )
                    .markup( "<button type='button' class='a-GV-expandCollapse' tabindex='-1'><span aria-hidden='true' class='a-Icon'></span></button>" +
                             TAB_TRAP + "<div class='" + C_FLOATING_ITEM_CONT + "'></div>" + TAB_TRAP + "</div>" );
            }
            this._renderFooter( out );

            if ( o.columnSort ) {
                out.markup( "<div class='" + C_GRID_COL_CONTROLS + "' style='display:none'>" );
                this._renderButton( out, "a-Button js-asc", "icon-up-chevron", lang.getMessage( "APEX.GV.SORT_ASCENDING" ) );
                this._renderButton( out, "a-Button js-desc", "icon-down-chevron", lang.getMessage( "APEX.GV.SORT_DESCENDING" ) );
                out.markup( "</div>" );
            }

            if ( o.reorderColumns ) {
                out.markup( "<div class='" + C_GRID_COL_HANDLE + "' style='display:none'></div>" )
                    .markup( "<div class='a-GV-colMoveMarker' style='display:none'>" )
                    .markup( "<span class='a-GV-colMoveMarker-x'></span></div>" );
            }

            this.element.html(out.toString());
            this.pageKey = null; // cause the pagination if any to get updated
            tb$ = this.element.find( "tbody,thead" );
            if ( hasFrozenColumns ) {
                this.tableFrozenHeader$ = tb$.eq(0);
                this.tableHeader$ = tb$.eq(1);
                this.tableFrozenData$ = tb$.eq(3);
                this.tableData$ = tb$.eq(5);
            } else {
                this.tableFrozenHeader$ = null;
                this.tableFrozenData$ = null;
                this.tableHeader$ = tb$.eq(0);
                this.tableData$ = tb$.eq(2);
            }
            this.columnControls$ = this.element.find( SEL_GRID_COL_CONTROLS );
            this.columnControls$.find( "button" ).attr( "tabindex", -1 );
            this.columnHandle$ = this.element.find( "." + C_GRID_COL_HANDLE );
            this.columnMoveMarker$ = this.element.find( ".a-GV-colMoveMarker" );
            this.floatingItem$ = this.element.find( ".a-GV-floatingItem" );
        },

        _calcSpanAllColumns: function( frozen ) {
            var i, span, start, end;

            span = 0;
            if ( frozen ) {
                start = 0;
                end = this.lastFrozenColIndex + 1;
                if ( this.options.rowHeader !== NONE ) {
                    span += 1;
                }
            } else {
                start = this.lastFrozenColIndex + 1;
                end = this.columns.length;
            }
            for ( i = start; i < end; i++ ) {
                if ( !this.columns[i].hidden ) {
                    span += 1;
                }
            }
            return span;
        },

        //
        // Begin tableModelViewBase overrides
        //

        _getHeaderHeight: function() {
            return this.element.find( SEL_GRID_HEADER ).outerHeight();
        },

        _getStickyTop: function() {
            if ( $.isFunction( this.options.stickyTop ) ) {
                return this.options.stickyTop();
            }
            return 0;
        },

        _selectedCount: function() {
            if ( this.options.persistSelection ) {
                // if selection is persisted in the model the ask the model for the count
                // in this case the selection could span multiple pages
                return this.model.getSelectedCount();
            } // else
            return this.tableData$.find( SEL_SELECTED ).length;
        },

        _deletedCount: function() {
            return this.tableData$.find( SEL_DELETED ).length;
        },

        _hasControlBreaks: function() {
            return this.breakColumns.length > 0;
        },

        _getRecordHeight: function() {
            var r$, rowHeight;

            // This only works well for fixed height rows
            // This can be called when the table is empty so add an empty row just to measure then clean it up.
            r$ = this.tableData$.find( "tr" ).filter( SEL_VISIBLE ).first();
            if ( !r$.length ) {
                r$ = $("<tr class='" + C_GRID_ROW + "'><td class='" + C_GRID_CELL + "'></td></tr>");
                r$ = this.tableData$.append(r$);
                rowHeight = r$.outerHeight();
                this.tableData$.empty(); // todo r$ got rid of tbody, but what if it wasn't empty?
            } else {
                rowHeight = r$.outerHeight();
            }
            return rowHeight;
        },

        _getDataRenderContext: function () {
            return {
                hasFrozenColumns: this.lastFrozenColIndex >= 0 || this.options.rowHeader !== NONE,
                dataOut: util.htmlBuilder(),
                frozenOut:  util.htmlBuilder()
            };
        },

        _getDataContainer: function() {
            if ( this.tableFrozenData$ ) {
                return $( [this.tableFrozenData$[0], this.tableData$[0]] );
            } // else
            return this.tableData$;
        },

        _insertFiller: function( out, curFiller$ ) {
            var filler$, frozenFiller$;

            if ( out.hasFrozenColumns ) {
                frozenFiller$ = $( out.frozenOut.toString() );
                if ( curFiller$ ) {
                    curFiller$.first().before( frozenFiller$ );
                } else {
                    this.tableFrozenData$.html( frozenFiller$ );
                }
            }
            filler$ = $( out.dataOut.toString() );
            if ( curFiller$ ) {
                curFiller$.last().before( filler$ );
            } else {
                this.tableData$.html( filler$ );
            }
            if ( out.hasFrozenColumns ) {
                filler$ = $( [frozenFiller$[0], filler$[0]] );
            }
            return filler$;
        },

        _insertData: function( out, offset, count, filler$, how ) {
            var body$, dataRows$, frozenRows$,
                self = this;

            // after the rendering any anchors or buttons (any tab stop) in cells need to have the tab stops disabled this happens in getRows$
            if ( out.hasFrozenColumns ) {
                frozenRows$ = getRows$( out.frozenOut );
            }
            dataRows$ = getRows$( out.dataOut );
            if ( !filler$ ) {
                if ( out.hasFrozenColumns ) {
                    this.tableFrozenData$.append( frozenRows$ );
                }
                this.tableData$.append( dataRows$ );
            } else {
                // else must have filler$ and how must be before or after
                if ( out.hasFrozenColumns ) {
                    filler$.first()[how]( frozenRows$ );
                }
                filler$.last()[how]( dataRows$ );
            }

            if ( this.options.hasSize ) {
                body$ = this.element.children( SEL_GRID_BODY );
                this._adjustSizeForScrollBars( body$.height(),
                    body$.find( ".a-GV-w-scroll" ).width(),
                    this.tableData$.prev().height() );
            }

            if ( this.gotoCellPending ) {
                setTimeout( function() {
                    self.gotoCell( self.gotoCellPending.record, self.gotoCellPending.column );
                }, 1);
            }

            this._trigger( EVENT_PAGE_CHANGE, null, {
                offset: offset,
                count: count
            });
        },

        _controlBreakLabel: function( rowItem ) {
            var i, col, value, columnItem,
                o = this.options,
                label = "";

            for ( i = 0; i < this.breakColumns.length; i++ ) {
                col = this.breakColumns[i];
                if ( i > 0 ) {
                    label += ", ";
                }
                value = this.model.getValue( rowItem, col.property );
                // this is a subset of what _renderFieldDataValue does, for example links and cell templates are not supported
                if ( value === null ) {
                    value = "";
                }
                // check to see if the model has a display value
                if ( typeof value === "object" && value.hasOwnProperty( "d" ) ) {
                    value = value.d;
                } else {
                    // otherwise if there is a column item it may have a display value
                    columnItem = this.columnItems[ col.property ];
                    if ( columnItem ) {
                        value = columnItem.item.displayValueFor( value );
                    }
                }
                if ( value === "" && o.showNullAs ) {
                    value = o.showNullAs;
                }
                // if escape is false then value can contain markup otherwise it is escaped as element content
                // note that the === false test is so that anything else defaults to true.
                if ( col.escape === false ) {
                    value = util.escapeHTML( value );
                }
                label += "<span class='a-GV-breakLabel'>" + ( col.label ? util.escapeHTML( col.label ) : col.heading ) + ":</span> <span class='a-GV-breakValue'> " + value + "</span>";
            }
            return label;
        },

        _renderBreakRecord: function ( out, expandControl, breakLabel ) {
            if ( out.hasFrozenColumns ) {
                this._renderBreakRow( out.frozenOut, true, expandControl, breakLabel );
                expandControl = false;
            }
            this._renderBreakRow( out.dataOut, false, expandControl, breakLabel );
        },

        _renderFillerRecord: function ( out, cssClass ) {
            if ( out.hasFrozenColumns ) {
                out.frozenOut.markup( "<tr class='" + cssClass + "'><td " )
                    .attr( ATTR_COLSPAN, this._calcSpanAllColumns( true ) )
                    .markup( "></td></tr>" );
            }
            out.dataOut.markup( "<tr class='" + cssClass + "'><td " )
                .attr( ATTR_COLSPAN, this._calcSpanAllColumns( false ) )
                .markup( "></td></tr>" );
        },

        _renderRecord: function( out, rowItem, index, id, meta ) {
            if ( out.hasFrozenColumns ) {
                this._renderRow( out.frozenOut, rowItem, index, id, meta, true );
            }
            this._renderRow( out.dataOut, rowItem, index, id, meta, false );
        },

        _insertRecord: function( row, record, id, meta, where ) {
            var newRow$, after$,
                hasFrozenColumns = this.lastFrozenColIndex >= 0 || this.options.rowHeader !== NONE,
                out = util.htmlBuilder(),
                inserted = [];

            if ( hasFrozenColumns ) {
                // An inserted row doesn't have an index because it would cause all the other indexes to shift down!
                this._renderRow( out, record, null, id, meta, true );
                newRow$ = getRows$( out );
                inserted.push( newRow$[0] );
                if ( where === "after" && row ) {
                    after$ = row.first();
                    after$.after( newRow$ );
                } else {
                    this.tableFrozenData$.prepend( newRow$ );
                }
                out.clear();
            }
            // An inserted row doesn't have an index because it would cause all the other indexes to shift down!
            this._renderRow( out, record, null, id, meta, false );
            newRow$ = getRows$( out );
            inserted.push( newRow$[0] );
            if ( where === "after" && row ) {
                after$ = row.last();
                after$.after( newRow$ );
            } else {
                this.tableData$.prepend( newRow$ );
            }
            return $(inserted);
        },

        _afterInsert: function( insertedElements ) {
            var i, col,
                focus = !!this.element.find( "." + C_FOCUSED ).length,
                cell$ = insertedElements[0].find(".a-GV-cell" ).first();

            // find the first editable cell
            for ( i = 0; i < this.columns.length; i++ ) {
                col = this.columns[i];
                if ( !col.hidden && !col.readonly && this.columnItems[col.property] ) {
                    cell$ = cellFromColumnIndex( insertedElements[0], this.fieldToColumnIndex[col.property] );
                    break;
                }
            }
            // select and edit
            this._select( cell$,  null, focus, false, false );
            if ( focus && this.options.editable && !this.editMode ) {
                this.setEditMode( true );
            }
        },

        _updateRecordField: function( row, record, field, meta ) {
            var cell$, columnItem, col, cellMeta,
                index = this.fieldToColumnIndex[field],
                out = util.htmlBuilder();

            if ( index === undefined ) {
                return; // there is no cell to update
            }
            cell$ = cellFromColumnIndex( row, index );
            // update just the changed cell
            if ( this.editMode && record === this.activeRecord && !this.ignoreItemChange ) {
                columnItem = this.columnItems[field];
                if ( columnItem ) {
                    this._setColumnItemValue( columnItem.item, record, field );
                }
            }
            if ( !cell$.hasClass( C_ACTIVE ) ) {
                // update the readonly/display value
                col = this.options.columns[0][field];
                cellMeta = null;
                if ( meta.fields ) {
                    cellMeta = meta.fields[col.property];
                }
                out = util.htmlBuilder();
                this._renderFieldDataValue( out, col, record, meta, cellMeta );
                cell$.html( out.toString() );
            }
        },

        _updateRecordState: function( row$, itemId, meta ) {
            var f, cellMeta, cell$, cls, index,
                noTooltip = !this.tooltipOptions;

            row$.removeClass( C_DELETED + " is-inserted is-updated is-error is-warning");
            cls = "";
            if ( meta.deleted ) {
                cls += " " + C_DELETED;
            } else {
                if ( meta.inserted ) {
                    cls += " is-inserted";
                } else if ( meta.updated ) {
                    cls += " is-updated";
                }
                // don't care about errors on deleted records
                if ( meta.error ) {
                    cls += " is-error";
                } else if ( meta.warning ) {
                    cls += "is-warning";
                }
            }
            if ( cls ) {
                row$.addClass( cls );
            }
            if ( noTooltip ) {
                // NOTE this assumes/requires that there be a row header in order to see these record level error messages
                if ( meta.message ) {
                    row$.first().find( ".a-GV-rowHeader" ).attr( ATTR_TITLE, meta.message );
                } else {
                    row$.first().find( ".a-GV-rowHeader" ).removeAttr( ATTR_TITLE );
                }
            }
            // todo highlight can't apply to updates right?
            // todo this may make volatile columns stale

            // handle field level metadata but don't care if the record is deleted
            if ( meta.fields ) {
                for ( f in meta.fields ) {
                    if ( meta.fields.hasOwnProperty( f ) ) {
                        cellMeta = meta.fields[f];
                        index = this.fieldToColumnIndex[f];
                        if ( index === undefined ) {
                            continue; // there is no cell for this field, move on
                        }
                        cell$ = cellFromColumnIndex( row$, index );

                        cell$.removeClass( "is-changed is-error is-warning");
                        if ( !meta.deleted ) {
                            cls = "";
                            if ( cellMeta.error ) {
                                cls += " is-error";
                            } else if ( cellMeta.warning ) {
                                cls += " is-warning";
                            } else if ( cellMeta.changed ) {
                                cls += " is-changed";
                            }
                            if ( cls ) {
                                cell$.addClass( cls );
                            }
                        }
                        if ( noTooltip ) {
                            if ( cellMeta.message && !meta.deleted ) {
                                cell$.attr( ATTR_TITLE, cellMeta.message );
                            } else {
                                cell$.removeAttr( ATTR_TITLE );
                            }
                        }
                    }
                }
            }
        },

        _removeRecord: function( row ) {
            row.remove();
        },

        _identityChanged: function( prevId, id ) {
            // need to update the dom and the activeRecordId if it applies
            var row$ = this._elementsFromRecordIds( [prevId] )[0];
            row$.attr( ATTR_DATA_ID, id );
            if ( this.activeRecordId === prevId ) {
                this.activeRecordId = id;
            }
        },

        _replaceRecord: function( row, rec, oldId, id, meta ) {
            var newRow$, wasSelected, wasFocused, rowIndex = null, colIndex = null,
                newRow = [],
                index = row.first().children().first().find( ".a-GV-rownum" ).text() || null,
                hasFrozenColumns = this.lastFrozenColIndex >= 0 || this.options.rowHeader !== NONE,
                out = util.htmlBuilder();

            this.activeRecord = null; // if any records are replaced don't take any chances of holding on to stale record
            if ( index ) {
                index = parseInt( index, 10 ) - 1;
            }

            wasSelected = row.hasClass( C_SELECTED );
            wasFocused = row.find( this.lastFocused ).length;
            if ( wasFocused ) {
                rowIndex = 0;
                colIndex = domIndex( getContainingGridCell$( this.lastFocused ) );
            }

            if ( hasFrozenColumns ) {
                if ( wasFocused && row.last().find( this.lastFocused ).length ) {
                    rowIndex = 1;
                }
                this._renderRow( out, rec, index, id, meta, true );
                newRow$ = getRows$( out );
                newRow.push( newRow$[0] );
                row.first().replaceWith( newRow$ );
                out.clear();
            }
            this._renderRow( out, rec, index, id, meta, false );
            newRow$ = getRows$( out );
            newRow.push( newRow$[0] );
            row.last().replaceWith( newRow$ );

            newRow$ = $( newRow );
            if ( wasFocused ) {
                this._select( cellFromColumnIndex( newRow$.eq( rowIndex ), colIndex ), {}, false, false );
            } else if ( wasSelected ) {
                newRow$.addClass( C_SELECTED );
            }
        },

        _removeFocus: function( row ) {
            var curRow$, cell$, next$;

            // If the row has focus then move it to the next row (or prev row if at end) keeping in the same column
            curRow$ = $( this.lastFocused ).closest( "tr" );
            if ( curRow$[0] === row[0] || ( row.length > 1 && curRow$[0] === row[1] ) ) {
                cell$ = getContainingDataCell$( this.lastFocused );
                next$ = this._getNextCellDown( cell$, 1 );
                if ( !next$ ) {
                    next$ = this._getPrevCellUp( cell$, 1 );
                }
                if ( next$ ) {
                    this._select( next$, {}, true, false );
                }
            }
        },

        _elementsFromRecordIds: function( ids ) {
            var i, frozenRows$,
                includeFrozen = true,
                rows = [],
                keys = {};

            for ( i = 0; i < ids.length; i++ ) {
                keys[ids[i]] = i;
                rows[i] = null;
            }
            if ( !this.tableFrozenData$ ) {
                includeFrozen = false;
            }
            if ( includeFrozen ) {
                frozenRows$ = this.tableFrozenData$.children();
            }
            this.tableData$.children().each(function( index ) {
                var row$ = $( this ),
                    id = row$.attr( ATTR_DATA_ID );

                if ( keys[id] !== undefined ) {
                    if ( includeFrozen ) {
                        rows[keys[id]] = $( [ frozenRows$[index], this ] );
                    } else {
                        rows[keys[id]] = row$;
                    }
                }
            });
            return rows;
        },

        //
        // End tableModelViewBase overrides
        //

        _renderBreakRow: function ( out, frozen, expandControl, breakLabel ) {
            var span;

            span = this._calcSpanAllColumns( frozen );

            out.markup( "<tr role='row' class='a-GV-controlBreak is-expanded'><th class='a-GV-controlBreakHeader' scope='colgroup'" )
                .attr( ATTR_COLSPAN, span )
                .markup( ">" );
            if ( expandControl ) {
                // xxx may have initial collapsed state
                // xxx a-Icon icon-right-arrow icon-down-arrow, aria-expanded
                this._renderButton( out, "a-Button js-toggleBreak", "icon-down-arrow", lang.getMessage( "APEX.GV.BREAK_COLLAPSE" ) );
            }
            out.markup( "<span class='a-GV-controlBreakLabel'>")
                .markup( breakLabel )
                .markup( "</span></th></tr>" );
        },

        _renderRow: function( out, rowItem, index, id, meta, frozen ) {
            var i, col, cls, highlight, cellMeta, cellType,
                title = null,
                rowHeaderTitle = null,
                cellTitle = null,
                o = this.options,
                columns = this.columns,
                start = this.lastFrozenColIndex + 1,
                end = columns.length,
                noTooltip = !this.tooltipOptions;

            cls = C_GRID_ROW;
            if ( meta.agg ) {
                cls += " is-aggregate";
                cls += " a-GV-aggregate--" + meta.agg;
                if ( meta.grandTotal ) {
                    cls += " is-grandTotal";
                    title = o.aggregateLabels[meta.agg].overallLabel;
                } else {
                    title = o.aggregateLabels[meta.agg].label;
                }
                // todo mark as potentially stale if the model has changes
            } else {
                if ( meta.sel ) {
                    cls += " " + C_SELECTED;
                }
                if ( meta.deleted ) {
                    cls += " " + C_DELETED;
                } else {
                    if ( meta.inserted ) {
                        cls += " is-inserted";
                    } else if ( meta.updated ) {
                        cls += " is-updated";
                    }
                    if ( meta.error ) {
                        cls += " is-error";
                    } else if ( meta.warning ) {
                        cls += " is-warning";
                    }
                    if ( meta.highlight ) {
                        highlight = o.highlights[meta.highlight];
                        if ( highlight && highlight.cssClass ) {
                            cls += " " + highlight.cssClass;
                        } else {
                            cls += " " + meta.highlight;
                        }
                    }
                    if ( meta.message && noTooltip ) {
                        rowHeaderTitle = meta.message;
                    }
                    if ( !this.model.allowEdit( rowItem ) ) {
                        cls += " is-readonly";
                    }
                }
            }
            out.markup( "<tr role='row'" )
                .attr( ATTR_DATA_ID, id)
                .attr( "class", cls )
                .optionalAttr( ATTR_TITLE, title )
                .optionalAttr( ARIA_SELECTED, meta.sel ? "true" : null )
                .optionalAttr( "style", this.controlBreakCollapsed ? "display:none;": null )
                .markup( ">" );

            if ( frozen && o.rowHeader !== NONE ) {
                out.markup( "<th role='rowheader'" )
                    .attr( "class",  C_GRID_CELL + " " + C_GRID_ROW_HEADER )
                    .optionalAttr( ATTR_TITLE, rowHeaderTitle )
                    .optionalAttr( "scope", ( meta.agg || o.rowHeader === LABEL) ? "row" : null ) // only include scope=row if there will be a label xxx should sequence count?
                    .markup( ">" );
                // xxx accessible label for row header heading?
                if ( o.rowHeaderCheckbox && !meta.agg ) {
                    out.markup( "<span" )
                        .attr( "class", o.multiple ? "a-GV-rowSelector" : "a-GV-rowSelector a-GV-rowSelector--single" )
                        .markup( "><input class='" + C_GRID_ROW_SELECT + "' tabindex='-1'" )
                        .attr( "type", o.multiple ? "checkbox" : "radio" )
                        .optionalBoolAttr( "checked", !!meta.sel)
                        .markup( "/></span>" );
                }
                if ( o.rowHeader === SEQUENCE && !meta.agg ) {
                    out.markup( "<span class='a-GV-rownum'>" )
                        .content( meta.serverOffset !== undefined ? meta.serverOffset + 1 : "" )
                        .markup( "</span>" );
                }
                if ( meta.agg || o.rowHeader === LABEL ) {
                    out.markup( "<span class='a-GV-rowLabel'>" );
                    if ( meta.agg ) {
                        out.content( o.aggregateLabels[meta.agg][ meta.grandTotal ? "overallLabel" : "label" ] );
                    } else if ( o.rowHeaderLabelColumn ) {
                        out.markup( this.model.getValue( rowItem, o.rowHeaderLabelColumn ) );
                    }
                    out.markup( "</span>" );
                }
                out.markup("</th>" );
            }
            if ( frozen ) {
                start = 0;
                end = this.lastFrozenColIndex + 1;
            }
            for ( i = start; i < end; i++) {
                col = columns[i];
                cellMeta = null;
                if ( meta.fields ) {
                    cellMeta = meta.fields[col.property];
                }
                if ( !col.hidden ) {
                    cls = C_GRID_CELL;
                    cls += alignmentClass( col.alignment );
                    if ( col.columnCssClasses ) {
                        cls += " " + col.columnCssClasses;
                    }
                    cellTitle = null;
                    if ( cellMeta ) {
                        if ( cellMeta.error ) {
                            cls += " is-error";
                        } else if ( cellMeta.warning ) {
                            cls += " is-warning";
                        } else if ( cellMeta.changed ) {
                            cls += " is-changed";
                        } else if ( cellMeta.highlight ) {
                            highlight = o.highlights[cellMeta.highlight];
                            if ( highlight && highlight.cssClass ) {
                                cls += " " + highlight.cssClass;
                            } else {
                                cls += " " + cellMeta.highlight;
                            }
                        }
                        if ( cellMeta.disabled ) {
                            cls += " is-disabled";
                        }
                        // a checksum on the cell means that it is readonly
                        if ( cellMeta.ck ) {
                            cls += " is-readonly";
                        }
                        // todo mark volatile columns as stale if the item is edited
                    }
                    if ( noTooltip ) {
                        if ( cellMeta && cellMeta.message ) {
                            cellTitle = cellMeta.message;
                        } else if ( meta.agg ) {
                            cellTitle = o.aggregateTooltips[ meta.agg + "|" + col.property ];
                        }
                    }

                    cellType = "td";
                    if ( col.usedAsRowHeader ) {
                        cellType = "th";
                    }

                    out.markup( "<" + cellType + " role='gridcell' tabindex='-1'" )
                        .attr( "class", cls )
                        .optionalAttr( ATTR_TITLE, cellTitle )
                        .optionalAttr( "scope", col.usedAsRowHeader ? "row" : null )
                        .markup( ">" );
                    this._renderFieldDataValue( out, col, rowItem, meta, cellMeta );
                    out.markup("</" + cellType + ">" );
                }
            }
            out.markup( "</tr>" );
        },

        _getColumnMetadata: function( cell$ ) {
            var index, headerTable$, hdr$;

            if ( cell$.hasClass( C_GRID_CELL ) ) {
                index = domIndex( cell$ );
                if ( cell$.closest( "." + C_GRID_WRAP_FROZEN ).length ) {
                    headerTable$ = this.tableFrozenHeader$;
                } else {
                    headerTable$ = this.tableHeader$;
                }
                hdr$ = headerTable$.children().last().children().eq( index );
            } else if ( cell$.hasClass( C_GRID_COL_HEADER ) ) {
                hdr$ = cell$;
            } else {
                return null;
            }
            index = parseInt( hdr$.attr( ATTR_DATA_IDX ), 10 );
            if ( !isNaN( index ) ) {
                return this.columns[index];
            } // else
            return null;
        },

        _activateColumnHeader: function( event, cell$ ) {
            var col = this._getColumnMetadata( cell$ );

            if ( !col.noHeaderActivate ) {
                this._trigger( EVENT_ACTIVATE_COLUMN_HEADER, event, {
                    header$: cell$,
                    column: col
                });
            }
        },

        _activateRow: function( cell$ ) {
            var i, columnItem, prop, activeRecord, activeRecordId, meta, editable,
                row$ = cell$.parent(),
                id = row$.attr( ATTR_DATA_ID );

            activeRecord = this.model.getRecord( id );
            activeRecordId = this.model.getRecordId( activeRecord );
            editable = this.model.allowEdit( activeRecord );
            meta = this.model.getRecordMetadata( activeRecordId );
            row$.addClass( C_ACTIVE ); // xxx shouldn't this add active class to both rows?
            // load all values
            this._beginSetColumnItems();
            for ( i = 0; i < this.columns.length; i++ ) {
                prop = this.columns[i].property;
                columnItem = this.columnItems[ prop ];
                if ( columnItem ) {
                    this._setColumnItemValue( columnItem.item, activeRecord, prop,
                        editable ? meta : null );
                }
            }

            // change detection for non-active columns is gated by activeRecord, which is cleared in _deactivateRow, so don't set until after values are set
            // do set before calling reinit incase there is a callback into this for getActiveRecordId
            this.activeRecord = activeRecord;
            this.activeRecordId = activeRecordId;
            this._endSetColumnItems();

            this._triggerBeginEditing( this.activeRecord, this.activeRecordId );
        },

        _deactivateRow: function( callback ) {
            var i, columnItem, prop, validity, meta, colMeta, isDisabled,
                row$ = this.element.find( ".a-GV-row.is-active" );

            if ( this.activeLockCount > 0 ) {
                this.activeUnlockCallback = callback;
                return;
            }
            row$.removeClass( C_ACTIVE );
            meta = this.model.getRecordMetadata(( this.activeRecordId ) );
            // do final validation mainly for the benefit of inserted rows
            // don't validate rows that are deleted or can't be edited
            if ( meta && !meta.deleted && this.model.allowEdit( this.activeRecord ) ) {
                for ( i = 0; i < this.columns.length; i++ ) {
                    prop = this.columns[i].property;
                    columnItem = this.columnItems[ prop ];

                    // TODO Only check validity for enabled and visible items
                    if ( columnItem ) {
                        isDisabled = columnItem.item.isDisabled();
                        if ( !isDisabled ) {
                            validity = columnItem.item.getValidity();
                            if ( !validity.valid ) {
                                this.model.setValidity( "error", this.activeRecordId, prop, columnItem.item.getValidationMessage() );
                            }
                        } else {
                            this.model.setValidity( "valid", this.activeRecordId, prop );
                        }
                        // update disabled metadata
                        colMeta = meta.fields ? meta.fields[prop] : null;
                        if ( ( colMeta && colMeta.hasOwnProperty( "disabled" ) ) || isDisabled ) {
                            if ( !colMeta ) {
                                if ( !meta.fields ) {
                                    meta.fields = {};
                                }
                                colMeta = {};
                                meta.fields[prop] = colMeta;
                            }
                            colMeta.disabled = isDisabled;
                            // todo think is a call to metadataChanged needed?
                        }
                    }
                }
            }
            this._triggerEndEditing( this.activeRecord, this.activeRecordId );
            this.activeRecord = null;
            this.activeRecordId = null;
            callback();
        },

        _beginDeactivate: function( cell$, focus ) {
            var self = this;

            this.deactivateDelayTimer = setTimeout(function() {
                self.lastEditModeCell = null;
                self._deactivateCell( cell$ );
                if ( focus ) {
                    // put focus back in cell without focusing the editable field.
                    self.dontFocusEditableItem = true;
                    cell$.focus();
                }
                self.deactivateDelayTimer = null;
            }, 10);
        },

        _clearDeactivate: function() {
            if ( this.deactivateDelayTimer ) {
                clearTimeout( this.deactivateDelayTimer );
                this.deactivateDelayTimer = null;
            }
        },

        // must only be called while in edit mode
        _activateCell: function( cell$ ) {
            var column, next$, prev$, popupSelector, floatingTrap, endTrap, oneCellInRow,
                o = this.options,
                self = this,
                columnItem = null,
                isBreak =  cell$.hasClass( C_GRID_BREAK_HEADER ),
                traps$ = this.tabTraps$;

            function activate() {
                column = self._getColumnMetadata( cell$ );
                if ( column ) {
                    columnItem = self.columnItems[column.property];
                }
                // Check if editing is allowed for this cell. Trust the cell class added during rendering because don't have access to the model cell metadata
                if ( columnItem && self.model.allowEdit( self.activeRecord ) && !column.readonly && !cell$.hasClass( "is-readonly" ) ) {
                    cell$.addClass( C_ACTIVE );
                    // If an item has a popup that is attached at the end of the body (anywhere outside this widget) then
                    // we need to know about the popup so that we can manage focus and know that the cell is still active.
                    // see also the focusin, focusout handlers
                    popupSelector = columnItem.item.getPopupSelector();
                    if ( popupSelector ) {
                        $( document.body ).on( "focusin.gridpopup", function( event ) {
                            // ignore focus changes within a popup
                            if ( $( event.target ).closest( popupSelector ).length ) {
                                self._clearDeactivate();
                            }
                        } ).on( "focusout.gridpopup", function( event ) {
                            var cell$;
                            if ( $( event.target ).closest( popupSelector ).length ) {
                                cell$ = $( self.lastEditModeCell );
                                self._beginDeactivate( cell$, true );
                            }
                        } );
                    }
                    if ( columnItem.element$.height() > cell$.height() ) {
                        cell$.empty();
                        self.floatingItem$.children( SEL_FLOATING_ITEM_CONT ).append( columnItem.element$ );
                        self._expandFloatingItem( cell$, true );
                    } else {
                        cell$.empty().append( columnItem.element$ );
                    }

                    self._activateColumnItem( columnItem, column.elementId + "_HDR" ); // xxx how to include column group and possibly control break and row header what if heading contains markup?

                    if ( !self.dontFocusEditableItem ) {
                        // item value has already been set when row was activated
                        columnItem.item.setFocus();
                    }
                    self.dontFocusEditableItem = false;
                }
            }

            this.lastEditModeCell = cell$[0];

            // if the current row is not selected (can happen because of tab into new row) then select it but don't focus.
            if ( !cell$.parent().hasClass( C_SELECTED ) ) {
                this._select( cell$, null, false, false );
            }

            // make next and previous cells tab stops
            this._setFocusable( cell$[0] ); // this cell has to have been focusable
            cell$.attr( "tabindex", -1 ); // but can't be a tabstop
            next$ = this._nextCellWithWrap( cell$ ).attr( "tabindex", 0 );
            prev$ = this._prevCellWithWrap( cell$ ).attr( "tabindex", 0 );
            endTrap = null;
            if ( this.tableFrozenData$ ) {
                endTrap = 3;
                floatingTrap = 5;
                traps$.eq( 0 ).prop( "tabIndex", prev$.length ? 0 : -1 );
                // when there is exactly one cell in a row of either the frozen or non-frozen table then the
                // next and prev tab stops will both be in the other table and the normal tab order breaks
                // that is what the 1 and 2 tabtraps are for.
                oneCellInRow = cell$.parent().children().length === 1;
                traps$.eq( 1 ).prop( "tabIndex", oneCellInRow ? 0 : -1 );
                traps$.eq( 2 ).prop( "tabIndex", oneCellInRow ? 0 : -1 );
            } else {
                if ( o.autoAddRecord ) {
                    endTrap = 0;
                    floatingTrap = 2;
                } else {
                    floatingTrap = 1;
                }
            }
            if ( endTrap !== null ) {
                if ( o.autoAddRecord && this.onLastPage && next$.length === 0) {
                    traps$.eq( endTrap ).prop( "tabIndex", 0 ).addClass( "js-autoAdd" );
                } else {
                    traps$.eq( endTrap ).prop( "tabIndex", next$.length ? 0 : -1 ).removeClass( "js-autoAdd" );
                }
            }
            if ( o.autoAddRecord && this.onLastPage && next$.length === 0) {
                traps$.eq( floatingTrap ).prop( "tabIndex", 0 )
                    .toggleClass( "js-last", !next$.length )
                    .addClass( "js-autoAdd" );
            } else {
                traps$.eq( floatingTrap ).prop( "tabIndex", next$.length ? 0 : -1 )
                    .toggleClass( "js-last", !next$.length )
                    .removeClass( "js-autoAdd" );
            }
            // the floating prev tab trap is still needed

            // if this cell is in a new row need to deactivate this row and activate the new one, or if no active row then need to activate the new row
            // but deactivate row is a possibly async operation (if the active row is locked).
            if ( !this.activeRecord || this.activeRecordId !== cell$.parent().attr( ATTR_DATA_ID ) ) {
                if ( this.activeRecord ) {
                    this._deactivateRow( function() {
                        if ( !isBreak ) {
                            self._activateRow( cell$ );
                            activate();
                        }
                    });
                    return;
                }
                if ( !isBreak ) {
                    this._activateRow( cell$ );
                }
            }
            if ( !isBreak ) {
                activate();
            }
        },

        _expandFloatingItem: function( cell$, initial ) {
            var column, columnItem,
                cellWidth = cell$.outerWidth();

            this.floatingItem$
                .css( "position", "" )
                .css( "min-width", cellWidth + "px" )
                .insertAfter( this.element.find( SEL_GRID_BODY ) ).show()
                .addClass( C_EXPANDED )
                .draggable( "enable" )
                .find( ".a-GV-expandCollapse span" ).removeClass( "icon-ig-expand" ).addClass( "icon-ig-restore" );
            this.floatingItem$.children( SEL_FLOATING_ITEM_CONT ).show()
                .css( {width: "", height: ""} );

            if ( initial ) {
                this.floatingItem$.position( {
                    my: "left top",
                    at: "left top",
                    of: cell$,
                    collision: "fit fit"
                } );
            } else {
                this.floatingItem$.css( { left: this.lastFloatingPos.left, top: this.lastFloatingPos.top } );
            }
            this.floatingItem$.find( SEL_TAB_TRAP ).not( ".js-last" ).attr( "tabindex", 0 );
            // focus the item
            column = this._getColumnMetadata( cell$ );
            if ( column ) {
                columnItem = this.columnItems[column.property];
            }
            columnItem.item.setFocus();
        },

        _collapseFloatingItem: function( cell$ ) {
            var w = cell$.innerWidth(),
                h = cell$.innerHeight();

            this.lastFloatingPos = { left: this.floatingItem$.css("left"), top: this.floatingItem$.css("top") };
            this.floatingItem$
                .css( {"position": "relative", "left": "", "top": ""} )
                .appendTo( cell$ )
                .draggable( "disable" )
                .removeClass( C_EXPANDED + " ui-state-disabled" ) // want it disabled for dragging but don't want the content to look disabled
                .find( ".a-GV-expandCollapse span" ).removeClass( "icon-ig-restore" ).addClass( "icon-ig-expand" );
            this.floatingItem$.children( SEL_FLOATING_ITEM_CONT ).hide()
                .width( w )
                .height( h );
            this.floatingItem$.find( ".a-GV-expandCollapse" ).focus();
            this.floatingItem$.find( SEL_TAB_TRAP ).not( ".js-last" ).attr( "tabindex", -1 );
        },

        // must only be called while in edit mode
        _deactivateCell: function( cell$ ) {
            var column, out, meta, cellMeta,
                columnItem = null;

            if ( cell$.length <= 0 ) {
                return;
            }

            // remove tab stops
            this._setFocusable( cell$[0] );
            this._nextCellWithWrap( cell$ ).attr( "tabindex", -1 );
            this._prevCellWithWrap( cell$ ).attr( "tabindex", -1 );
            this.tabTraps$.prop( "tabIndex", -1 );

            column = this._getColumnMetadata( cell$ );
            if ( column ) {
                columnItem = this.columnItems[column.property];
            }
            if ( columnItem && cell$.hasClass( C_ACTIVE ) ) {
                $( document.body ).off( ".gridpopup" );

                this._setModelValue( cell$, columnItem.item, this.activeRecord, column.property );

                meta = this.model.getRecordMetadata( this.activeRecordId );
                cellMeta = null;
                if ( meta.fields ) {
                    cellMeta = meta.fields[column.property];
                }

                if ( !this.debugDontDeactivateCell ) { // see debugCellEdit
                    this.floatingItem$.insertAfter( this.element.find( SEL_GRID_BODY ) );
                    this.floatingItem$.hide();

                    this._deactivateColumnItem( columnItem );

                    out = util.htmlBuilder();
                    this._renderFieldDataValue( out, column, this.activeRecord, meta, cellMeta );
                    cell$.html( out.toString() );
                } else {
                    cell$.addClass( C_FOCUSED + " " + C_ACTIVE ); // put this back to make it look like it can be edited
                }

                // Also need to update the cell and row state
                // we know that itemId is not used also cell$ must be in the lastRow$ and using parent() will
                // include the row from both tables.
                this._updateRecordState(  this.lastRow$.parent(), null, meta );
            }
        },

        /*
         * Focus a cell or some content of a cell.
         * el is usually a grid table cell including headers, group headers etc. but it could also be an element
         * such as a button or link inside a cell. _setFocusable will soon be called because of focusin handler.
         */
        _setFocus: function( el ) {
            el.tabIndex = 0;
            el.focus();
        },

        /*
         * Make a cell or contents of a cell focusable.
         * el is usually a grid table cell including headers, group headers etc. but it could also be an element
         * such as a button or link inside a cell. When a cell is made focusable all actionalble content (e.g. buttons) in the
         * cell is made a tab stop. Any previous cell tab stops are removed.
         */
        _setFocusable: function ( el ) {
            var column, cell$, focusableContentLen,
                tabContent = this.options.tabbableCellContent,
                newCell$ = getContainingTableCell$( el );

            if ( this.lastFocused ) {
                $( this.lastFocused ).removeAttr( ARIA_LBL_BY );
                cell$ = getContainingTableCell$( this.lastFocused );
                // if the cell has changed remove the previous tab stops
                if ( newCell$[0] !== cell$[0] ) {
                    cell$.add( cell$.find( tabContent ).not( ":disabled,.js-asc,.js-desc" ) ).prop( "tabIndex", -1 );
                }
            }
            // if the cell contains any buttons focus them
            focusableContentLen = newCell$.find( tabContent ).not( ":disabled,.js-asc,.js-desc" ).prop( "tabIndex", 0 ).length;
            if ( !focusableContentLen ) {
                // otherwise focus the cell
                newCell$.prop( "tabIndex", 0 );
            }
            // it is expected that el is either the cell or a link or button in the cell
            this.lastFocused = el;
            column = this._getColumnMetadata( newCell$ );
            if ( column && column.elementId ) {
                // todo ACC need to also include the rowheader and maybe control break
                $( el ).attr( ARIA_LBL_BY, column.elementId + "_HDR" );
            }
        },

        _getCellFocusable: function( cell$ ) {
            var a$ = cell$.find( this.options.tabbableCellContent ).not( ":disabled,.js-asc,.js-desc" );
            if ( a$.length ) {
                return a$[0];
            }
            return cell$[0];
        },

        _getPrevCellUp: function( cur$, n, noHeader ) {
            var row$,
                next$ = null,
                colIndex = null;

            if ( cur$.hasClass( C_GRID_BREAK_HEADER ) || cur$.hasClass( C_GRID_COL_GROUP )) {
                colIndex = this.lastColIndex;
            }
            if ( colIndex === null ) {
                colIndex = columnIndexFromCell( cur$ );
                this.lastColIndex = colIndex;
            }

            // find the nth previous row
            cur$.closest( SEL_ROW ).prevAll( SEL_VISIBLE ).each( function( i ) {
                var tr$ = $( this );
                // don't go past a filler row
                if ( tr$.hasClass( C_GRID_SCROLL_FILLER ) ) {
                    return false;
                }
                row$ = tr$;
                return i < n - 1;
            } );

            if ( !this.focusInHeader && !noHeader && !row$ ) {
                this.focusInHeader = true;
                if ( cur$.closest( "tr" ).closest( "table" ).parent().hasClass( C_GRID_WRAP_FROZEN ) ) {
                    row$ = this.tableFrozenHeader$.children().last();
                } else {
                    row$ = this.tableHeader$.children().last();
                }
            }
            if ( row$ ) {
                if ( row$.hasClass( C_CONTROL_BREAK )) {
                    next$ = row$.children().first();
                } else {
                    next$ = cellFromColumnIndex( row$, colIndex );
                }
            }
            return next$;
        },

        _getNextCellDown: function( cur$, n ) {
            var row$,
                next$ = null,
                colIndex = null;

            if ( cur$.hasClass( C_GRID_BREAK_HEADER ) || cur$.hasClass( C_GRID_COL_GROUP )) {
                colIndex = this.lastColIndex;
            }
            if ( colIndex === null ) {
                colIndex = columnIndexFromCell( cur$ );
                this.lastColIndex = colIndex;
            }

            // find the nth next row
            cur$.closest( SEL_ROW ).nextAll( SEL_VISIBLE ).each( function( i ) {
                var tr$ = $( this );
                // don't go past a filler row
                if ( tr$.hasClass( C_GRID_SCROLL_FILLER ) ) {
                    return false;
                }
                row$ = tr$;
                return i < n - 1;
            } );

            if ( this.focusInHeader && !row$ ) {
                this.focusInHeader = false;
                if ( cur$.closest( "tr" ).closest( "table" ).parent().hasClass( C_GRID_WRAP_FROZEN ) ) {
                    row$ = this.tableFrozenData$.children().first();
                } else {
                    row$ = this.tableData$.children().first();
                }
            }
            if ( row$ ) {
                if ( row$.hasClass( C_CONTROL_BREAK )) {
                    next$ = row$.children().first();
                } else {
                    next$ = cellFromColumnIndex( row$, colIndex );
                }
            }
            return next$;
        },

        _select: function ( cells$, event, focus, delayTrigger, noNotify ) {
            var prevSelected, start, end, temp, toFocus, rows$, selectableRows$, allRows$, checked,
                otherTable$, otherRows, id, modelIds, inSameRow,
                selectionMade = false,
                o = this.options,
                action = "set",
                self = this,
                prevSel$ = this.element.children( SEL_GRID_BODY ).find( SEL_SELECTED );

            // this just tracks the selection state in modelIds the model is updated later
            function updateModelSelectionState( rows$, selectState ) {
                var last;

                if ( o.persistSelection ) {
                    last = rows$.length;
                    if ( self.tableFrozenHeader$ ) {
                        last = last / 2;
                    }
                    rows$.each( function( i ) {
                        var id;

                        if ( i >= last ) {
                            return false; // break
                        }
                        id = $( this ).attr( ATTR_DATA_ID );
                        if ( selectState && modelIds[id] === false ) {
                            delete modelIds[id]; // unselect then select = do nothing
                        } else {
                            modelIds[id] = selectState;
                        }
                    } );
                }
            }

            if ( o.persistSelection ) {
                modelIds = {}; // stores model selection state true = select, false = unselect
            }

            // can't select something that isn't visible
            cells$ = cells$.filter( SEL_VISIBLE );
            rows$ = cells$.parent( "tr" );

            inSameRow = this.lastRow$.index( cells$.first() ) >= 0;

            // if there are frozen columns
            if ( this.tableFrozenHeader$ ) {
                // todo this could be expensive especially if there are cells from multiple rows and/or cells from each table.
                // need to add corresponding rows from the other table
                otherRows = [];
                rows$.each(function() {
                    otherTable$ = self._getOtherTable( $( this ) );
                    otherRows.push( otherTable$.children()[domIndex( $( this ))] );  // todo perf!
                });
                rows$ = rows$.add( $( otherRows ) );
            }

            if ( event && this.options.multiple ) {
                if ( event.type === "click" ) {
                    // control+click for Windows and command+click for Mac
                    if ( event.ctrlKey || event.metaKey ) {
                        action = "toggle";
                    } else if ( event.shiftKey ) {
                        action = "range";
                    }
                } else if ( event.type === "keydown" ) {
                    // Mac has no concept of toggle with the keyboard
                    if ( event.keyCode === keys.SPACE ) {
                        if ( event.ctrlKey ) {
                            action = "toggle";
                        } else if ( event.shiftKey ) {
                            action = "range";
                        } else {
                            action = "add";
                        }
                    } else if ( event.ctrlKey ) {
                        action = "none";
                    } else if ( event.shiftKey ) {
                        action = "range";
                    }
                }
            }

            if ( action === "range" && ( !this.selectAnchor || this.focusInHeader ) ) {
                action = "set"; // when there is no anchor (or in header) turn range selection into set
            }

            // if the cell is in the header don't change the selection
            if ( !this.focusInHeader ) {
                // clear out previous selection if needed
                if ( action === "set" || action === "range" ) {
                    if ( o.rowHeaderCheckbox ) {
                        prevSel$.find( SEL_GRID_ROW_SELECT ).each(function() {
                            this.checked = false;
                        });
                    }
                    prevSel$.removeClass( C_SELECTED ).removeAttr( ARIA_SELECTED );
                    if ( o.persistSelection ) {
                        this.model.clearSelection();
                    }
                }

                // perform selection action
                prevSelected = rows$.hasClass( C_SELECTED );
                selectableRows$ = rows$.filter( ".a-GV-row" ).not( ".is-aggregate" );
                if ( action === "set" ||  action === "add" || (action === "toggle" && !prevSelected) ) {
                    selectableRows$.addClass( C_SELECTED ).attr( ARIA_SELECTED, true );
                    this.selectAnchor = selectableRows$[0];
                    selectionMade = true;
                } else if ( action === "range" ) {
                    start = domIndex( $( this.selectAnchor )); // todo perf?
                    end = domIndex( rows$.last() );
                    if ( start > end ) {
                        temp = end;
                        end = start;
                        start = temp;
                    }
                    otherRows = [];
                    // todo perf?
                    this.tableData$.children().each( function( index ) {
                        if ( index >= start && index <= end && $( this ).hasClass( "a-GV-row" ) && !$( this ).hasClass( "is-aggregate" ) ) {
                            otherRows.push( this );
                        }
                    } );
                    if ( this.tableFrozenData$ ) {
                        this.tableFrozenData$.children().each( function( index ) {
                            if ( index >= start && index <= end && $( this ).hasClass( "a-GV-row" ) && !$( this ).hasClass( "is-aggregate" ) ) {
                                otherRows.push( this );
                            }
                        } );
                    }
                    selectableRows$ = $( otherRows );
                    selectableRows$.addClass( C_SELECTED ).attr( ARIA_SELECTED, true );
                    selectionMade = true;
                } else if ( action === "toggle" && prevSelected ) {
                    selectableRows$.removeClass( C_SELECTED ).removeAttr( ARIA_SELECTED );
                    this.selectAnchor = selectableRows$[0];
                    if ( o.rowHeaderCheckbox ) {
                        selectableRows$.find( SEL_GRID_ROW_SELECT ).each(function() {
                            this.checked = false;
                        });
                    }
                    updateModelSelectionState( selectableRows$, false );
                }

                if ( selectionMade ) {
                    if ( o.rowHeaderCheckbox ) {
                        selectableRows$.find( SEL_GRID_ROW_SELECT ).each(function() {
                            this.checked = true;
                        });
                    }
                    updateModelSelectionState( selectableRows$, true );
                }

                if ( o.rowHeaderCheckbox && o.multiple ) {
                    // set all checked
                    // xxx for now assume select all applies to all rows currently rendered
                    allRows$ = this.tableFrozenData$.children( ".a-GV-row" );
                    checked = (allRows$.length === allRows$.filter( SEL_SELECTED ).length);
                    this.tableFrozenHeader$.find( SEL_GRID_ROW_SELECT ).first().each(function() {
                        this.checked = checked;
                    } ).closest( SEL_GRID_ROW_HEADER ).toggleClass( C_SELECTED, checked );
                }

                if ( o.persistSelection ) {
                    // Update selection state in the model
                    for ( id in modelIds ) {
                        if ( modelIds.hasOwnProperty( id ) ) {
                            this.model.setSelectionState( id, modelIds[id] );
                        }
                    }
                }

            }

            // focus if needed
            if ( cells$.length > 0 ) {
                toFocus = this._getCellFocusable( cells$.first() );
                if ( focus ) {
                    this._setFocus( toFocus );
                } else {
                    this._setFocusable( toFocus );
                }
                // set the current row after focus has been set so that lastRow$ is correct during any cell or row deactivation
                if ( !inSameRow ) {
                    this._setCurrentRow( cells$[0] );
                }
                /*xxx this is not correct and doesn't seem to be needed
                // scroll into view if needed xxx
                sp = cells$.first().closest( ".a-GV-w-hdr,.a-GV-w-scroll" )[0];
                if ( sp ) {
                    spOffset = $(sp).offset();
                    glOffset = this.element.offset();
                    offset = cells$.first().offset();
                    console.log("xxx spOffset ", glOffset, spOffset, offset, sp.scrollLeft);
                    // Don't use scrollIntoView because it applies to all potentially scrollable ancestors, we only
                    // want to scroll within the immediate scroll parent.
                    // Chrome scrolls parents other than the immediate scroll parent even though it seems that it shouldn't
                    if ( ( offset.top < spOffset.top ) || ( offset.top > spOffset.top + sp.offsetHeight ) ) {
                        sp.scrollTop = offset.top - glOffset.top;
                    }
                    if ( ( offset.left + cells$[0].offsetWidth < spOffset.left ) || ( offset.left > spOffset.left + sp.offsetWidth ) )  {
                        sp.scrollLeft = offset.left - glOffset.left;
                        console.log("xxx set scroll left to " + sp.scrollLeft);
                    }
                }
                */
            }

            // notify if needed
            if ( !this.focusInHeader && ( action === "toggle" ||
                (action === "range" && !prevSelected) ||
                (action === "add" && !prevSelected) ||
                (action === "set" && (prevSel$[0] !== rows$[0] || prevSel$.length !== rows$.length)) ) ) {
                // use a timer to make sure the focus happens first and also throttle
                // rapid changes from keyboard navigation.
                if ( self.triggerTimerId ) {
                    clearTimeout( self.triggerTimerId );
                    self.triggerTimerId = null;
                }

                self.triggerTimerId = setTimeout( function () {
                    self._updateStatus();
                    self.triggerTimerId = null;

                    // don't fire selection change for click events when in navigation mode xxx does this apply when clicking on links in cells?
                    if ( !(noNotify || ( self.options.navigation && event.type === "click" )) ) {
                        self._trigger( EVENT_SELECTION_CHANGE, event );
                    }
                }, delayTrigger ? 350 : 1 );
            }
        },

        _getOtherTable: function( row$ ) {
            var otherTable$;

            if ( !row$.parents( SEL_GRID_BODY ).length ) {
                otherTable$ = row$.closest( "table" ).parent().hasClass( C_GRID_WRAP_FROZEN ) ? this.tableHeader$ : this.tableFrozenHeader$;
            } else {
                otherTable$ = row$.closest( "table" ).parent().hasClass( C_GRID_WRAP_FROZEN ) ? this.tableData$ : this.tableFrozenData$;
            }
            return otherTable$;
        },

        _setCurrentRow: function( cell ) {
            var cell$ = $( cell ),
                row$ = cell$.closest( "tr" );

            if ( this.tableFrozenHeader$ ) {
                row$ = row$.add( this._getOtherTable( row$ ).children().eq( domIndex( row$ ) ) ); // todo perf?
            }
            this.lastRow$ = row$.children();
        },

        _nextCellWithWrap: function( cell$ ) {
            var row$, otherTable$,
                nextCell$ = cell$.next();

            if ( !nextCell$.length ) {
                row$ = cell$.closest( "tr" );
                // if there are frozen columns
                if ( this.tableFrozenHeader$ ) {
                    otherTable$ = this._getOtherTable( row$ );
                    // if currently in last frozen column
                    if ( row$.closest( "table" ).parent().hasClass( C_GRID_WRAP_FROZEN ) ) {
                        // go to first non-frozen cell in same row of other table
                        nextCell$ = otherTable$.children().eq( domIndex( row$ ) ).children().first(); // todo perf?
                    } else {
                        // go to first cell in next row of other table
                        nextCell$ = otherTable$.children().eq( domIndex( row$ )  + 1 ).children().first(); // todo perf?
                    }
                } else {
                    // go to first cell in next row
                    nextCell$ = row$.next().children().first();
                }
            }
            return nextCell$;
        },

        _prevCellWithWrap: function( cell$ ) {
            var row$, otherTable$, index,
               prevCell$ = cell$.prev();

            if ( !prevCell$.length ) {
                row$ = cell$.closest( "tr" );
                // if there are frozen columns
                if ( this.tableFrozenHeader$ ) {
                    otherTable$ = this._getOtherTable( row$ );
                    index = domIndex( row$ ); // todo perf?
                    // if currently in first non-frozen column
                    if ( !row$.closest( "table" ).parent().hasClass( C_GRID_WRAP_FROZEN ) ) {
                        // go to last non-frozen cell in same row of other table
                        prevCell$ = otherTable$.children().eq( index ).children().last();
                    } else {
                        // go to last cell in prev row of other table
                        index -= 1;
                        if ( index >= 0 ) {
                            prevCell$ = otherTable$.children().eq( domIndex( row$ )  - 1 ).children().last(); // todo perf?
                        }
                    }
                } else {
                    // go to last cell in prev row
                    prevCell$ = row$.prev().children().last();
                }
            }
            return prevCell$;
        },

        _clearHoverStates: function() {
            this.element.find( SEL_HOVER ).removeClass( C_HOVER );
            this.columnControls$.hide();
            this.columnHandle$.hide();
        },

        _setRowHoverState: function( row$ ) {
            var otherTable$;

            this.element.find( SEL_HOVER ).removeClass( C_HOVER );
            row$.addClass( C_HOVER );
            // if there are frozen columns then include other table row
            if ( this.tableFrozenHeader$ ) {
                otherTable$ = this._getOtherTable( row$ );
                otherTable$.children().eq( domIndex( row$ ) ).addClass( C_HOVER ); // todo perf?
            }
        },

        /*
         * The possible directions are "asc" and "desc"
         * The possible actions are:
         * - set: set this column to given direction clearing all others
         * - clear: don't sort any columns
         * - remove: don't sort this column leave other columns alone
         * - add: sort this column in given direction leave other columns alone (column wasn't sorted)
         * - change: sort this column in given direction leave other columns alone (column was sorted)
         * How the action is determined:
         *  Column       | User action             | columnSortMultiple setting
         *  current sort | direction   | Shift Key | false     | true
         *  -----------------------------------------------------------------
         *  asc/desc     | asc/desc    | true      | clear     | remove
         *  asc/desc     | asc/desc    | false     | clear     | clear
         *  asc/desc     | desc/asc    | true      | set       | change
         *  asc/desc     | desc/asc    | false     | set       | set
         *  none         | desc/asc    | true      | set       | add
         *  none         | desc/asc    | false     | set       | set
         */
        _sortChange: function( event, cell$, dir ) {
            var same,
                action = "set",
                column = this._getColumnMetadata( cell$ );

            if ( column.canSort ) {
                same = dir === column.sortDirection;
                if ( this.options.columnSortMultiple ) {
                    if ( event.shiftKey ) {
                        action = "add";
                        if ( same ) {
                            action = "remove";
                        } else if ( !column.sortDirection ) {
                            action = "add";
                        } else {
                            action = "change";
                        }
                    } else {
                        if ( same ) {
                            action = "clear";
                        }
                    }
                } else {
                    if ( same ) {
                        action = "clear";
                    }
                }
                this._trigger( EVENT_SORT_CHANGE, event, {
                    header$: cell$,
                    column: column,
                    direction: dir,
                    action: action
                });
            }
        },

        // todo think it might be nice to include based on shift key if the sort will be additive and what number
        // the problem is that this widget has no idea what the caller will do with the sort event and also the
        // next sort number is not readily known
        // also think about if the action should show up on the status bar
        _updateColumnControls: function( column ) {
            var titleAsc, titleDesc,
                ascActive = false,
                descActive = false;

            if ( column.sortDirection === "asc" ) {
                titleAsc = lang.getMessage( "APEX.GV.SORT_OFF" );
                titleDesc = lang.getMessage( "APEX.GV.SORT_DESCENDING" );
                ascActive = true;
            } else if ( column.sortDirection === "desc" ) {
                titleAsc = lang.getMessage( "APEX.GV.SORT_ASCENDING" );
                titleDesc = lang.getMessage( "APEX.GV.SORT_OFF" );
                descActive = true;
            } else {
                titleAsc = lang.getMessage( "APEX.GV.SORT_ASCENDING" );
                titleDesc = lang.getMessage( "APEX.GV.SORT_DESCENDING" );
            }
            this.columnControls$.find( ".js-asc" )
                .attr( ATTR_TITLE, titleAsc ).attr( ARIA_LABEL, titleAsc ).toggleClass( C_ACTIVE, ascActive );
            this.columnControls$.find( ".js-desc" )
                .attr( ATTR_TITLE, titleDesc ).attr( ARIA_LABEL, titleDesc ).toggleClass( C_ACTIVE, descActive );
        },

        _adjustSizeForScrollBars: function( h, w, headerHeight ) {
            var c$ = this.element.children(),
                bodyFrozen$ = c$.filter( SEL_GRID_BODY ).children( "." + C_GRID_WRAP_FROZEN ),
                headerScroll$ = c$.filter( SEL_GRID_HEADER ).children(  ".a-GV-w-hdr" );

            // if content has a horizontal scrollbar Note this gets confused if overflow is scroll rather than auto
            if ( this.tableData$.parent()[0].clientWidth > this.tableData$.parent().parent()[0].clientWidth ) {
                h -= util.getScrollbarSize().height;
            }
            bodyFrozen$.height( h );
            // if content has a vertical scrollbar Note this gets confused if overflow is scroll rather than auto
            if ( this.tableData$.parent()[0].clientHeight - headerHeight > this.tableData$.parent().parent()[0].clientHeight ) {
                w -= util.getScrollbarSize().width;
            }
            headerScroll$.width( w );
        },

        //
        // Column Header scroll support
        //

        _headerScrollCheck: function( pos, width, update ) {
            var sLeft, scrolled, offsetLeft,
                deltaX = 0,
                sp = this.hdrScrollParent;

            if ( sp ) {
                offsetLeft = $( this.hdrScrollParent ).offset().left;

                if ( this.element.hasClass( C_RTL ) ) {
                    if ( pos - width - offsetLeft < HEADER_SCROLL_SENSITIVITY ) {
                        deltaX = -HEADER_SCROLL_SPEED;
                    } else if ( ( offsetLeft + sp.offsetWidth ) - pos < HEADER_SCROLL_SENSITIVITY && ( offsetLeft + sp.offsetWidth ) - pos > 0 ) {
                        deltaX = HEADER_SCROLL_SPEED;
                    }
                    if ( update && deltaX ) {
                        sLeft = sp.scrollLeft + deltaX;
                        if ( sLeft > 0 ) {
                            sp.scrollLeft = 0;
                            deltaX = 0;
                        } else if ( sLeft > sp.scrollWidth - sp.clientWidth ) {
                            sp.scrollLeft = sp.scrollWidth - sp.clientWidth;
                            deltaX = 0;
                        } else {
                            sp.scrollLeft = sLeft;
                        }
                    }
                } else {
                    if ( ( offsetLeft + sp.offsetWidth ) - ( pos + width ) < HEADER_SCROLL_SENSITIVITY ) {
                        deltaX = HEADER_SCROLL_SPEED;
                    } else if ( pos - offsetLeft < HEADER_SCROLL_SENSITIVITY && pos - offsetLeft > 0 ) {
                        deltaX = -HEADER_SCROLL_SPEED;
                    }
                    if ( update && deltaX ) {
                        sLeft = sp.scrollLeft + deltaX;
                        if ( sLeft < 0 ) {
                            sp.scrollLeft = 0;
                            deltaX = 0;
                        } else if ( sLeft > sp.scrollWidth - sp.clientWidth ) {
                            sp.scrollLeft = sp.scrollWidth - sp.clientWidth;
                            deltaX = 0;
                        } else {
                            sp.scrollLeft = sLeft;
                        }
                    }
                }
                scrolled = !!deltaX;
            }
            return scrolled;
        },

        _headerScrollStart: function( pos, width, callback ) {
            var self = this,
                timeIndex = 0,
                times = [ 151, 134, 119, 106, 95, 86, 79, 74, 71]; // 70 + (9 - i)^2

            function scroll() {
                self.scrollTimerId = setTimeout( function() {
                    if ( self._headerScrollCheck( pos, width, true ) ) {
                        callback();
                        scroll();
                    } else {
                        self._headerScrollStop();
                    }
                }, times[timeIndex] );
                if ( timeIndex < times.length - 1 ) {
                    timeIndex += 1;
                }
            }

            if ( this.scrollTimerId ) {
                this._headerScrollStop();
            }
            scroll();
        },

        _headerScrollStop: function() {
            clearTimeout( this.scrollTimerId );
            this.scrollTimerId = null;
        },

        _expandCollapseControlBreak: function( pRows, expand ) {
            var i, row$, otherTable$, title;

            function doRow( _, tr ) {
                var tr$ = $(tr);
                if ( tr$.hasClass( "a-GV-row" ) && !tr$.hasClass( "is-aggregate" ) ) {
                    tr$[ expand ? "show" : "hide" ]();
                } else {
                    return false;
                }
            }

            if ( !$.isArray( pRows ) ) {
                pRows = [ pRows ];
            }

            for ( i = 0; i < pRows.length; i++ ) {
                row$ = pRows[i];
                if ( row$.hasClass( C_CONTROL_BREAK ) ) {
                    if ( row$.length !== 2 && this.tableFrozenHeader$ ) {
                        otherTable$ = this._getOtherTable( row$ );
                        row$ = $([ row$[0], otherTable$.children()[domIndex( row$ )] ]); // todo perf?
                    }
                    row$.first().nextAll().each( doRow );
                    if ( row$.length === 2 ) {
                        row$.last().nextAll().each( doRow );
                    }
                    if ( expand ) {
                        title = lang.getMessage( "APEX.GV.BREAK_COLLAPSE" );
                        row$.addClass( C_EXPANDED )
                            .find( ".js-toggleBreak" ).attr( ARIA_EXPANDED, "true" )
                            .attr( ATTR_TITLE, title ).attr( ARIA_LABEL, title )
                            .children( ".a-Icon" ).removeClass( "icon-right-arrow" ).addClass("icon-down-arrow");
                    } else {
                        title = lang.getMessage( "APEX.GV.BREAK_EXPAND" );
                        row$.removeClass( C_EXPANDED )
                            .find( ".js-toggleBreak" ).attr( ARIA_EXPANDED, "false" )
                            .attr( ATTR_TITLE, title ).attr( ARIA_LABEL, title )
                            .children( ".a-Icon" ).removeClass( "icon-down-arrow" ).addClass("icon-right-arrow");
                    }
                }
            }
        },

        /**
         * This method is for developer debugging only.
         * When developing a item plug-in that is to be used in a grid cell it can be difficult to debug on CSS styling
         * because as soon as the cell looses focus the item is moved in the DOM to a hidden area. Calling this with
         * a true argument to turn on cell edit debugging leaves the item in the cell after it looses focus so that
         * the DOM and styles can be inspected with browsers developer tools. This is not effective for end user use.
         * Because the cell is not fully deactivated it can affect editing in general. Call it again with false to
         * turn off cell edit debugging.
         * @param {Boolean} value
         * @example
         * From the console
         * var view$ = ... // this is the grid widget jQuery object.
         * view$.grid("debugCellEdit", true)
         * // do your debugging
         * view$.grid("debugCellEdit", false)
         */
        debugCellEdit: function( value ) {
            this.debugDontDeactivateCell = !!value;
        }

        /* todo remove when no longer needed
        //
        // For debugging
        //
        dumpColumnWidths: function() {
            var self = this,
                ctrl$ = this.element,
                dataWidths  = "  data width: ",
                hdrWidths   = "header width: ",
                colWidths   = "   col width: ",
                dataCols$ = ctrl$.children( SEL_GRID_BODY ).find( "thead" ).find( ".a-GV-header" ),
                headerCols$ = ctrl$.children( SEL_GRID_HEADER ).find( "thead" ).find( ".a-GV-header" );

            dataCols$.each(function( i ) {
                var col$ = $( this ),
                    col = self.columns[ col$.attr(ATTR_DATA_IDX) ];

                if (i > 0) {
                    hdrWidths += ", ";
                    dataWidths += ", ";
                    colWidths += ", ";
                }
                hdrWidths += headerCols$.eq(i).width() + "/" + headerCols$.eq(i).outerWidth();
                dataWidths += col$.width() + "/" + col$.outerWidth();
                if ( col ) {
                    colWidths += (col.width || "-") + "/" + (col.curWidth || "-");
                } else {
                    colWidths += "- / -";
                }
            });
            console.log("Dump column widths:");
            console.log(colWidths);
            console.log(hdrWidths);
            console.log(dataWidths);
            ctrl$.find("table" ).each(function() {
                console.log("table widths: " + $(this ).width());
            });
        }
        */
    });

})( apex.util, apex.model, apex.debug, apex.lang, apex.item, apex.jQuery );
